/*
 * tkTreeTable.c 
 *
 * =====================================================================
 *                          Copyright (c) 1997
 *                       Cygnus Multix Software GmbH
 *                        Autor: Khamis Abuelkomboz
 *
 * This tool is modified to mirror a realy full tree. The procedures are
 * modified for quickly viewing, selecting, scrolling etc.
 * This tool is faster than the standard list of Tcl/Tk and is still
 * fast, when more items added to the tree table (100 000 and more),
 * when at least not a linear list is mirrored (without using accelerator)
 * and the cache is used.
 * Supported Feautures:
 * Selection:
 *     Fully compatible to the listbox widget, with "see", "activate", ...
 * Movement:
 *     Fully compatible to the listbox.
 * Toggle: 
 *     A sub tree can be toggled to view/hide it.
 * Images: -image
 *     This is modified to support images too.
 * Tab positions:
 *     With the flag "-tabs" or "-columns" you can add tab stops to
 *     place text on this tab stops. You must give the tab size and
 *     not the x-position like text widget. I think it's better so.
 * Search:
 *     compatible to the text widget and supports extra flag "-begins" to
 *     search only a number of characters.
 * Sort:
 *     This is a usefull method to resort the tree. you can give the
 *     column number to sort items on this. The good thing here, that
 *     every sub tree is sorted as local list and the tree will be not
 *     damaged after sorting, only indeces are changed.
 * Fullname:
 *     with the method 'fullname' you can get the full tree name from
 *     the item back to it's latest root.
 * Truncate:
 *     With this flag you can let the widget cut the rest of string
 *     in a tab stop, if the string is too long for it.
 * Fit States:
 *     -bestfit:
 *         Change tab stop sizes so that all texts fit in there tab stops.
 *     -autofit
 *         Change tab stop sizes so that the current displayed texts fit
 *         in there tab stops.
 * Toggle viewing a column:
 *     column toggle/view/hide column#:
 *         it's possible to hide a column in the tab stop list.
 *         when a cloumn is hidden, there is no text displayed
 *         on it's position and the tab stop size is changed to the value
 *         in 'tabsMinSpace'
 * Accelerator (Pathfinder):
 * Now the treetable is perfect in it's performance. It uses an accelerator
 * to speed up the display and finding an item, especialy when the treetable
 * contains some thouthends or some hundert thouthend of entries. It will
 * still be fast (too fast) like a russan atom roket.
 *
 * The treetable as 'a tree' is fast without this accelerator. But when we
 * use this treetable as list, the entries are localized by linear searching,
 * this is exactly the problem because the display and selection manipulating
 * (without accelerator) will be too slow for alot of entries.
 * That's meen this accelerator works only for the items on the first level
 * (root).
 *
 * This accelerator is disabled by default and must be enabled with
 * '-accelerator 1'
 *
 * It is different as the built in cache. The used cache is only the latest
 * handled items in the treetable.
 *
 * Color managment:
 *     Supports a wide level of selection/fore-, background and bitmap colors.
 *
 * Speed of Treetable:
 * n     = number of all items in the tree
 * depth = depth of the longest sub tree
 * nr    = number of root items (without parents)
 * ni    = max. number items of sons in a sub tree (only one level)
 * ns    = max. number of all items in a sub tree
 *
 * TreeTable without CACHE and without Accel.    Regular linear list
 *          One Item        All items:           One Item:  All items
 * Search:  O(nr + ni)      O(n * (nr + ni))     O(n)       O(n^2)
 * Insert:  O(nr + ni)      O(n * (nr + ni))     O(n)       O(n^2)
 * Delete:  O(nr + ni + ns) O(n * (nr + ni))     O(n)       O(n^2)
    0-end:  -               O(n)                 -          O(n)
 * Display: O(nr + ni)      O(n)                 O(n)       O(n)
 * Select:  O(nr + ni)      O(n)                 O(n)       O(n)
 * Toggle:  O(nr + ni)      -                    -          -   
 * Sort:    -               O(n^2)               -          -
 *
 * Speed of Treetable with accelerator:
 * a = accelerator step
 * Search/Insert/Delete/Display/Select/Toggle: O(nr / a + ni)
 *
 * As example a = 1000, nr = 300,000, ni = 0 (a list):
 * O(300,000/1000 + 0) = 300
 *
 **************************************************************************/

#include <config.h>

#include "tclInt.h"
#include "tkInt.h"

#include <ctype.h>

#ifndef HAVE_STRNCASECMP_DECL
extern int strncasecmp ();
#endif

/*#include "tkConfig.h"*/
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "default.h"
#include "tkTreeTable.h"

#define CONFIG_OPTIONS_ONLY	0x4000

#define ITEM_HIDDEN		-1
#define ITEM_VIEWED		-2
#define ITEM_UNKNOWN		-3
#define ITEM_TEXT		-4
#define ITEM_NOT_FOUND		-5

#define ACTIVE_LINE_WIDTH       2

#define Min(a,b) (a<b ? a : b)
#define Max(a,b) (a>b ? a : b)

/* search for a char. in a string and return it's pointer, otherwise NULL */
#define NextChar(q, p, c, idx) q = p; idx = 0; while (*q != c && *q) q++, idx++; if (*q == 0) q = NULL

#ifdef USE_CACHE
#define RET_CACHED_INDEX(itemPtr) \
{ \
    int i = 0; \
    while (i<MAX_CACHED) \
    { \
	if (tablePtr->cachedItem[i] == itemPtr) \
	{ \
	    return tablePtr->cachedPos[i]; \
	} \
	i++; \
    } \
}
	
#define VIEW_CACHE() \
{ \
    int i = 0; \
    while (i<MAX_CACHED) \
    { \
	fprintf (stderr, "cache[%i]=(%i, %s)\n", i, \
		tablePtr->cachedPos[i], \
		tablePtr->cachedItem[i] ? tablePtr->cachedItem[i]->text : "NULL"); \
	i++; \
    } \
}
	
#define SET_CACHED(item, pos) \
{ \
    if (pos >= 0) \
    { \
    	if ( \
        tablePtr->cachedItem [0] != item && \
	tablePtr->cachedItem [1] != item && \
        tablePtr->cachedItem [2] != item && \
	tablePtr->cachedItem [3] != item && \
	tablePtr->cachedItem [4] != item && \
	tablePtr->cachedItem [5] != item && \
	tablePtr->cachedItem [6] != item && \
	tablePtr->cachedItem [7] != item \
	) \
	{ \
	    tablePtr->cachedPos [tablePtr->cacheIndex] = pos; \
	    tablePtr->cachedItem[tablePtr->cacheIndex] = item; \
	    tablePtr->cacheIndex = (tablePtr->cacheIndex+1) % MAX_CACHED; \
	} \
    } \
}

#define RET_CACHED_ITEM(index) \
{ \
    int i = 0; \
    while (i<MAX_CACHED) \
    { \
	if (tablePtr->cachedItem[i] != NULL && tablePtr->cachedPos[i] == index) \
	{ \
	    return tablePtr->cachedItem[i]; \
	} \
	i++; \
    } \
}

#define GET_CACHED_ITEM(item, index) \
{ \
    int ii=0; \
    while (ii<MAX_CACHED) \
    { \
	if (tablePtr->cachedItem[ii] != NULL && tablePtr->cachedPos[ii] == index) \
	{ \
	    item = tablePtr->cachedItem[ii]; \
	} \
	ii++; \
    } \
}

#define FREE_CACHE() \
{ \
    int i=0; \
    while (i<MAX_CACHED) \
    { \
	tablePtr->cachedItem[i] = NULL; \
	tablePtr->cachedPos [i] = -1; \
	i++; \
    } \
    tablePtr->cacheIndex = 0; \
}

#define FREE_RONG_CACHE(index) \
{ \
    int i=0; \
    while (i<MAX_CACHED) \
    { \
	if (tablePtr->cachedPos [i] >= index) \
	{ \
	tablePtr->cachedItem[i] = NULL; \
	tablePtr->cachedPos [i] = -1; \
	} \
	i++; \
    } \
}
#else
#define RET_CACHED_INDEX(itemPtr)
#define SET_CACHED(itemPtr, pos)
#define RET_CACHED_ITEM(index)
#define GET_CACHED_ITEM(itemPtr, index)
#define FREE_RONG_CACHE(index)
#define FREE_CACHE()
#endif

#ifdef USE_PARENT_CACHE
#define FREE_PARENT_CACHE() \
    tablePtr->cachedParent = -5; \
    tablePtr->cachedParentPtr = NULL
#define FREE_RONG_PARENT_CACHE(num) \
    if (tablePtr->cachedParent >= num) \
    {\
	tablePtr->cachedParent = -5; \
	tablePtr->cachedParentPtr = NULL; \
    }
#define GET_CACHED_PARENT(tablePtr, index) \
    ((tablePtr->cachedParentPtr != NULL && tablePtr->cachedParent == index) \
	? tablePtr->cachedParentPtr \
	: TreeTableFindItem (tablePtr, index))
#define SET_PARENT_CACHE(Ptr, num) \
	    tablePtr->cachedParentPtr = Ptr, \
	    tablePtr->cachedParent    = num
#else
#define GET_CACHED_PARENT(tablePtr, index)
#define FREE_PARENT_CACHE()
#define FREE_RONG_PARENT_CACHE(num)
#define SET_PARENT_CACHE(Ptr, num)
#endif


#define TREETABLE_WIDTH(tablePtr)  (Tk_Width(tablePtr->tkwin) \
			- 2*(tablePtr->inset+tablePtr->selBorderWidth))
#define TREETABLE_NUM_LINES(tablePtr) \
		((Tk_Height(tablePtr->tkwin)-2*(tablePtr->borderWidth))) \
    		/ tablePtr->lineHeight;

/*
 * Treetable item coordinates
 */
#define ITEM_X(Ptr) (tablePtr->inset \
                     - tablePtr->xOffset \
                     + Ptr->indent*tablePtr->indentWidth)
/* the index in the following command is the virtual position and not the
 * realy position
 */
#define ITEM_Y(Ptr, vindex) (((vindex - tablePtr->topIndex) * tablePtr->lineHeight) \
				+ tablePtr->borderWidth \
				+ (tablePtr->lineHeight-Ptr->bitmapHeight)/2)

#define ITEM_HEIGHT()          tablePtr->lineHeight
#define ITEM_BITMAP_WIDTH(Ptr) (Ptr->bitmapWidth + tablePtr->bitmapSpace)
#define ITEM_TEXT_WIDTH(Ptr)   Ptr->lineWidth
#define ITEM_WIDTH(Ptr)        (ITEM_BITMAP_WIDTH(Ptr)+ITEM_TEXT_WIDTH(Ptr))

#define ITEM_TEXT_INDENT(itemPtr) (tablePtr->inset \
			+ itemPtr->indent*tablePtr->indentWidth \
			+ tablePtr->bitmapSpace \
			+ ((itemPtr->bitmap != None || itemPtr->image) \
	    			? itemPtr->bitmapWidth \
				: 0))
#define ITEM_TEXT_X(itemPtr) (ITEM_TEXT_INDENT(itemPtr)-tablePtr->xOffset)

#ifdef USE_PATHFINDER
#define FIND_IN_PATHFINDER(pos, retPtr, retPos) \
    if (tablePtr->pathFinder) \
    { \
    	PathFinder_t *pf; \
	for (pf=tablePtr->pathFinder->next; pf; pf=pf->next) \
	{ \
	    if (pf->itemPos < pos) \
	    { \
	    	retPos = pf->itemPos; \
		retPtr = pf->itemPtr; \
	    } \
	} \
    }
#define DELETE_FROM_PATHFINDER(pos) \
    if (tablePtr->pathFinder) \
    { \
    	PathFinder_t *pf, *pf1=tablePtr->pathFinder; \
	for (pf=pf1->next; pf; pf=pf->next) \
	{ \
	    if (pos >= pf->itemPos) \
	    { \
		pf1->next = NULL; \
		for (pf1=pf; pf1; ) \
		{ \
		    pf = pf1; \
		    pf1=pf1->next; \
		    free (pf); \
		} \
		break; \
	    } \
	    pf1 = pf; \
	} \
    }
#else
#define FIND_IN_PATHFINDER(pos, retPtr, retPos)
#define DELETE_FROM_PATHFINDER(pos)
#endif

#define TreeTableRedrawRange(tablePtr) \
    if ((tablePtr->tkwin != NULL) && Tk_IsMapped(tablePtr->tkwin) \
	&& ! (tablePtr->flags & REDRAW_PENDING)) \
    { \
	Tk_DoWhenIdle(DisplayTreeTable, (ClientData) tablePtr); \
	tablePtr->flags |= REDRAW_PENDING; \
    }

static int  TreeTableCmd _ANSI_ARGS_((ClientData clientData,
					 Tcl_Interp *interp,
					 int argc, char **argv));
static int  TreeTableWidgetCmd _ANSI_ARGS_((ClientData clientData,
						Tcl_Interp *interp,
						int argc, char **argv));
static void DestroyTreeTable _ANSI_ARGS_((char *memPtr));
static int  ConfigureTreeTable _ANSI_ARGS_((TreeTable *tablePtr,
						int argc, char **argv,
						int flags));
static void DisplayTreeTable _ANSI_ARGS_((ClientData clientData));
static int  TreeTableFindIndex _ANSI_ARGS_((TreeTable *tablePtr,
						TableItem *itemPtr));
static TableItem * TreeTableFindItem _ANSI_ARGS_((TreeTable *tablePtr, int index));
static int  ConfigureTreeTableItem _ANSI_ARGS_((
					     TreeTable *tablePtr,
					     TableItem *itemPtr,
					     int argc, char **argv,
					     int flags,
					     int mode));
static TableItem* TreeTableInsertItem _ANSI_ARGS_((TreeTable *tablePtr,
					  int index,
					  TableItem* prevPtr,
					  int argc,
					  char **argv));
static int  TreeTableInsertItems (TreeTable* tablePtr,
				 int index, char* itemlist,
				int argc, char **argv);
static void TreeTableDeleteRange _ANSI_ARGS_((TreeTable *tablePtr,
					   int start, int end));
static int  TreeTableRemoveItem _ANSI_ARGS_((TreeTable *tablePtr,
					  TableItem *itemPtr,
					  int lineNum,
					  int* width,
					  int original_item,
					  int children));
static void TreeTableFreeItem _ANSI_ARGS_((TreeTable *tablePtr,
					       TableItem *itemPtr));
static void TreeTableEventProc _ANSI_ARGS_((ClientData clientData,
						XEvent *eventPtr));
static int  GetTreeTableIndex _ANSI_ARGS_((TreeTable *tablePtr,
					       char *string, int endAfter,
					       int *indexPtr));
static void ChangeTreeTableView _ANSI_ARGS_((TreeTable *tablePtr,
						 int index, int redraw));
static void ChangeTreeTableOffset _ANSI_ARGS_((TreeTable *tablePtr,
						   int offset));
static void TreeTableScanTo _ANSI_ARGS_((TreeTable *tablePtr,
					     int x, int y));
static int  NearestTreeTableItem _ANSI_ARGS_((TreeTable *tablePtr,
						  int y));
static void TreeTableComputeLineHeight _ANSI_ARGS_((TreeTable *tablePtr, TableItem*thisPtr, int cnt));
static void TreeTableComputeWidths _ANSI_ARGS_((TreeTable *tablePtr,
						    TableItem *itemPtr,
						    int checkAll));

static int  TreeTableFetchSelection _ANSI_ARGS_((ClientData clientData,
						     int offset, char *buffer,
						     int maxBytes));
static void TreeTableLostSelection _ANSI_ARGS_((ClientData clientData));
#if 0
static void TreeTableRedrawRange (TreeTable *tablePtr);
#endif
static void TreeTableUpdateVScrollbar (TreeTable *tablePtr);
static void TreeTableUpdateHScrollbar (TreeTable *tablePtr);

static int  TreeTableIndex (TableItem *itemPtr, int index, int pos, TableItem **itemRet);
static int  TreeTableInsertListItem (TreeTable* tablePtr,
					TableItem*newPtr,
					TableItem*prevPtr,
					int index, int *newIndex);
static int  TreeTableFindIndex_x (TableItem *itemPtr, TableItem *srcPtr, int index, int* pos);
static int  TreeTableCountNotHidden (TreeTable*tablePtr, int flag);
static int  TreeTableLostSel (TreeTable* tablePtr, int first, int last);
static int  TreeTableLostSel_x (register TreeTable *tablePtr, TableItem* itemPtr,
				int *pos, int first, int last);
static int  TreeTableRealyIndex (register TreeTable *tablePtr, int index);
static int  DisplayRecursive (register TreeTable *tablePtr,
					register TableItem * itemPPtr,
					register int* pos,
					register int* rpos,
    					register int limit,
					register Pixmap pixmap,
					int calc);
static int TreeTableCountNotHidden_x (TableItem*itemPtr);
static void TreeTableImageProc(ClientData clientData,
				int x, int y,
				int width, int height,
				int imgWidth, int imgHeight);
static void TreeTableGetItems (TreeTable *tablePtr, int from, int to);
static int TreeTableSearchCmd (TreeTable *tablePtr,  int argc,
				char **argv);
static int TreeTableFindNotHiddenItem_x (TableItem *itemPtr,
					int index, int pos,
					TableItem **itemRet);
static int TreeTableViewedIndex (register TreeTable *tablePtr, int index);
static int TreeTableSort(TreeTable *tablePtr, int argc, char **argv);
static int TreeTableSelectFromTo (TreeTable*tablePtr, TableItem *itemPtr,
				int pos,
				int start, int end);
static int TreeTableGetBBox (TreeTable *tablePtr, int index, int *x, int *y, int *w, int *h);
static TableItem * TreeTableFindNotHiddenItem (TreeTable *tablePtr, int index);
static TableItem * TreeTableSortFindPrev(TableItem *listPtr, TableItem *itemPtr,
					int col, int noCase, int *index);
static void ComputeTabStops (TreeTable*tablePtr, TableItem *itemPtr);

static int TreeTableToggle     (TreeTable* tablePtr, int xx, int yy, int testonly);
static int TreeTableToggleSubTree (TreeTable* tablePtr, TableItem* itemPtr, int* rpos, int* pos, int xx, int yy, int testonly);
static int TreeTableToggleItem (TreeTable* tablePtr, TableItem *itemPtr, int hide);
static void TreeTableToggleIt   (TreeTable* tablePtr, TableItem*Ptr);
					
void ViewArgs (char *reason, int argc, char *argv[], int mode);

#undef TEST_TREE
#ifdef TEST_TREE
#ifdef USE_CACHE
void ViewCache(TreeTable*tablePtr);
#endif
#endif

static int InsertedNewPosition = 0;

/*
 * Information used for argv parsing:
 */
static Tk_ConfigSpec itemConfigSpecs[] = {
    {TK_CONFIG_PIXELS, "-linewidth", "lineWidth", "LineWidth",
       NO_TABLEITEM_LINEWIDTH, Tk_Offset(TableItem, lineWidth), 0},
       
    {TK_CONFIG_STRING, "-text", "text", "Text",
       DEF_TABLEITEM_TEXT, Tk_Offset(TableItem, text), 
       TK_CONFIG_NULL_OK},
    {TK_CONFIG_BITMAP, "-bitmap", "bitmap", "Bitmap",
       DEF_TABLEITEM_BITMAP, Tk_Offset(TableItem, bitmap), 
       TK_CONFIG_NULL_OK},
       
    /* image */
    {TK_CONFIG_STRING, "-image", (char *) NULL, (char *) NULL,
	DEF_TABLEITEM_IMAGE, Tk_Offset(TableItem, imageString),
	TK_CONFIG_NULL_OK},

       
    {TK_CONFIG_FONT, "-font", "font", "Font",
       DEF_TABLEITEM_FONT, Tk_Offset(TableItem, fontPtr), 
       TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-textforeground", "textForeground", "Background",
       DEF_TABLEITEM_TEXTFOREGROUND, Tk_Offset(TableItem, textFgColor), 
       TK_CONFIG_NULL_OK},
       
#if USE_BITMAP_COLORS
    {TK_CONFIG_COLOR, "-bitmapforeground", "bitmapForeground", "Background",
       DEF_TABLEITEM_BITMAPFOREGROUND, Tk_Offset(TableItem, bitmapFgColor), 
       TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-bitmapbackground", "bitmapBackground", "Background",
       DEF_TABLEITEM_BITMAPBACKGROUND, Tk_Offset(TableItem, bitmapBgColor), 
       TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-bitmapselectfg", "bitmapSelectFg", "Background",
       DEF_TABLEITEM_BITMAPSELECTFOREGROUND, 
       Tk_Offset(TableItem, bitmapSelectFgColor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-bitmapselectbg", "bitmapSelectBg", "Background",
       DEF_TABLEITEM_BITMAPSELECTBACKGROUND,
       Tk_Offset(TableItem, bitmapSelectBgColor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-lineforeground", "lineForeground", "Background",
       DEF_TABLEITEM_LINEFOREGROUND, Tk_Offset(TableItem, lineFgColor), 
       TK_CONFIG_NULL_OK},
#endif
    {TK_CONFIG_INT, "-indent", "indent", "Indent",
       DEF_TABLEITEM_INDENT, Tk_Offset(TableItem, indent), 0},
       
    {TK_CONFIG_INT, "-parent", "parent", "Parent",
       DEF_TABLEITEM_PARENT, Tk_Offset(TableItem, parent), 0},
       
    {TK_CONFIG_INT, "-succnum", "succnum", "SuccNum",
       DEF_TABLEITEM_SUCCNUM, Tk_Offset(TableItem, succNum), 0},
    {TK_CONFIG_SYNONYM, "-children", "succnum", (char *) NULL,
       (char *) NULL, 0, 0},
    
    {TK_CONFIG_INT, "-seennum", "seennum", "SeenNum",
       DEF_TABLEITEM_SEENNUM, Tk_Offset(TableItem, seenNum), 0},
    /*
     * Client data
     */
    {TK_CONFIG_STRING, "-data", "data", "data",
       DEF_TABLEITEM_DATA, Tk_Offset(TableItem, data),
       TK_CONFIG_NULL_OK},
       
#if 0
    {TK_CONFIG_INT, "-flags", "flags", "Flags",
       DEF_TABLEITEM_FLAGS, Tk_Offset(TableItem, flags), 0},
#endif

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}
};

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background",
				DEF_TREETABLE_BG_COLOR,
				Tk_Offset(TreeTable, normalBorder),
				TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-background", "background", "Background",
				DEF_TREETABLE_BG_MONO,
				Tk_Offset(TreeTable, normalBorder),
				TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
				(char *) NULL, 0, 0},
       
    {TK_CONFIG_COLOR, "-bitmapforeground", "bitmapForeground",
				"Foreground",
				DEF_TREETABLE_BITMAPFG,
				Tk_Offset(TreeTable, defBitmapFgColor),
				0},
    {TK_CONFIG_COLOR, "-bitmapbackground", "bitmapBackground",
				"Foreground",
				DEF_TREETABLE_BITMAPBG,
				Tk_Offset(TreeTable, defBitmapBgColor),
				0},
    {TK_CONFIG_COLOR, "-bitmapselectfg", "bitmapSelectFg", "Foreground",
				DEF_TREETABLE_BITMAPSELECTFOREGROUND, 
       				Tk_Offset(TreeTable, defBitmapSelectFgColor),
				TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-bitmapselectbg", "bitmapSelectBg", "Background",
				DEF_TREETABLE_BITMAPSELECTBACKGROUND,
				Tk_Offset(TreeTable, defBitmapSelectBgColor),
				TK_CONFIG_NULL_OK},
    {TK_CONFIG_PIXELS, "-bitmapspace", "bitmapSpace", "BitmapSpace",
				DEF_TREETABLE_BITMAPSPACE,
				Tk_Offset(TreeTable, bitmapSpace),
				0},

    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
				DEF_TREETABLE_BORDER_WIDTH,
				Tk_Offset(TreeTable, borderWidth), 
				0},
    {TK_CONFIG_SYNONYM, "-bw", "borderWidth", (char *) NULL,
				(char *) NULL, 0, 0},
       
    {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
				DEF_TREETABLE_CURSOR,
				Tk_Offset(TreeTable, cursor),
				TK_CONFIG_NULL_OK},
       
    {TK_CONFIG_BOOLEAN, "-exportselection", "exportSelection",
				"ExportSelection",
				DEF_TREETABLE_EXPORT_SELECTION,
				Tk_Offset(TreeTable, exportSelection),
				0},
       
    {TK_CONFIG_SYNONYM, "-fg", "foreground", (char *) NULL,
				(char *) NULL, 0, 0},
    {TK_CONFIG_BOOLEAN, "-fillselection", "fillSelection",
				"FillSelection",
				DEF_TREETABLE_FILL_SELECTION,
				Tk_Offset(TreeTable, fillSelection),
				0},
				
    {TK_CONFIG_BOOLEAN, "-sortedinsertion", "sortedInsertion",
				"SortedInsertion",
				DEF_TREETABLE_SORTED_INSERTION,
				Tk_Offset(TreeTable, sortedInsertion),
				0},
    {TK_CONFIG_BOOLEAN, "-sortnocase", "sortNoCase", "SortNoCase",
				DEF_TREETABLE_SORT_NOCASE,
				Tk_Offset(TreeTable, sortNoCase),
				0},
    {TK_CONFIG_SYNONYM, "-nocase", "sortNoCase", (char *) NULL,
				(char *) NULL, 0, 0},
    {TK_CONFIG_BOOLEAN, "-sortcolumn", "sortColumn", "SortColumn",
				DEF_TREETABLE_SORT_NOCASE,
				Tk_Offset(TreeTable, sortColumn),
				0},
    {TK_CONFIG_SYNONYM, "-col", "sortColumn", (char *) NULL,
				(char *) NULL, 0, 0},
				
    {TK_CONFIG_FONT, "-font", "font", "Font",
				DEF_TREETABLE_FONT,
				Tk_Offset(TreeTable, defFontPtr),
				0},
    {TK_CONFIG_STRING, "-geometry", "geometry", "Geometry",
				DEF_TREETABLE_GEOMETRY,
				Tk_Offset(TreeTable, geometry),
				0},
    {TK_CONFIG_INT, "-width", "width", "Width",
				DEF_TREETABLE_WIDTH,
				Tk_Offset(TreeTable, width), 0},
    {TK_CONFIG_INT, "-height", "height", "Height",
				DEF_TREETABLE_WIDTH,
				Tk_Offset(TreeTable, height), 0},
       
    /*
     * hidden bitmap
     */
    {TK_CONFIG_BITMAP, "-hiddenbitmap", "hiddenbitmap", "HiddenBitmap",
				DEF_TABLEITEM_BITMAP,
				Tk_Offset(TreeTable, hiddenBitmap), 
				TK_CONFIG_NULL_OK},
       
    /* hidden image */
    {TK_CONFIG_STRING, "-hiddenimage", (char *) NULL, (char *) NULL,
				DEF_TABLEITEM_IMAGE,
				Tk_Offset(TreeTable, hiddenImageString),
				TK_CONFIG_NULL_OK},

    {TK_CONFIG_STRING, "-idlecommand", "idleCommand", "IdleCommand",
				DEF_TABLEITEM_IDLECOMMAND,
				Tk_Offset(TreeTable, idlecommand),
				TK_CONFIG_NULL_OK},
				
#if 0
    {TK_CONFIG_STRING, "-hiddencommand", "hiddencommand", "HiddenCommand",
				DEF_HIDDEN_COMMAND,
				Tk_Offset(TreeTable, hiddencommand),
				TK_CONFIG_NULL_OK},
#endif
       
    {TK_CONFIG_COLOR, "-highlightbackground", "highlightBackground",
				"HighlightBackground", NORMAL_BG,
				Tk_Offset(TreeTable, highlightBgColor),
				0},
    {TK_CONFIG_COLOR, "-highlightColor", "highlightcolor",
				"HighlightColor", BLACK, 
				Tk_Offset(TreeTable, highlightColor),
				0},
    {TK_CONFIG_PIXELS, "-highlightwidth", "highlightWidth", "HighlightWidth",
				DEF_TREETABLE_HIGHLIGHT_WIDTH,
				Tk_Offset(TreeTable, highlightWidth),
				0},
    {TK_CONFIG_PIXELS, "-highlightthickness", "highlightThickness", "HighlightThickness",
				DEF_TREETABLE_HIGHLIGHT_WIDTH,
				Tk_Offset(TreeTable, highlightWidth),
				0},
       
    {TK_CONFIG_PIXELS, "-indentwidth", "indentWidth", "BorderWidth",
				DEF_TREETABLE_INDENTWIDTH,
				Tk_Offset(TreeTable, indentWidth),
				0},
    {TK_CONFIG_COLOR, "-lineforeground", "lineForeground",
				"Foreground",
				DEF_TREETABLE_LINEFG,
				Tk_Offset(TreeTable, defLineFgColor),
				0},
    {TK_CONFIG_PIXELS, "-linewidth", "lineWidth", "LineWidth",
				NO_TREETABLE_LINEWIDTH,
				Tk_Offset(TreeTable, defLineWidth),
				0},
       
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
				DEF_TREETABLE_RELIEF,
				Tk_Offset(TreeTable, relief),
				0},
	/* Color for multi colors */
    {TK_CONFIG_BORDER, "-selectbackground", "selectBackground",
				"Foreground",
				DEF_TREETABLE_SELECT_COLOR, 
				Tk_Offset(TreeTable, selBorder),
				TK_CONFIG_COLOR_ONLY},
	/* color for black/white option */
    {TK_CONFIG_BORDER, "-selectbackground", "selectBackground",
				"Foreground", DEF_TREETABLE_SELECT_MONO,
				Tk_Offset(TreeTable, selBorder),
				TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_PIXELS, "-selectborderwidth", "selectBorderWidth",
				"BorderWidth",
				DEF_TREETABLE_SELECT_BD, 
				Tk_Offset(TreeTable, selBorderWidth),
				0},
				
	/* Color for multi colors */
    {TK_CONFIG_COLOR, "-selectforeground", "selectForeground",
				"Background",
				DEF_TREETABLE_SELECT_FG_COLOR,
				Tk_Offset(TreeTable, selFgColorPtr),
				TK_CONFIG_COLOR_ONLY},
	/* color for black/white option */
    {TK_CONFIG_COLOR, "-selectforeground", "selectForeground",
				"Background",
				DEF_TREETABLE_SELECT_FG_MONO,
				Tk_Offset(TreeTable, selFgColorPtr),
				TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_BOOLEAN, "-setgrid", "setGrid", "SetGrid",
				DEF_TREETABLE_SET_GRID,
				Tk_Offset(TreeTable, setGrid),
				0},
    {TK_CONFIG_STRING, "-selectmode", "selectMode", "SelectMode",
				DEF_TREETABLE_SELECTMODE,
				Tk_Offset(TreeTable, selectMode), 
				TK_CONFIG_NULL_OK},
    
    {TK_CONFIG_STRING, "-tabs", "tabs", "Tabs",
				DEF_TREETABLE_TABS,
				Tk_Offset(TreeTable, tabsList), 
				TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-deftabs", "defTabs", "DefTabs",
				DEF_TREETABLE_DEFTABS,
				Tk_Offset(TreeTable, defTabsList), 
				TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-justify", "justify", "Justify",
				DEF_TREETABLE_JUSTIFY,
				Tk_Offset(TreeTable, justify), 
				TK_CONFIG_NULL_OK},
    {TK_CONFIG_BOOLEAN, "-bestfit", "bestFit", "BestFit",
				DEF_TREETABLE_BESTFIT,
				Tk_Offset(TreeTable, BestFit), 
				0},
    {TK_CONFIG_BOOLEAN, "-autofit", "autoFit", "AutoFit",
				DEF_TREETABLE_AUTOFIT,
				Tk_Offset(TreeTable, AutoFit), 
				0},
    {TK_CONFIG_BOOLEAN, "-truncate", "truncate", "Truncate",
				DEF_TREETABLE_SPLITLINES,
				Tk_Offset(TreeTable, Truncate), 
				0},
    {TK_CONFIG_COLOR, "-splitlineforeground", "splitLineForeground",
				"Foreground",
				DEF_TREETABLE_SPLITLINEFG,
				Tk_Offset(TreeTable, defSplitLineFgColor),
				0},
    {TK_CONFIG_BOOLEAN, "-splitlines", "splitLines", "SplitLines",
				DEF_TREETABLE_TRUNCATE,
				Tk_Offset(TreeTable, SplitLines), 
				0},
    {TK_CONFIG_STRING, "-takefocus", "takeFocus", "TakeFocus",
				DEF_TREETABLE_TAKE_FOCUS,
				Tk_Offset(TreeTable, takeFocus),
				TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-textforeground", "textForeground",
				"Foreground",
				DEF_TREETABLE_TEXTFG, 
				Tk_Offset(TreeTable, defTextFgColor),
				0},
    {TK_CONFIG_STRING, "-xscrollcommand", "xScrollCommand", "ScrollCommand",
				DEF_TREETABLE_SCROLL_COMMAND,
				Tk_Offset(TreeTable, xScrollCmd),
				TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-yscrollcommand", "yScrollCommand", "ScrollCommand",
				DEF_TREETABLE_SCROLL_COMMAND,
				Tk_Offset(TreeTable, yScrollCmd),
				TK_CONFIG_NULL_OK},
       
    {TK_CONFIG_PIXELS, "-tabfreespace",
    				"tabFreeSpace", "TabFreeSpace",
       DEF_TABLEITEM_TABFREESPACE, Tk_Offset(TreeTable, tabsMinSpace), 0},
    
#ifdef USE_PATHFINDER
    {TK_CONFIG_BOOLEAN, "-accelerator", "accelerator", "Accelerator",
				DEF_TREETABLE_USE_ACCELERATOR,
				Tk_Offset(TreeTable, UseAccelerator), 
				0},
#endif

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}
};

static void
TreeTableDeletedProc(clientData)
    ClientData clientData;      /* Pointer to widget record for widget. */
{
    register TreeTable *tablePtr = (TreeTable *) clientData;
    Tk_Window tkwin = tablePtr->tkwin;

    /*
     * This procedure could be invoked either because the window was
     * destroyed and the command was then deleted (in which case tkwin
     * is NULL) or because the command was deleted, and then this procedure
     * destroys the widget.
     */

    if (tkwin != NULL) {
        if (tablePtr->setGrid) {
            Tk_UnsetGrid(tablePtr->tkwin);
        }
        tablePtr->tkwin = NULL;
        Tk_DestroyWindow(tkwin);
    }
}

/*
 *-------------------------------------------------------------------
 *
 * TreeTableCmd --
 *
 *      This procedure is invoked to process the "treetable" Tcl
 *      command.  See the user documentation for details on what
 *      it does.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side Effects:
 *      See the user documentation
 *
 *-------------------------------------------------------------------
TreeTableCmd
 */

static int
TreeTableCmd(clientData, interp, argc, argv)
     ClientData clientData;       /* Main window associated with interpreter */
     Tcl_Interp *interp;          /* Current interpreter */
     int        argc;             /* number of arguments */
     char       **argv;           /* argument strings */
{
    register TreeTable *tablePtr;
    Tk_Window new;
    Tk_Window tkwin = (Tk_Window) clientData;
    
    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }
    
    new = Tk_CreateWindowFromPath(interp, tkwin, argv[1], (char *) NULL);
    if (new == (Tk_Window) NULL) {
	return TCL_ERROR;
    }
    
    /*
     * Initialize the fields of the structure that won't be initialized
     * by ConfigureTreeTable, or that ConfigureTreeTable requires to be
     * initialized already (e.g. resource pointers).
     */
    tablePtr = (TreeTable *) calloc(sizeof(TreeTable), 1);
    if (tablePtr==NULL)
    {
	Tcl_SetResult (interp, "can't allocate memory for treetable\n", TCL_STATIC);
    	return TCL_ERROR;
    }
    tablePtr->tkwin = new;
    tablePtr->display = Tk_Display(new);
    tablePtr->interp = interp;
    
    tablePtr->relief = TK_RELIEF_RAISED;
    tablePtr->defLineWidth = -1;
    tablePtr->maxWidth = 10;
    tablePtr->selectFirst = -1;
    tablePtr->active=-1;

    FREE_CACHE();
    FREE_PARENT_CACHE();
    
#if 0
    Tk_CreateEventHandler(tablePtr->tkwin, 
			  ButtonPressMask,
			  TreeTablePressProc, (ClientData) tablePtr);
    Tk_CreateEventHandler(tablePtr->tkwin, 
			  ButtonReleaseMask,
			  TreeTableReleaseProc, (ClientData) tablePtr);
#endif
			  
    Tk_SetClass(tablePtr->tkwin, "TreeTable");
    Tk_CreateEventHandler(tablePtr->tkwin, 
			  ExposureMask|StructureNotifyMask|FocusChangeMask,
			  TreeTableEventProc, (ClientData) tablePtr);
    Tk_CreateSelHandler(tablePtr->tkwin, XA_PRIMARY, XA_STRING, 
			TreeTableFetchSelection, (ClientData) tablePtr, 
			XA_STRING);
    tablePtr->widgetCmd = Tcl_CreateCommand(interp, Tk_PathName(tablePtr->tkwin),
		TreeTableWidgetCmd,
		(ClientData) tablePtr, TreeTableDeletedProc);
    if (ConfigureTreeTable(tablePtr, argc-2, argv+2, 0) != TCL_OK) {
	Tk_DestroyWindow(tablePtr->tkwin);
	return TCL_ERROR;
    }
    interp->result = Tk_PathName(tablePtr->tkwin);
    return TCL_OK;
}

/* This is how the treetable widget is created.  This returns a
   standard Tcl result.  */

int
create_treetable_command (Tcl_Interp *interp)
{
  if (Tcl_CreateCommand (interp, "treetable", TreeTableCmd,
			 (ClientData) Tk_MainWindow (interp),
			 NULL) == NULL)
    return TCL_ERROR;
  return TCL_OK;
}

static int
TreeTableBuildSelection (Tcl_Interp *interp, TableItem*itemPtr, int pos)
{
    TableItem* Ptr;
    int i = pos;
    char index[20];
    for (Ptr=itemPtr; Ptr != NULL; Ptr=Ptr->nextPtr)
    {
	if (Ptr->flags&ITEM_SELECTED)
	{
	    sprintf (index, "%d", i);
	    Tcl_AppendElement(interp, index);
	}
	i++;
	if (Ptr->succPtr != NULL)
	{
	    i = TreeTableBuildSelection (interp, Ptr->succPtr, i);
	}
    }
    return i;
}

/*
 *--------------------------------------------------------------
 *
 * TreeTableWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */
static int
TreeTableWidgetCmd(clientData, interp, argc, argv)
     ClientData clientData;             /* Information about the treetable
					 * widget */
     Tcl_Interp *interp;                /* Current interpreter */
     int argc;                          /* Number of arguments. */
     char **argv;                       /* Argument strings */
{
    register TreeTable *tablePtr = (TreeTable *) clientData;
    register TableItem *itemPtr, *Ptr;
    register char c;
    register int result = TCL_OK;
    register int length;
    int index;
    char tmp [256];

#if 0
if (argv[1][0] != 'i' || strcmp (argv[1], "insert") != 0)
{
    fprintf (stderr, "%i ", clock()); ViewArgs (NULL, argc, argv, 2);
}
#endif

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_Preserve((ClientData) tablePtr);

    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'a') && (strncmp(argv[1], "activate", length) == 0))
    {
	if (argc != 3)
	{
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " activate index\"",
			 (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) != TCL_OK)
	{
	    goto error;
	}
	/* an item in a hidden sub tree can't be activeted */
	Ptr = TreeTableFindNotHiddenItem (tablePtr, index);
	if (Ptr == NULL)
	{
	    Ptr = TreeTableFindItem (tablePtr, index);
	    if (Ptr) for (Ptr=Ptr->parentPtr; Ptr; Ptr=Ptr->parentPtr)
	    {
		if (Ptr->flags & ITEM_HIDDEN_SUBTREE)
		{
		    index = TreeTableFindIndex (tablePtr, Ptr);
		    if (Ptr->nextPtr != NULL)
			index ++;
		    break;
		}
	    }
	}
	tablePtr->active = index;
	TreeTableRedrawRange(tablePtr);
    }
    else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0) && (length >= 2))
    {
    	char *p;
	int i;
	int tw;
	
	if (argc != 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " cget option\"",
			(char *) NULL);
	        goto error;
	}
	
	/* the tab stops are changed
	 * immediatly and we must reset the tab stop
	 * string before returning the values
	 */
	if (tablePtr->tabsList != NULL)
	{
	        Tcl_Free ((char *) tablePtr->tabsList);
		tablePtr->tabsList = Tcl_Alloc (sizeof (char) * tablePtr->tabsNum * 32);
		p = tablePtr->tabsList;
		p[0] = 0;
		for (i=0; i<tablePtr->tabsNum; i++)
		{
		    if (tablePtr->tabsHidden[i])
		    {
			if (i==0 && tablePtr->itemPtr!=NULL)
			{
			    tw = Max (0, ITEM_TEXT_X(tablePtr->itemPtr));
			}
			else
			{
			    tw = tablePtr->tabsMinSpace;
			}
		    }
		    else
		    {
			tw = tablePtr->tabs[i];
		    }
		    sprintf (p+strlen(p), *p ? " %i" : "%i", tw);
		}
	}
	
	result = Tk_ConfigureValue(interp, tablePtr->tkwin, configSpecs,
		(char *) tablePtr, argv[2], 0);
    } else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)
	&& (length > 2)) {
	
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, tablePtr->tkwin, configSpecs,
				      (char *) tablePtr, (char *) NULL, 0);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, tablePtr->tkwin, configSpecs,
				      (char *) tablePtr, argv[2], 0);
	} else {
	    result = ConfigureTreeTable(tablePtr, argc-2, argv+2,
					TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "curselection", length) == 0)
	       && length >= 2) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " curselection\"", (char *) NULL);
	    goto error;
	}

	/* build and return selection list */
	TreeTableBuildSelection (interp, tablePtr->itemPtr, 0);
	
    } else if ((c == 'r') && (strncmp(argv[1], "remove", length) == 0)) {
	int i, lineWidth=0;
	int deleteChildren = 0, cnt = 0;
	TableItem*Ptr;
	
	if (argc < 3 || argc > 4) {
	    RemoveError:
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " remove ?children? index\"",
			     (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[argc-1], 0, &i) 
	    != TCL_OK) {
	    goto error;
	}
	
	/* verify if delete children only */
	if (argc == 4)
	{
	    if (argv[2][0] == 'c' && strncmp(argv[2], "children", strlen (argv[2])) == 0)
	    {
		deleteChildren = 1;
	    }
	    else
	    {
		goto RemoveError;
	    }
	}
	itemPtr = TreeTableFindItem(tablePtr, i);
	if (itemPtr == (TableItem *) NULL) {
	    Tcl_AppendResult(interp, "Error obtaining item of index : ",
			     argv[argc-1], (char *) NULL);
	    goto error;
	}
	if (!deleteChildren)
	{
	    TreeTableRemoveItem(tablePtr, itemPtr, i, &lineWidth, 1, 0);
	    cnt = 1;
	}
	/* delete all children, but don't delete the item */
	else
	{
	    int start = i+1, end = i;
	    for (Ptr=itemPtr->succPtr; Ptr; Ptr=Ptr->nextPtr)
	    {
		end += 1 + Ptr->succNum;
	    }
	    if (end >= start)
	    {
		TreeTableDeleteRange(tablePtr, start, end);
		cnt = end - start + 1;
	    }
	}
	
	sprintf (tmp, "%i", cnt);
	Tcl_ResetResult (interp);
	Tcl_AppendResult (interp, tmp, NULL);
    } else if ((c == 'd') && (strncmp(argv[1], "delete", length) == 0)) {
	int start,end;
	
	if (argc < 3 || argc > 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " delete start ?end?\"",
			     (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &start) 
	    != TCL_OK) {
	    Tcl_AppendResult(interp, "Bad starting index.", (char *) NULL);
	    goto error;
	}
	if (argc > 3) {
	    if (GetTreeTableIndex(tablePtr, argv[3], 0, &end)
		!= TCL_OK) {
		Tcl_AppendResult(interp, "Bad ending index.", (char *) NULL);
		goto error;
	    }
	} else {
	    end = start;
	}
	if (end < start) {
	    end = start;
	}
	TreeTableDeleteRange(tablePtr, start, end);
    }
    else if ((c == 'g') && (strncmp(argv[1], "get", length) == 0))
    {
	int to;
	
	if (argc < 3 || argc > 4)
	{
	    get_invalid_args:
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " get index ?to?\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) != TCL_OK)
	{
	    goto get_invalid_args;
	}
	to = index;
	if (argc == 4 && GetTreeTableIndex(tablePtr, argv[3], 0, &to) != TCL_OK)
	{
	    goto get_invalid_args;
	}
	if (index < 0)
	{
	    index = 0;
	}
	if (index >= tablePtr->numItems)
	{
	    index = tablePtr->numItems-1;
	}
	/* find item */
	if (argc == 3)
	{
	    itemPtr = TreeTableFindItem (tablePtr, index);
	    if (itemPtr != NULL)
	    {
		interp->result = itemPtr->text; /*CHG*/
	    }
	}
	/* find a couple of items */
	else
	{
	    TreeTableGetItems (tablePtr, index, to);
	}
    }
    else if ((c == 'i') && (strncmp(argv[1], "insert", length) == 0))
    {
    	TableItem* newPtr;
	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " insert index ?list <list>? ?option1 option2 ...?\"",
			     (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 1, &index)
	    != TCL_OK) {
	    goto error;
	}
	
	/* add a copple of items with same options */
	if (argv[3][0] == 'l' && strcmp (argv[3], "list") == 0)
	{
	    if (TreeTableInsertItems (tablePtr, index,
				      argv[4], argc-5, argv+5) != TCL_OK)
	    	goto error;
	    goto done;
	}
	InsertedNewPosition = index;
	newPtr = TreeTableInsertItem(tablePtr, index, NULL, argc-3, argv+3);
	if (newPtr == NULL)
	{
	    Tcl_SetResult (interp, "can't add item", TCL_STATIC);
	    goto error;
	}
	sprintf (tmp, "%i", InsertedNewPosition);
	Tcl_ResetResult (interp);
	Tcl_AppendResult (interp, tmp, (char*) NULL);
    }
    else if ((c == 'd') && (strncmp(argv[1], "disp", length) == 0))
    {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " disp\"",
			     (char *) NULL);
	    goto error;
	}
	DisplayTreeTable (tablePtr);
    } else if ((c == 'i') && (strncmp(argv[1], "itemconfigure", length) == 0)) {
	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " itemconfigure index ?option val ...?",
			     "\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index)
	    != TCL_OK) {
	    goto error;
	}
	itemPtr = (TableItem *) TreeTableFindItem(tablePtr, index);
	if (itemPtr == (TableItem *) NULL) {
	    char s[20];

	    sprintf(s, "%d", index);
	    Tcl_AppendResult(interp, "Unable to locate item with index: \"",
			     s, "\"", (char *) NULL);
	    goto error;
	}
	
	/* set realy parent */
	itemPtr->parent = TreeTableFindIndex (tablePtr, itemPtr->parentPtr);
	
	if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, tablePtr->tkwin, itemConfigSpecs,
				      (char *) itemPtr, (char *) NULL, 0);
	} else if  (argc == 4) {
	    result = Tk_ConfigureInfo(interp, tablePtr->tkwin, itemConfigSpecs,
				      (char *) itemPtr, argv[3], 0);
	} else {
	    result = ConfigureTreeTableItem(tablePtr, itemPtr, argc-3,
					    argv+3, TK_CONFIG_ARGV_ONLY, 0);
	}
    }
    else if ((c == 'i') && (strncmp(argv[1], "itemcget", length) == 0))
    {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " itemcget index ?option?",
			     "\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index)
	    != TCL_OK) {
	    goto error;
	}
	itemPtr = (TableItem *) TreeTableFindItem(tablePtr, index);
	if (itemPtr == (TableItem *) NULL) {
	    char s[20];

	    sprintf(s, "%d", index);
	    Tcl_AppendResult(interp, "Unable to locate item with index: \"",
			     s, "\"", (char *) NULL);
	    goto error;
	}
	
	/* set realy parent */
	itemPtr->parent = TreeTableFindIndex (tablePtr, itemPtr->parentPtr);
	result = Tk_ConfigureValue (interp, tablePtr->tkwin, itemConfigSpecs,
				      (char *) itemPtr, argv[3], 0);
    }
    else if ((c == 'n') && (strncmp(argv[1], "nearest", length) == 0)) {
	int y;
	
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " nearest y\"", (char *) NULL);
	    goto error;
	}
	if (Tcl_GetInt(interp, argv[2], &y) != TCL_OK) {
	    goto error;
	}
	index = NearestTreeTableItem(tablePtr, y);
	sprintf(interp->result, "%d", index);
    } else if ((c == 'i') && (strncmp(argv[1], "index", length) == 0)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " index index\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex (tablePtr, argv[2], 0, &index) != TCL_OK) {
	    goto error;
	}
	sprintf(interp->result, "%d", index);
    } else if ((c == 'p') && (length >= 2) && (strncmp(argv[1], "parent", length) == 0)) {
	int parent;

	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " parent <line>\"",
			     (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) 
	    != TCL_OK) {
	    goto error;
	}
	itemPtr = (TableItem *) TreeTableFindItem(tablePtr, index);
	if (itemPtr == (TableItem *) NULL) {
	    Tcl_AppendResult(interp, "invalid treetable index : \"",
			     argv[2], "\"", (char *) NULL);
	    goto error;
	}
	if (itemPtr->parentPtr == (TableItem *) NULL) {
	    interp->result[0] = (char) NULL;
	} else {
	    parent = TreeTableFindIndex(tablePtr, itemPtr->parentPtr);
	    if (parent == -1) {
		Tcl_AppendResult(interp, "invalid treetable index : \"",
				 argv[2], "\"", (char *) NULL);
		goto error;
	    }
	    sprintf(interp->result, "%d", parent);
	}
    } else if ((c == 's') && (length >= 2) 
	       && (strncmp(argv[1], "scan", length) == 0)) {
	int x,y;
	
	if (argc != 5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " scan mark|dragto x y\"",
			     (char *) NULL);
	    goto error;
	}
	if ((Tcl_GetInt(interp, argv[3], &x) != TCL_OK) ||
	    (Tcl_GetInt(interp, argv[4], &y) != TCL_OK)) {
	    goto error;
	}
	if ((argv[2][0] == 'm') 
	    && (strncmp(argv[2], "mark", strlen(argv[2])) == 0)) {
	    tablePtr->scanMarkX = x;
	    tablePtr->scanMarkY = y;
	    tablePtr->scanMarkXOffset = tablePtr->xOffset;
	    tablePtr->scanMarkYIndex = tablePtr->topIndex;
	} else if ((argv[2][0] == 'd')
		   && (strncmp(argv[2], "dragto", strlen(argv[2])) == 0)) {
	    TreeTableScanTo(tablePtr, x, y);
	} else {
	    Tcl_AppendResult(interp, "bad scan option \"", argv[2],
			     "\": must be mark or dragto", (char *) NULL);
	    goto error;
	}
    }
    else if ((c == 's') && (length >= 2) && (strncmp(argv[1], "selection", length) == 0))
    {
	if (argc < 3)
	{
	    goto selection_invalid_args;
	}
	
	length = strlen(argv[2]);
	c = argv[2][0];
	
	/* Only require information */
	if (argc == 4 && (c == 'i') && (strncmp(argv[2], "includes", length) == 0))
	{
	    if (GetTreeTableIndex(tablePtr, argv[3], 0, &index) != TCL_OK)
	    {
		goto selection_invalid_args;
	    }
	    itemPtr = TreeTableFindItem (tablePtr, index);
	    if (itemPtr != NULL && (itemPtr->flags&ITEM_SELECTED))
	    {
	    	sprintf (interp->result, "1");
	    }
	    else
	    {
	    	sprintf (interp->result, "0");
	    }
	    
	    goto done;
	}
	
	else if (argc >= 4 && c == 'c' && argv[2] != NULL && (strncmp(argv[2], "clear", length) == 0))
	{
	    int last;
	    if (argc > 5)
	    {
		goto selection_invalid_args;
	    }
	    if (GetTreeTableIndex (tablePtr, argv[3], 0, &index) != TCL_OK)
	    {
		goto error;
	    }
	    last = index;
	    if (argc == 5 && GetTreeTableIndex (tablePtr, argv[4], 0, &last) != TCL_OK)
	    {
		goto error;
	    }
	    if (TreeTableLostSel (tablePtr, Min (index, last), Max (index, last)))
	    {
		TreeTableRedrawRange(tablePtr);
	    }
	    goto done;
	}
	if (argc > 5)
	{
	    goto selection_invalid_args;
	}

	if ((c == 'a') && (strncmp(argv[2], "anchor", length) == 0))
	{
	    if (argc != 4 || GetTreeTableIndex(tablePtr, argv[3], 0, &index) 
		!= TCL_OK)
	    {
		goto selection_invalid_args;
	    }
	    tablePtr->selectAnchor = index;
	}
	else if ((c == 's') && (strncmp(argv[2], "set", length) == 0))
	{
	    int to;
	    if (GetTreeTableIndex(tablePtr, argv[3], 0, &index)
		!= TCL_OK)
	    {
	        goto selection_invalid_args;
	    } 
	    to = index;
	    if (argc == 5 &&
	        GetTreeTableIndex(tablePtr, argv[4], 0, &to) != TCL_OK)
	    {
		goto selection_invalid_args;
	    }
	    TreeTableSelectFromTo
		    (tablePtr, tablePtr->itemPtr, 0, Min (index, to), Max (index, to));
	    TreeTableRedrawRange(tablePtr);
	}
	else
	{
	    selection_invalid_args:
	    Tcl_AppendResult(interp,
	    	"bad select option or parameter: must be \"",
		argv[0],
		"\" anchor index, clear index ?last?, includes index, set index ?last?",
		(char *) NULL);
	    goto error;
	}
    }
    else if ((c == 's') && (length > 2)
	       && (strncmp(argv[1], "size", length) == 0))
    {
	sprintf(interp->result, "%d", tablePtr->numItems);
    }
    else if ((c == 'x') && (strncmp(argv[1], "xview", length) == 0))
    {
	if (argc == 2)
	{
	    if (tablePtr->maxWidth == 0) {
		interp->result = "0 1";
	    } else {
		double fraction,fraction2;
		double windowWidth;

		windowWidth = (double) TREETABLE_WIDTH(tablePtr);

		fraction = tablePtr->xOffset/((double) tablePtr->maxWidth);
		fraction2 = (tablePtr->xOffset + windowWidth)
			/((double) tablePtr->maxWidth);
			
		if (fraction2 > 1.0) {
			fraction2 = 1.0;
		}
		sprintf(interp->result, "%g %g", fraction, fraction2);
	    }
	    goto done;
	}
	else if (argc == 3)
	{
	    if (Tcl_GetInt (interp, argv[2], &index) != TCL_OK)
	    {
		goto xview_invalid_num;
	    }
	    ChangeTreeTableOffset(tablePtr, index*tablePtr->xScrollUnit);
	    goto done;
	}
	else if (argc == 4 && argv[2][0] == 's' && strcmp (argv[2], "see") == 0)
	{
	    if (Tcl_GetInt (interp, argv[3], &index) != TCL_OK)
	    {
		goto xview_invalid_num;
	    }
	    index = index * tablePtr->xScrollUnit - Tk_Width(tablePtr->tkwin) / 2;
	    ChangeTreeTableOffset(tablePtr, index);
	    goto done;
	}
	else
	{
	    int number, units = 1;
	    double pos;
	    switch (Tk_GetScrollInfo(interp, argc, argv, &pos, &number))
	    {
	    case TK_SCROLL_MOVETO:
	    	pos = pos * (double) tablePtr->maxWidth + 0.5;
		index = (int) pos;
		break;
	    case TK_SCROLL_PAGES:
	    	units = Tk_Width(tablePtr->tkwin) / tablePtr->xScrollUnit;
	        index = tablePtr->xOffset + number * units * tablePtr->xScrollUnit;
		break;
	    case TK_SCROLL_UNITS:
		index = tablePtr->xOffset + number * units * tablePtr->xScrollUnit;
		break;
	    case TK_SCROLL_ERROR:
	    	goto error;
		break;
	    }
	    if (index < 0) index = 0;
	    if (index > tablePtr->maxWidth) index = tablePtr->maxWidth;
	    ChangeTreeTableOffset (tablePtr, index);
	    goto done;
	}
	xview_invalid_num:
	Tcl_AppendResult(interp, "xview: invalid number", (char *) NULL);
	goto error;
    }
    else if ((c == 'y') && (strncmp(argv[1], "yview", length) == 0))
    {
	if (argc == 2)
	{
	    if (tablePtr->numItems == 0)
	    {
		interp->result = "0 1";
	    }
	    else
	    {
		double count;
		double fraction, fraction2;

		count = (double)TreeTableCountNotHidden (tablePtr, 0);
		fraction = tablePtr->topIndex/count;
		fraction2 = (tablePtr->topIndex + tablePtr->numLines) /
			    ((double) count);
		if (fraction2 > 1.0)
		{
		    fraction2 = 1.0;
		}
		sprintf(interp->result, "%g %g", fraction, fraction2);
	    }
	    goto done;
	}
	else if (argc == 3)
	{
	    if (GetTreeTableIndex(tablePtr, argv[2], 0, &index)
		!= TCL_OK)
	    {
		goto yview_invalid_num;
	    }
	    ChangeTreeTableView(tablePtr, index, 0);
	    goto done;
	}
	else if (argc == 4 && argv[2][0] == 's' && strcmp (argv[2], "see") == 0)
	{
	    if (GetTreeTableIndex(tablePtr, argv[3], 0, &index) != TCL_OK)
	    {
		goto yview_invalid_num;
	    }
	    if (index < 0)
	    {
		index = 0;
	    }
	    else if (index >= tablePtr->numItems)
	    {
		index = tablePtr->numItems-1;
	    }
	    tablePtr->active = index;
	    
	    /* convert realy index to index relative to viewed items */
	    index = TreeTableViewedIndex (tablePtr, index);
	    
	    if (index < 0)
	    {
		index = 0;
	    }
	    else if (index >= TreeTableCountNotHidden (tablePtr, 0))
	    {
		index = TreeTableCountNotHidden (tablePtr, 0) - 1;
	    }
	    index -= tablePtr->numLines/2;
	    ChangeTreeTableView(tablePtr, index, 1);
	    goto done;
	}
	else
	{
	    int number;
	    double pos;
	    switch (Tk_GetScrollInfo(interp, argc, argv, &pos, &number))
	    {
	    case TK_SCROLL_MOVETO:
	    	pos = pos * tablePtr->numViewed + 0.5;
		index = (int) pos;
		break;
	    case TK_SCROLL_PAGES:
		index = tablePtr->topIndex + number * tablePtr->numLines;
		break;
	    case TK_SCROLL_UNITS:
	    	index = tablePtr->topIndex + number;
		break;
	    case TK_SCROLL_ERROR:
	    	goto error;
		break;
	    }
	    ChangeTreeTableView (tablePtr, index, 1);
	    goto done;
	}
	    
	yview_invalid_num:
	Tcl_AppendResult(interp, "yview: invalid number", (char *) NULL);
	goto error;
    }
    else if (argc == 3 && (c == 's') && strncmp (argv[1], "see", length) == 0)
    {
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) != TCL_OK)
	{
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0],
		" see index\"",
		(char *) NULL);
	    goto error;
	}
	if (index < tablePtr->topIndex)
	{
	    ChangeTreeTableView(tablePtr, index, 0);
	}
	else if (index > tablePtr->topIndex+tablePtr->numLines-1)
	{
	    ChangeTreeTableView(tablePtr, index-tablePtr->numLines+1, 0);
	}
    }
    else if ((c == 'i') && (strncmp(argv[1], "identify", length) == 0))
    {
	int x, y, ret;
	char*p;
	
	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0],
			     " identify x y\"",
			     (char *) NULL);
	    goto error;
	} 
	if (Tcl_GetInt (interp, argv[2], &x) != TCL_OK) {
	    goto error;
	}
	if (Tcl_GetInt (interp, argv[3], &y) != TCL_OK) {
	    goto error;
	}
	
	ret = TreeTableToggle (tablePtr, x, y, 1);
	switch (ret)
	{
	case ITEM_VIEWED:
		p = "view";
		break;
	case ITEM_HIDDEN:
		p = "hide";
		break;
	case ITEM_TEXT:
		p = "text";
		break;
	case ITEM_UNKNOWN:
		p = "noroot";
		break;
	case ITEM_NOT_FOUND:
		p = "";
		break;
	default:
		p = "";
		break;
	}
	strcpy (interp->result, p);
    }
    else if ((c == 't') && (strncmp(argv[1], "toggle", length) == 0))
    {
	int hide = 0, real_index=0;
	
	if (argc < 3 || argc > 5) {
	    ToggleError:
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0],
			     " toggle index ?hide|view? ?virtual|real?\"",
			     (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) 
	    != TCL_OK) {
	    goto error;
	}
	if (argc == 5) /* verify if we want to toggle a virtual/real position*/
	{
	    if (argv[4][0] == 'v')
		real_index = 0;
	    else if (argv[4][0] == 'r')
	    	real_index = 1;
	    else
	    	goto ToggleError;
	}
	if (real_index)
	{
	    itemPtr = TreeTableFindItem (tablePtr, index);
	}
	else
	{
	    itemPtr = TreeTableFindNotHiddenItem (tablePtr, index);
	}
	if (itemPtr == NULL)
	{
	    goto error;
	}
	if (argc >= 4)
	{
	    if (argv[3][0] == 'v')
		hide = 0;
	    else if (argv[3][0] == 'h')
	    	hide = 1;
	    else
	    	goto ToggleError;
	}
	else 
	{
	    hide = (itemPtr->flags&ITEM_HIDDEN_SUBTREE) ? 0 : 1;
	}
	TreeTableToggleItem (tablePtr, itemPtr, hide);
    }
    else if ((c == 'f') && (strncmp(argv[1], "fullname", length) == 0))
    {
	int len, alen;
	char *cchar = " ";
	
	if (argc < 3 || argc > 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " fullname index ?join string?\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) 
	    != TCL_OK) {
	    goto error;
	}
	if (argc == 4)
	{
	    cchar = argv[3];
	}
	if (index < 0) {
	    index = 0;
	}
	if (index >= tablePtr->numItems) {
	    index = tablePtr->numItems-1;
	}
	/* find item */
	itemPtr = TreeTableFindItem (tablePtr, index);
	if (itemPtr != NULL)
	{
	    char *buf;
	    /** calculate length of used buffer */
	    alen = strlen (cchar);
	    len = strlen(itemPtr->text) + alen + 1;
	    for (Ptr=itemPtr->parentPtr; Ptr != NULL; Ptr=Ptr->parentPtr)
	    {
	    	len += Ptr->textLength + alen;
	    }
	    buf = Tcl_Alloc (len);
	    if (buf == NULL)
	    {
		goto error;
	    }
	    /* build full path name and join it with join string */
	    strcpy (buf, itemPtr->text);
	    for (Ptr=itemPtr->parentPtr; Ptr != NULL; Ptr=Ptr->parentPtr)
	    {
		if (*buf)
		{
	    	    memmove (buf + alen, buf, strlen (buf) + 1);
		    strncpy (buf, cchar, alen);
		}
		memmove(buf + Ptr->textLength, buf, strlen (buf) + 1);
		strncpy (buf, Ptr->text, Ptr->textLength);
	    }
	    
	    Tcl_SetResult(interp, buf, TCL_DYNAMIC);
	}
    }
    else if (c == 's' && strncmp (argv[1], "search", length) == 0)
    {
	if (TreeTableSearchCmd(tablePtr, argc, argv) != TCL_OK)
	    goto error;
    }
    else if ((c == 's') && (length > 5)
	       && (strncmp(argv[1], "seenall", length) == 0))
    {
	sprintf(tmp, "%d", tablePtr->numViewed);
	Tcl_SetResult(interp, tmp, TCL_VOLATILE);
    }
    else if ((c == 'd') && (strncmp(argv[1], "data", length) == 0))
    {
	if (argc != 3)
	{
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " data index\"", (char *) NULL);
	    goto error;
	}
	if (Tcl_GetInt(interp, argv[2], &index) != TCL_OK) {
	    goto error;
	}
	itemPtr = TreeTableFindItem(tablePtr, index);
	if (itemPtr == NULL)
	{
	    Tcl_AppendResult(interp, "wrong item #", (char *) NULL);
	    goto error;
	}
	Tcl_SetResult (interp,
			(itemPtr->data) ? itemPtr->data : "",
			TCL_STATIC);
    }
    else if ((c == 's') && (strncmp(argv[1], "sort", length) == 0))
    {
	if (TreeTableSort (tablePtr, argc-2, argv+2) != TCL_OK)
	    goto error;
    }
    else if ((c == 'x') && (strncmp(argv[1], "xoffset", length) == 0))
    {
	sprintf(interp->result, "%d", tablePtr->xOffset);
    }
    else if ((c == 'b') && (strncmp(argv[1], "bbox", length) == 0))
    {
	int x, y, x2, y2;
	if (argc != 3)
	{
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " bbox index\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) != TCL_OK)
	{
	    goto error;
	}
	if (TreeTableGetBBox (tablePtr, index, &x, &y, &x2, &y2) != TCL_OK)
	{
	    goto error;
	}
	sprintf (tmp, "%i %i %i %i", x, y, x2, y2);
	Tcl_ResetResult (interp);
	Tcl_AppendResult (interp, tmp, NULL);
    }
    else if ((c == 'c') && (strncmp(argv[1], "column", length) == 0))
    {
	int num, hide;
	if (argc != 4)
	{
	    column_error:
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " column hide/view/toggle column#\"", (char *) NULL);
	    goto error;
	}
	if (Tcl_GetInt(interp, argv[3], &num) != TCL_OK)
	{
	    goto column_error;
	}
	if (tablePtr->tabs == NULL)
	{
	    Tcl_AppendResult(interp,
	    			"wrong usage of \"",
	    			argv[0],
				"\" there are no tab stops to handle with",
				(char*) NULL);
	    goto error;
	}
	if (num < 0)
	{
	    num = 0;
	}
	if (num > tablePtr->tabsNum)
	{
	    num = tablePtr->tabsNum;
	}
	if (strcmp (argv[2], "hide") == 0)
	{
	    hide = 1;
	}
	else if (strcmp (argv[2], "view") == 0)
	{
	    hide = 0;
	}
	else if (strcmp (argv[2], "toggle") == 0)
	{
	    if (tablePtr->tabsHidden[num])
	    {
		hide = 0;
	    }
	    else
	    {
		hide = 1;
	    }
	}
	else
	{
	    goto column_error;
	}
	tablePtr->tabsHidden[num] = hide;
	
	TreeTableRedrawRange (tablePtr);
    }
    /*
     * Calculate how many parents does an item have
     */
    else if ((c == 'l') && (strncmp(argv[1], "levels", length) == 0))
    {
    	int level = 0;
	if (argc != 3)
	{
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " level index\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) != TCL_OK)
	{
	    goto error;
	}
	
	itemPtr = TreeTableFindItem (tablePtr, index);
	if (itemPtr != NULL)
	{
	   while (itemPtr->parentPtr != NULL)
	   {
		level++;
		itemPtr = itemPtr->parentPtr;
	   }
	}
	
	sprintf (tmp, "%i", level);
	Tcl_ResetResult (interp);
	Tcl_AppendResult (interp, tmp, NULL);
    }
    else if ((c == 'c') && (strncmp(argv[1], "children", length) == 0))
    {
	if (argc != 3)
	{
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " children index\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) != TCL_OK)
	{
	    goto error;
	}
	
	itemPtr = TreeTableFindItem (tablePtr, index);
	if (itemPtr == NULL)
	{
	    goto error;
	}
	Tcl_ResetResult (interp);
	for (index += 1, Ptr=itemPtr->succPtr; Ptr; Ptr=Ptr->nextPtr)
	{
	    sprintf (tmp, "%i", index);
	    Tcl_AppendElement (interp, tmp);
	    index += 1 + Ptr->succNum;
	}
    }
    else if ((c == 'r') &&
    		(strncmp(argv[1], "root", length) == 0 ||
		 strncmp(argv[1], "rootname", length) == 0))
    {
	if (argc != 3)
	{
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " root index\"", (char *) NULL);
	    goto error;
	}
	if (GetTreeTableIndex(tablePtr, argv[2], 0, &index) != TCL_OK)
	{
	    goto error;
	}
	itemPtr = TreeTableFindItem (tablePtr, index);
	if (itemPtr == NULL)
	{
	    goto error;
	}
	/* get base root of item */
	for (Ptr=itemPtr; Ptr->parentPtr; Ptr=Ptr->parentPtr)
	{
	}
	if (strncmp(argv[1], "rootname", length) == 0)
	{
	    Tcl_SetResult (interp, Ptr->text, TCL_STATIC);
	}
	else
	{
	    sprintf (tmp, "%i", TreeTableFindIndex (tablePtr, Ptr));
	    Tcl_SetResult (interp, tmp, TCL_VOLATILE);
	}
    }
#if defined (TEST_TREE) && defined (USE_CACHE)
    else if ((c == 't') && (strncmp(argv[1], "test", length) == 0))
    {
	ViewCache (tablePtr);
    }
#endif
    else
    {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
			 "\": must be "
			 "activate, "
			 "bbox, "
			 "cget, "
			 "configure, "
			 "curselection, "
			 "delete, "
			 "fullname, "
			 "get, "
			 "hidecolumn, "
			 "identify, "
			 "index, "
			 "insert, "
			 "itemcget, "
			 "levels, "
			 "nearest, "
			 "remove, "
			 "root, "
			 "rootname, "
			 "scan, "
			 "search, "
			 "see, "
			 "seenall, "
			 "select, "
			 "selection, "
			 "size, "
			 "sort, "
			 "toggle, "
			 "xview, "
			 "or "
			 "yview", (char *) NULL);
	goto error;
    }
 
/*
if (tabsAccess)
{
fprintf (stderr, "\ttabs now <%s>\n", tablePtr->tabsList?tablePtr->tabsList:"");
}
*/

  done:
    Tcl_Release((ClientData) tablePtr);
    
#if 0
if (argv[1][0] != 'i' || strcmp (argv[1], "insert") != 0)
{
    fprintf (stderr, "%i\n", clock());
}
#endif

    return result;
    
  error:
    Tcl_Release((ClientData) tablePtr);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyTreeTable --
 *
 *	This procedure is invoked by Tcl_EventuallyFree or Tcl_Release
 *	to clean up the internal structure of a treetable at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the treetable is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyTreeTable(memPtr)
    char *memPtr;       /* Info about widget. */
{
    register TreeTable *tablePtr = (TreeTable *) memPtr;
    
    /*
     * Free up all of the list elements.
     */
    TreeTableDeleteRange (tablePtr, 0, tablePtr->numItems-1);

    /*
     * Free up all the stuff that requires special handling,
     * then let Tk_FreeOptions handle all the standard option-related
     * stuff.
     */

    if (tablePtr->defTextGC != None) {
	Tk_FreeGC(tablePtr->display, tablePtr->defTextGC);
    }
    if (tablePtr->defBitmapGC != None) {
	Tk_FreeGC(tablePtr->display, tablePtr->defBitmapGC);
    }
    if (tablePtr->defBitmapSelectGC != None) {
	Tk_FreeGC(tablePtr->display, tablePtr->defBitmapSelectGC);
    }
    if (tablePtr->defLineGC != None) {
	Tk_FreeGC(tablePtr->display, tablePtr->defLineGC);
    }
    if (tablePtr->selectGC != None) {
	Tk_FreeGC(tablePtr->display, tablePtr->selectGC);
    }
    /* this function doesn't delete the image icons */
    Tk_FreeOptions(configSpecs, (char *) tablePtr, tablePtr->display, 0);
    
    /* also free image if availiable */
    if (tablePtr->hiddenImage != NULL)
    {
	Tk_FreeImage (tablePtr->hiddenImage);
    }
    /* free tab stops, if availiable */
    if (tablePtr->tabs != NULL) {
	free ((char *)tablePtr->tabs);
    }
    if (tablePtr->defTabs != NULL) {
	free ((char *)tablePtr->defTabs);
    }
    if (tablePtr->tabsJustify != NULL) {
	free ((char *)tablePtr->tabsJustify);
    }
    if (tablePtr->idlecommand != NULL) {
	ckfree ((char *)tablePtr->idlecommand);
    }
    if (tablePtr->tabsHidden != NULL) {
	free ((char *)tablePtr->tabsHidden);
    }
    free((char *) tablePtr);
}


/*
 *----------------------------------------------------------------------
 *
 * ConfigureTreeTable --
 *
 *      This procedure is called to process an argv/argc list, plus
 *      the Tk option database, in order to configure (or reconfigure)
 *      a treetable widget.
 *
 * Results:
 *      The return value is a standard Tcl result.  If TCL_ERROR is
 *      returned, then interp->result contains an error message.
 *
 * Side effects:
 *      Configuration information, such as colors, border width,
 *      etc. get set for tablePtr;  old resources get freed,
 *      if there were any.
 *
 *----------------------------------------------------------------------
 */

static int
ConfigureTreeTable(tablePtr, argc, argv, flags)
     register TreeTable *tablePtr;   /* Information about widget; may
				      * or may not already have values 
				      * for some fields. */
     int argc;                       /* Number of valid entries in argv */
     char **argv;                    /* Arguments */
     int flags;                      /* Flags to pass to Tk_ConfigureWidget */
{
    XGCValues gcValues;
    GC new;
    int i, width, height, oldExport;
    Tk_FontMetrics metrics;
    int pixelWidth, pixelHeight;
    Tk_Font oldfont;
    Pixmap oldhiddenBitmap;
    char
    	*oldhiddenImageString,
	*oldTabsList,
	*oldJustify,
	*oldDefTabsList,
	*oldSelectModeStr;
    int oldBestFit;
    int compute_width = 0;
#ifdef USE_PATHFINDER
    int oldUseAccel = tablePtr->UseAccelerator;
#endif
    int oldWidth, oldHeight;

    oldExport		= tablePtr->exportSelection;
    oldfont		= tablePtr->defFontPtr;
    oldhiddenBitmap	= tablePtr->hiddenBitmap;
    oldhiddenImageString= tablePtr->hiddenImageString;
    oldTabsList		= tablePtr->tabsList;
    oldJustify          = tablePtr->justify;
    oldDefTabsList	= tablePtr->defTabsList;
    oldSelectModeStr    = tablePtr->selectModeStr;
    oldBestFit          = tablePtr->BestFit;
    oldHeight           = tablePtr->height;
    oldWidth            = tablePtr->width;

    if (Tk_ConfigureWidget(tablePtr->interp, tablePtr->tkwin, configSpecs, 
			   argc, argv, (char *) tablePtr, flags) != TCL_OK)
    {
	return TCL_ERROR;
    }

    /* get geometry for hidden bitmap/image **/
    if (tablePtr->hiddenImageString != NULL && oldhiddenImageString != tablePtr->hiddenImageString)
    {
    	Tk_Image image;
	image = Tk_GetImage (tablePtr->interp, tablePtr->tkwin,
				tablePtr->hiddenImageString,
				TreeTableImageProc, (ClientData) NULL);
	if (image == NULL)
	{
	    Tcl_AppendResult(tablePtr->interp, "bad image name \"", tablePtr->hiddenImageString,
			 "\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (tablePtr->hiddenImage != NULL)
	{
	    Tk_FreeImage (tablePtr->hiddenImage);
	}
	tablePtr->hiddenImage = image;
	Tk_SizeOfImage (tablePtr->hiddenImage,
			&tablePtr->hiddenWidth,
			&tablePtr->hiddenHeight);
	compute_width = 1;
    }
    else if (oldhiddenBitmap != tablePtr->hiddenBitmap)
    {
	Tk_SizeOfBitmap(tablePtr->display, tablePtr->hiddenBitmap,
			&tablePtr->hiddenWidth, &tablePtr->hiddenHeight);
	compute_width = 1;
    }

    /*
     * Some options need special processing, such as parsing the
     * geometry and setting the background from a 3-D border.
     */
    Tk_SetBackgroundFromBorder(tablePtr->tkwin, tablePtr->normalBorder);

    if (tablePtr->highlightWidth < 0)
    {
	tablePtr->highlightWidth = 0;
    }
    tablePtr->inset = tablePtr->highlightWidth + tablePtr->borderWidth;
    gcValues.foreground = tablePtr->defTextFgColor->pixel;
    gcValues.font = Tk_FontId (tablePtr->defFontPtr);
    gcValues.graphics_exposures = False;
    new = Tk_GetGCColor(tablePtr->tkwin,
		   GCForeground|GCFont|GCGraphicsExposures,
		   &gcValues, tablePtr->defTextFgColor, NULL);
    if (tablePtr->defTextGC != None)
    {
	Tk_FreeGC(tablePtr->display, tablePtr->defTextGC);
    }
    tablePtr->defTextGC = new;
    gcValues.foreground = tablePtr->selFgColorPtr->pixel;
    gcValues.background = Tk_3DBorderColor(tablePtr->selBorder)->pixel;
    gcValues.font = Tk_FontId (tablePtr->defFontPtr);
    
    new = Tk_GetGCColor(tablePtr->tkwin, GCForeground|GCBackground|GCFont,
		   &gcValues, tablePtr->selFgColorPtr,
		   Tk_3DBorderColor(tablePtr->selBorder));
    if (tablePtr->selectGC != None)
    {
	Tk_FreeGC(tablePtr->display, tablePtr->selectGC);
    }
    tablePtr->selectGC = new;
    
    gcValues.foreground = tablePtr->defBitmapFgColor->pixel;
    gcValues.background = tablePtr->defBitmapBgColor->pixel;
    gcValues.graphics_exposures = False;
    if (tablePtr->defBitmapGC != None)
    {
	Tk_FreeGC(tablePtr->display, tablePtr->defBitmapGC);
    }
    tablePtr->defBitmapGC = Tk_GetGCColor(tablePtr->tkwin, 
					  GCForeground|GCBackground|
					  GCGraphicsExposures, &gcValues,
					  tablePtr->defBitmapFgColor,
					  tablePtr->defBitmapBgColor);

    if ((tablePtr->defBitmapSelectFgColor != None) ||
	(tablePtr->defBitmapSelectBgColor != None))
    {
	XColor *foreground, *background;

	if (tablePtr->defBitmapSelectFgColor != None) {
	    foreground = tablePtr->defBitmapSelectFgColor;
	} else {
	    foreground = tablePtr->selFgColorPtr;
	}
	gcValues.foreground = foreground->pixel;
	if (tablePtr->defBitmapSelectBgColor != None) {
	    background = tablePtr->defBitmapSelectBgColor;
	} else {
	    background = Tk_3DBorderColor(tablePtr->selBorder);
	}
	gcValues.background = background->pixel;
	gcValues.graphics_exposures = False;
	if (tablePtr->defBitmapSelectGC != None) {
	    Tk_FreeGC(tablePtr->display, tablePtr->defBitmapSelectGC);
	}
	tablePtr->defBitmapSelectGC = Tk_GetGCColor(tablePtr->tkwin,
						    GCForeground|GCBackground|
						    GCGraphicsExposures,
						    &gcValues, foreground,
						    background);
    }
    if (tablePtr->defLineWidth == atoi(NO_TREETABLE_LINEWIDTH))
    {
	tablePtr->defLineWidth = atoi(DEF_TREETABLE_LINEWIDTH);
    }
    gcValues.foreground = tablePtr->defLineFgColor->pixel;
    gcValues.graphics_exposures = False;
    gcValues.line_width = tablePtr->defLineWidth;
    if (tablePtr->defLineGC != None)
    {
	Tk_FreeGC(tablePtr->display, tablePtr->defLineGC);
    }
    tablePtr->defLineGC = Tk_GetGCColor(tablePtr->tkwin,
					(GCForeground|GCGraphicsExposures
					 |GCLineWidth),
					&gcValues, tablePtr->defLineFgColor,
					NULL);
    
    /* Split line color */
    gcValues.foreground = tablePtr->defSplitLineFgColor->pixel;
    if (tablePtr->defSplitLineGC != None)
    {
	Tk_FreeGC(tablePtr->display, tablePtr->defSplitLineGC);
    }
    tablePtr->defSplitLineGC = Tk_GetGCColor(tablePtr->tkwin,
				   GCForeground|GCGraphicsExposures|GCLineWidth,
				   &gcValues, tablePtr->defSplitLineFgColor,
				   NULL);
   
    /*
     * Claim the selection if we've suddenly started exporting it.
     */
    if (tablePtr->exportSelection && (!oldExport) && (tablePtr->selectFirst != -1))
    {
	Tk_OwnSelection(tablePtr->tkwin, XA_PRIMARY, TreeTableLostSelection,
			(ClientData) tablePtr);
    }
    
    /*
     * Register the desired geometry for the window, and arrange for
     * the window to be redisplayed.
     */
    if ((sscanf(tablePtr->geometry, "%dx%d", &width, &height) != 2) ||
	(width <= 0) || (height <= 0))
    {
	Tcl_AppendResult(tablePtr->interp, "bad geometry \"", tablePtr->geometry,
			 "\", should be \"widthxheight\"", (char *) NULL);
	return TCL_ERROR;
    }
    
    /*
     * width or height is given
     */
    if (oldHeight != tablePtr->height)
    {
    	height = tablePtr->height;
    }
    if (oldWidth != tablePtr->width)
    {
    	width = tablePtr->width;
    }

    if (oldfont != tablePtr->defFontPtr)
    {
        Tk_GetFontMetrics (tablePtr->defFontPtr, &metrics);
	
	tablePtr->defFontHeight = metrics.linespace;
	TreeTableComputeLineHeight(tablePtr, NULL, 0);
	if (metrics.linespace + 1 + 2 * tablePtr->selBorderWidth
	    > tablePtr->lineHeight) {
	  tablePtr->lineHeight = (metrics.linespace + 1
				  + 2 * tablePtr->selBorderWidth);
	}
	tablePtr->flags |= UPDATE_H_SCROLLBAR;
	compute_width = 1;
    }
    
    /*
     * actual tab stop list
     */
    if (oldTabsList != tablePtr->tabsList && tablePtr->tabsList != NULL)
    {
	char **tabs, *p;
	int num, oldnum, *oldtabs;
	if (Tcl_SplitList (tablePtr->interp, tablePtr->tabsList, &num, &tabs) != TCL_OK)
	    return TCL_ERROR;
	if (num == 0)
	{
	    Tcl_Free ((char *) tabs);
	    return TCL_ERROR;
	}
	oldtabs = tablePtr->tabs;
	oldnum  = tablePtr->tabsNum;
	
	tablePtr->tabs = malloc (sizeof (int*) * num);
	if (tablePtr->tabs == NULL)
	{
	    Tcl_Free ((char *) tabs);
	    return TCL_ERROR;
	}
	tablePtr->tabsNum = num;
	for (i=0, p=tabs[i]; i<num; i++, p=tabs[i])
	{
	    /* don't change value of a column if it usually hidden */
	    if (tablePtr->tabsHidden == NULL || !tablePtr->tabsHidden[i])
	    {
		tablePtr->tabs[i] = Max (tablePtr->tabsMinSpace, atoi (p));
	    }
	    else if (oldtabs != NULL && i < oldnum) /* calculated value */
	    {
	    	tablePtr->tabs[i] = oldtabs[i];
	    }
	    else if (tablePtr->defTabs != NULL)
	    {
		tablePtr->tabs[i] = tablePtr->defTabs [i];
	    }
	    else
	    {
		tablePtr->tabs[i] = tablePtr->tabsMinSpace;
	    }
	}
	if (oldtabs != NULL)
	{
	    free (oldtabs);
	}
    	/*
	 * save tab stops as default. this is usefull by bestfit and
	 * autofit
	 */
	if (tablePtr->defTabs == NULL)
	{
	    tablePtr->defTabs = malloc (sizeof (int*) * num);
	    if (tablePtr->defTabs == NULL)
	    {
		Tcl_Free ((char *) tabs);
		return TCL_ERROR;
	    }
	    for (i=0, p=tabs[i]; i<num; i++, p=tabs[i])
	    {
		tablePtr->defTabs[i] = tablePtr->tabs[i];
	    }
	}
	Tcl_Free ((char *) tabs);
	
    }
    
    /*
     * actual default tab stop list
     */
    if (oldDefTabsList != tablePtr->defTabsList && tablePtr->defTabsList != NULL)
    {
	char **tabs, *p;
	int num;
	if (Tcl_SplitList (tablePtr->interp, tablePtr->defTabsList, &num, &tabs) != TCL_OK)
	    return TCL_ERROR;
	if (num == 0)
	{
	    Tcl_Free ((char *) tabs);
	    return TCL_ERROR;
	}
	if (tablePtr->defTabs != NULL)
	{
	    free (tablePtr->defTabs);
	}
	
	tablePtr->defTabs = malloc (sizeof (int*) * num);
	if (tablePtr->defTabs == NULL)
	{
	    Tcl_Free ((char *) tabs);
	    return TCL_ERROR;
	}
	for (i=0, p=tabs[i]; i<num; i++, p=tabs[i])
	{
	    tablePtr->defTabs[i] = Max (0, atoi (p));
	}
    	/*
	 * if tab stop list isn't given set this tab stop to it
	 */
	if (tablePtr->tabs == NULL)
	{
	    tablePtr->tabs = malloc (sizeof (int*) * num);
	    if (tablePtr->tabs == NULL)
	    {
		Tcl_Free ((char *) tabs);
		return TCL_ERROR;
	    }
	}
	/*
	 * the tab stop number must be equevalent
	 * take minimun, else we get a 'core dump'
	 */
	if (tablePtr->tabsNum != num)
	{
	    tablePtr->tabsNum = Min (tablePtr->tabsNum, num);
	}
	/* set the tabs and deftabs as equevalent for now */
	for (i=0, p=tabs[i]; i<num; i++, p=tabs[i])
	{
	    tablePtr->tabs[i] = tablePtr->defTabs[i];
	}
	Tcl_Free ((char *) tabs);
	
    }
    
    /*
     * if justify is changed
     */
    if (oldJustify != tablePtr->justify)
    {
	char **jj, *p;
	int num;
	if (Tcl_SplitList (tablePtr->interp, tablePtr->justify, &num, &jj) != TCL_OK)
	    return TCL_ERROR;
	if (num != tablePtr->tabsNum)
	{
	    Tcl_Free ((char *) jj);
	    return TCL_ERROR;
	}
    	if (tablePtr->tabsJustify)
	{
	    free(tablePtr->tabsJustify);
	}
	tablePtr->tabsJustify = malloc (sizeof (int*) * tablePtr->tabsNum);
	for (i=0, p=jj[i]; i<num; i++, p=jj[i])
	{
	    if (Tcl_GetInt(tablePtr->interp, jj[i], &tablePtr->tabsJustify[i]) != TCL_OK)
	    {
		return TCL_ERROR;
	    }
	}
	Tcl_Free ((char *) jj);
    }
    
    /* allocate memory for tabs state (hidden/viewed) */
    if (tablePtr->tabsHidden == NULL && tablePtr->tabs && tablePtr->tabsNum > 0)
    {
	tablePtr->tabsHidden = calloc (sizeof (int*) * tablePtr->tabsNum, 1);
    }
    
    if (oldSelectModeStr != tablePtr->selectModeStr && tablePtr->selectModeStr != NULL)
    {
	tablePtr->selectMode =
		(strcmp (tablePtr->selectModeStr, "single") == 0 ||
		 strcmp (tablePtr->selectModeStr, "browse") == 0
				? SINGLE
				: MULTI);
    }
    
    /* only on this position we can look for bestfit */
    if (oldBestFit != tablePtr->BestFit && tablePtr->tabs && tablePtr->BestFit)
    {
	TreeTableComputeWidths (tablePtr, NULL, 3);
    }
    
#ifdef USE_PATHFINDER
    if (oldUseAccel != tablePtr->UseAccelerator)
    {
	if (tablePtr->UseAccelerator)
	{
	    if (tablePtr->pathFinder == NULL)
	    {
		tablePtr->pathFinder = calloc (sizeof (PathFinder_t), 1);
	    }
	}
	else
	{
	    PathFinder_t *pf, *pf1;
	    for (pf1=tablePtr->pathFinder; pf1; )
	    {
	    	pf = pf1;
	    	pf1 = pf1->next;
		free (pf);
	    }
	    tablePtr->pathFinder = NULL;
	}
    }
#endif

    tablePtr->numLines = TREETABLE_NUM_LINES(tablePtr);
			
    if (tablePtr->numLines <0)
    {
	tablePtr->numLines = 0;
    }
    
    /* compute width newely if needed, else it take long time */
    if (compute_width)
    {
	TreeTableComputeWidths(tablePtr, (TableItem *) NULL, 2);
	tablePtr->flags |= UPDATE_V_SCROLLBAR;
    }
    pixelWidth = width*tablePtr->xScrollUnit
    		+ 2*tablePtr->inset
		+ 2*tablePtr->selBorderWidth;
    pixelHeight = height*tablePtr->lineHeight + 2*tablePtr->borderWidth;
    
    if (((TkWindow *)tablePtr->tkwin)->reqWidth != pixelWidth ||
    	((TkWindow *)tablePtr->tkwin)->reqHeight != pixelHeight)
    {
	Tk_GeometryRequest   (tablePtr->tkwin, pixelWidth, pixelHeight);
    }
    Tk_SetInternalBorder (tablePtr->tkwin, tablePtr->borderWidth);
    if (tablePtr->setGrid)
    {
	Tk_SetGrid(tablePtr->tkwin, width, height, tablePtr->xScrollUnit,
		   tablePtr->lineHeight);
    }
    tablePtr->flags |= UPDATE_V_SCROLLBAR|UPDATE_H_SCROLLBAR;
    TreeTableRedrawRange(tablePtr);
    return TCL_OK;
}


static int
DisplayRecursive (register TreeTable *tablePtr,
		  register TableItem * itemPPtr,
		  register int* pos,
		  register int* rpos,
		  int limit,
		  Pixmap pixmap,
		  int disp)
{
    GC gc=0;
    Tk_Font font;
    Tk_FontMetrics metrics;
    TableItem *parentPtr;
    TableItem * itemPtr;
    int x, y, x1, y1, hidden;
    int redisplay = 0, windowWidth = -1;
    int tabx = tablePtr->inset - tablePtr->xOffset;
    int lineHeight = tablePtr->lineHeight;
    char tmp [1024];
    
#ifdef USE_PATHFINDER
    PathFinder_t *Pf=tablePtr->pathFinder;
#endif
    
    itemPtr = itemPPtr;

#ifdef USE_PATHFINDER
    if (Pf)
    {
	/* we only use the Pathfinder on the first level
	 * of the tree (root) ***/
    	if (itemPtr->parentPtr) /* parent availiable */
	{
	    Pf = NULL;
	}
	else
	{
	    PathFinder_t *pf=Pf;
	    for (pf=Pf->next; pf; pf=pf->next)
	    {
	        if (pf->itemPos < tablePtr->topIndex)
	        {
		    *pos = *rpos = pf->itemPos;
		    itemPtr = pf->itemPtr;
		}
	    }
	    /* find last path finder */
	    for (;Pf->next; Pf=Pf->next);
	}
    }
#endif

    while (*pos < tablePtr->topIndex && itemPtr != NULL)
    {
#ifdef USE_PATHFINDER
    	    if (Pf && Pf->itemPos < *pos && (*pos%PATHFINDER_STEP)==0)
	    {
	    	Pf->next = calloc (sizeof (PathFinder_t), 1);
		Pf=Pf->next;
		Pf->itemPos = *pos;
		Pf->itemPtr = itemPtr;
	    }
#endif

	    *pos += 1;
	    *rpos += 1;
	    
	    if (itemPtr->succPtr != NULL && ! (itemPtr->flags&ITEM_HIDDEN_SUBTREE))
	    {
		if (*pos+itemPtr->seenNum < tablePtr->topIndex)
		{
		    *rpos += itemPtr->succNum;
		    *pos += itemPtr->seenNum;
		}
		else
		{
		    redisplay += DisplayRecursive
			(tablePtr, itemPtr->succPtr, pos, rpos, limit, pixmap,
			disp);
		}
	    }
	    itemPtr = itemPtr->nextPtr;
    }
    while ((itemPtr != NULL) && (*pos <= limit))
    {
#ifdef USE_PATHFINDER
	/* Add only the root items in the pathfinder */
    	if (Pf && Pf->itemPos < *pos && (*pos%PATHFINDER_STEP)==0)
	{
	    Pf->next = calloc (sizeof (PathFinder_t), 1);
	    Pf=Pf->next;
	    Pf->itemPos = *pos;
	    Pf->itemPtr = itemPtr;
	}
#endif
	hidden = itemPtr->flags&ITEM_HIDDEN_SUBTREE;
	/*
	 * Compute starting x-coord. for the line.
	 */
	x = tabx + itemPtr->indent*tablePtr->indentWidth;
	/*
	 * Increment x position by separator space.
	 */
	x1 = x + tablePtr->bitmapSpace;
	if (itemPtr->bitmap != None || itemPtr->image) {
	    x1 += itemPtr->bitmapWidth;
	}
	
	/* y-coordinate for bitmap */
	y = ((*pos - tablePtr->topIndex) * lineHeight)
		+ tablePtr->borderWidth;
	/*
	 * Compute y coordinate to center the text vertically on the line.
	 */
	y1  = y + lineHeight;
	y  += (lineHeight-itemPtr->bitmapHeight)/2;
	y1 -= (lineHeight-itemPtr->fontHeight)/2;

	/*
	 * If this line is part of the selection, draw the higlight now 
	 */
	if (disp && (itemPtr->flags&ITEM_SELECTED))
	{
	    if (tablePtr->fillSelection && windowWidth == -1)
	    {
		windowWidth = TREETABLE_WIDTH(tablePtr);
	    }
	    Tk_Fill3DRectangle(tablePtr->tkwin, pixmap,
			       tablePtr->selBorder,
			       x1 - 2,
			       y-(lineHeight-itemPtr->bitmapHeight)/2,
			       (tablePtr->fillSelection)
					? Max (windowWidth, itemPtr->lineWidth) - x1 + 4
					: itemPtr->lineWidth - x1 + 4,
			       lineHeight,
			       
			       tablePtr->selBorderWidth,
			       TK_RELIEF_RAISED);
	}

	/*
	 * Draw the line back to the parent if necessary
	 */
	if (itemPtr->parentPtr != (TableItem *) NULL)
	{
	    register TableItem *parent;
	    register int px, py, mx, my;
	    
	    /*
	     * Compute x and y of parent bitmap
	     */
	    parent = itemPtr->parentPtr;
	    py = y - lineHeight + itemPtr->bitmapHeight;
	    px = (parent->indent*tablePtr->indentWidth) + 
	      tablePtr->borderWidth + parent->bitmapWidth/2 - 
		tablePtr->xOffset;
	    mx = x - 1;
	    my = y + itemPtr->bitmapHeight/2;
	    
	    if (py < 0) py = 0;
	    
	    /* horizontal line */
	    if (disp)
	      XDrawLine(tablePtr->display, pixmap,
#ifdef USE_BITMAP_COLOR
		itemPtr->lineGC != None ? itemPtr->lineGC : tablePtr->defLineGC,
#else
		tablePtr->defLineGC,
#endif
		mx, my, px, my);
		  
	    if (itemPtr->nextPtr != NULL)
	    {
	       my = y + lineHeight;
	    }
	    
	    /* vertical line */
	    if (disp)
	      XDrawLine(tablePtr->display, pixmap,
#ifdef USE_BITMAP_COLOR
	    	itemPtr->lineGC != None ? itemPtr->lineGC : tablePtr->defLineGC,
#else
		tablePtr->defLineGC,
#endif
		px, my, px, py);
	}
	
	/*
	 * Draw the bitmap 
	 */
	if (disp
	    && ((! hidden && (itemPtr->bitmap != None || itemPtr->image))
		|| (hidden && (tablePtr->hiddenImage != NULL
			       || tablePtr->hiddenBitmap != None))))
	  {
	    
	    /* draw image if available */
	    if (! hidden && itemPtr->image != NULL)
	    {
		Tk_RedrawImage(itemPtr->image,
				0, 0,
				itemPtr->bitmapWidth, itemPtr->bitmapHeight,
				pixmap,
				x, y);
	    }
	    else if (hidden && tablePtr->hiddenImage != NULL)
	    {
	    	Tk_RedrawImage(tablePtr->hiddenImage,
				0, 0,
				tablePtr->hiddenWidth, tablePtr->hiddenHeight,
				pixmap,
				x, y);
	    }
	    else
	    {
		register int width, height;
	    	/* 
	     	* If line in selection, use bitmapSelectGC
	     	*/
	    	if (itemPtr->flags&ITEM_SELECTED) {
#ifdef USE_BITMCAP_COLOR
		    gc = tablePtr->selectGC;
		    if (!hidden && itemPtr->bitmapSelectGC != None) {
		        gc = itemPtr->bitmapSelectGC;
		    } else {
		        if (tablePtr->defBitmapSelectGC != None) {
			    gc = tablePtr->defBitmapSelectGC;
		        } else {
			    gc = tablePtr->selectGC;
		        }
		    }
#else
		    if (!hidden && tablePtr->defBitmapSelectGC != None) {
			gc = tablePtr->defBitmapSelectGC;
		    } else {
			gc = tablePtr->selectGC;
		    }
#endif
	        } else {
		    if (!hidden && itemPtr->bitmapGC != None) {
		        gc = itemPtr->bitmapGC;
		    } else {
		        gc = tablePtr->defBitmapGC;
		    }
	        }
	    
	        if (hidden)
	        {
		    width  = tablePtr->hiddenWidth;
		    height = tablePtr->hiddenHeight;
	        }
	        else
	        {
	    	    width  = itemPtr->bitmapWidth;
		    height = itemPtr->bitmapHeight;
	        }
	        XCopyPlane(tablePtr->display,
	    	       /* view Bitmap for hidden sub tree, if availiable */
	    	       hidden
			   ? tablePtr->hiddenBitmap
		           : itemPtr->bitmap,
		       pixmap,
		       gc,
		       0, 0,
		       width, height,
		       x, y,
		       1);
	    }
	}

	if (disp)
	{
	if (itemPtr->flags&ITEM_SELECTED)
	{
	    gc = tablePtr->selectGC;
	    /* here we need to make sure our font matches the
	     * current line. */
	    if (itemPtr->fontPtr != NULL)
	    {
		XSetFont(tablePtr->display, gc, Tk_FontId(itemPtr->fontPtr));
	    }
	}
	else if (itemPtr->textGC != None)
	{
	    gc = itemPtr->textGC;
	}
	else
	{
	    gc = tablePtr->defTextGC;
	}
	}
	font = ((itemPtr->fontPtr != NULL) ? itemPtr->fontPtr
		: tablePtr->defFontPtr);

	/*
	 * we support tab stops too
	 */
	if (tablePtr->tabsNum > 0 && tablePtr->tabs != NULL)
	{
	    char *nextTab, *p = itemPtr->text;
	    int len;
	    int tx = tabx, px = x1;
	    int i = 0, width, width1;
	    
	    while (p != NULL)
	    {
		NextChar (nextTab, p, '\t', len);
		
		/*
		 * there is no tab stops availiable, so we don't need to
		 * recalc the list width
		 */
		width = Tk_TextWidth (font, p, len);
		
		/* Calculate auto fit */
		if (!disp && tablePtr->AutoFit)
		{
		    int xx = width+tablePtr->tabsMinSpace;
		    if (p == itemPtr->text)
		    {
			xx += ITEM_TEXT_INDENT(itemPtr);
		    }
		    if (tablePtr->tabs[i] < xx)
		    {
			tablePtr->tabs[i] = xx;
		    }
		}
		
		/* calculate max. width, if tab stops availiable */
		if (nextTab == NULL)
		{
		    if (itemPtr->lineWidth != px + width)
		    {
			itemPtr->lineWidth = px + width + tablePtr->xOffset;
		    }
		    if (tablePtr->maxWidth < itemPtr->lineWidth)
		    {
			tablePtr->maxWidth = itemPtr->lineWidth;
		    }
		}
		
		/* if truncate is enabled, truncate the rest of the
		 * text if it overrides the next tab stop.
		 * It uses a binary search
		 */
		if (tablePtr->Truncate && i<tablePtr->tabsNum-1)
		{
		    int x11;
		    int tabw;
		
		    if (i == 0)
		    {
		    	x11 = ITEM_TEXT_INDENT(itemPtr);
			width1 = x11+width;
		    }
		    else
		    {
			width1 = width;
			x11=0;
		    }
		    
		    if (tablePtr->tabsHidden[i])
		    {
			tabw = tablePtr->tabsMinSpace;
		    }
		    else
		    {
			tabw = tablePtr->tabs[i];
		    }
		    if (width1 > tabw-tablePtr->tabsMinSpace)
		    {
			char *q = p;
			int step;
			/*
			 * use binary algorithm O(log n) to truncate the text
			 */
			len += 2;
			step = len;
			do
			{
			    step /= 2;
			    if (width1 > tabw-tablePtr->tabsMinSpace)
			    {
			    	len -= step;
			    }
			    else
			    {
			    	len += step;
			    }
			    strncpy (tmp, q, len);
			    if (len > 2)
			    {
			    	p = tmp + len - 2;
			        *p ++ = '.';
			    	*p ++ = '.';
			    }
			    else if (len == 2)
			    {
			    	p = tmp + len - 1;
			    	*p ++ = '.';
			    }
			    else
			    {
			    	p = tmp + 1;
			    }
			    p[0] = 0;
			    p = tmp;
				
			    width = Tk_TextWidth (font, p, len);
			    if (i == 0)
			    {
			        width1 = x11+width;
			    }
			    else
			    {
			        width1 = width;
			    }

			} while (step > 1);
			
			/* 
			 * After the binary algorithm to truncate the text
			 * The text can be one or two characters larger than it
			 * fit in the tab column. Truncate it linear.
			 */
			while (len > 0 && width1 > tabw-tablePtr->tabsMinSpace)
			{
			    len --;
			    strncpy (tmp, q, len);
			    if (len > 2)
			    {
			    	p = tmp + len - 2;
			        *p ++ = '.';
			    	*p ++ = '.';
			    }
			    else if (len == 2)
			    {
			    	p = tmp + len - 1;
			    	*p ++ = '.';
			    }
			    else
			    {
			    	p = tmp + 1;
			    }
			    p[0] = 0;
			    p = tmp;
				
			    width = Tk_TextWidth (font, p, len);
			    if (i == 0)
			    {
			        width1 = x11+width;
			    }
			    else
			    {
			        width1 = width;
			    }
			};
		    }
		}

		if (disp)
		{
		    Tk_GetFontMetrics (font, &metrics);
		    /* right-justify */
		    if (tablePtr->tabsJustify && tablePtr->tabsJustify[i])
		    {
			int tabw, ww = Tk_TextWidth(font, p, len);
			if (tablePtr->tabsHidden[i])
			{
			     tabw = tablePtr->tabsMinSpace;
			}
			else
			{
			     tabw = tablePtr->tabs[i];
			}
			Tk_DrawChars (tablePtr->display, pixmap, gc, font,
			      p, len,
			      px+tabw-ww-tablePtr->tabsMinSpace/2, y1 - metrics.descent);
		    }
		    else
		    {
			Tk_DrawChars (tablePtr->display, pixmap, gc, font,
				  p, len,
				  px+(i==0 ? 0 : tablePtr->tabsMinSpace/2),
				  y1 - metrics.descent);
		    }
		}
		
		/* calculate the next tab stop position */
		if (nextTab)
		{
		    p = nextTab + 1;
		}
		else
		{
		    p = nextTab;
		}
		
		if (tablePtr->tabsHidden[i])
		{
		    if (i == 0)
		    {
			tx += ITEM_TEXT_X(itemPtr);
		    }
		    else
		    {
		        tx += tablePtr->tabsMinSpace;
		    }
		}
		else
		{
		    tx += tablePtr->tabs[i];
		}

		/* if truncate is seted, we cut the rest of the string */
		if (!tablePtr->Truncate)
		{
		    px = Max (px+width+tablePtr->tabsMinSpace, tx);
		}
		else
		{
		    px = tx;
		}
		
		if (i<tablePtr->tabsNum-1)
		{
		    i++;
		}
	    }
	}
	else
	{
	    if (disp)
	    {
	        Tk_GetFontMetrics (font, &metrics);
		Tk_DrawChars (tablePtr->display, pixmap, gc, font,
			      itemPtr->text, itemPtr->textLength,
			      x1, y1 - metrics.descent);
	    }
	    if (itemPtr->lineWidth > tablePtr->maxWidth)
	    {
		tablePtr->maxWidth = itemPtr->lineWidth;
	    }
	}
	
	/* draw line if have keyboard focus */
	if (disp && (*rpos == tablePtr->active) && (tablePtr->flags & GOT_FOCUS))
	{
	    XFillRectangle(tablePtr->display, pixmap,
	    	     (itemPtr->textGC != None)
			? itemPtr->textGC
			: tablePtr->defTextGC, 
		     x1,
		     y - (lineHeight-itemPtr->bitmapHeight)/2
		       + lineHeight
		       - 3,
		     (tablePtr->fillSelection)
			? Max (TREETABLE_WIDTH(tablePtr), itemPtr->lineWidth) - x1
			: itemPtr->lineWidth - x1 /*width*/ ,
		     ACTIVE_LINE_WIDTH);
	}
      
	/* if we have a selected line, we need to restore selectGC's font */
	if(disp && itemPtr->fontPtr != NULL)
	{
	    XSetFont(tablePtr->display, tablePtr->selectGC,
		 Tk_FontId (tablePtr->defFontPtr));
	}
	
	/* parse the parents upper first level (not first level)
	 * and draw the lines if a next item of a parent is availiable
	 */
	parentPtr = itemPtr->parentPtr;
	if (disp) for (; parentPtr != NULL; parentPtr = parentPtr->parentPtr)
	{
	   register int px, py, my;
	
	   
	   if (parentPtr->nextPtr == NULL)
	   {
	   	continue;
	   }
	   /*
	    * Compute x and y of parent bitmap
	    */
	   py = ((*pos - tablePtr->topIndex) * lineHeight) +
		tablePtr->borderWidth;
	   if (py < 0) py = 0;
	   my = py + lineHeight;
	   if (my < 0) my = 0;
	   
	   px = (parentPtr->indent-1)*tablePtr->indentWidth + 
		tablePtr->borderWidth + parentPtr->bitmapWidth/2 - 
		tablePtr->xOffset;

	   XDrawLine(tablePtr->display, pixmap,
#ifdef USE_BITMAP_COLOR
			parentPtr->lineGC != None ? parentPtr->lineGC : tablePtr->defLineGC,
#else
		tablePtr->defLineGC,
#endif
			px, my, px, py);
	}
	
	*pos += 1;
	*rpos += 1;
	if (itemPtr->succPtr != NULL && !hidden)
	{
	     redisplay += DisplayRecursive
			(tablePtr, itemPtr->succPtr, pos, rpos, limit, pixmap,
			disp);
	}
	else
	{
	     *rpos += itemPtr->succNum;
	}
	itemPtr = itemPtr->nextPtr;
    }
    return redisplay;
}

/*
 *--------------------------------------------------------------
 *
 * DisplayTreeTable --
 *
 *      This procedure redraws the contents of a treetable window.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Information appears on the screen.
 *
 *--------------------------------------------------------------
 */
#define USE_PIXMAP
static void
DisplayTreeTable(clientData)
     ClientData clientData;            /* Information about window. */
{
    register TreeTable *tablePtr = (TreeTable *) clientData;
    register Tk_Window tkwin = tablePtr->tkwin;
    int limit, pos, rpos, oldMaxWidth;
    Pixmap pixmap=0;
    int redisplay;
    int i;
    
    /*
     * Iterate through all of the items in the treetable,
     * displaying each in turn. 
     */
    limit = tablePtr->topIndex + tablePtr->numLines + 1;
    if (limit >= tablePtr->numItems)
    {
	limit = tablePtr->numItems-1;
    }
    
    /* calculate best auto fit for tab stops */
    if (tablePtr->AutoFit && tablePtr->tabsNum > 0)
    {
	for (i=0; i<tablePtr->tabsNum; i++)
	{
	    tablePtr->tabs[i] = tablePtr->defTabs[i];
	}
    	pos = rpos = 0;
	tablePtr->maxWidth = 0;
    	DisplayRecursive(tablePtr, tablePtr->itemPtr,
			&pos, &rpos,
			limit, pixmap,
			0);
    }
    
    tablePtr->flags &= ~REDRAW_PENDING;
    if (tablePtr->flags & UPDATE_V_SCROLLBAR)
    {
	TreeTableUpdateVScrollbar(tablePtr);
    }
    tablePtr->flags &= ~(REDRAW_PENDING|UPDATE_V_SCROLLBAR);
    if ((tablePtr->tkwin == NULL) || !Tk_IsMapped(tkwin)) {
	return;
    }

    /*
     * Redrawing is done in a temporary pixmap that is allocated here and freed
     * at the end of the procedure.  All drawing is done to the pixmap,
     * and the pixmap is copied to the screen at the end of the procedure.
     * This provides the smoothest possible visual effects (no flashing
     * on the screen).
     */
#ifndef USE_PIXMAP
    pixmap = Tk_WindowId(tablePtr->tkwin);
    Tcl_Preserve((ClientData) tablePtr->interp);
#else
    pixmap = Tk_GetPixmap(tablePtr->display, Tk_WindowId(tkwin),
			   Tk_Width(tkwin), Tk_Height(tkwin),
			   Tk_Depth(tkwin));
#endif

    Tk_Fill3DRectangle(tkwin, pixmap, tablePtr->normalBorder,
		       0, 0, Tk_Width(tkwin), Tk_Height(tkwin), 
		       tablePtr->borderWidth, tablePtr->relief);
    
    oldMaxWidth = tablePtr->maxWidth;
    tablePtr->maxWidth = 0;
    pos = 0;
    rpos = 0;
    redisplay = DisplayRecursive
    			(tablePtr, tablePtr->itemPtr,
			&pos, &rpos,
			limit, pixmap,
			1);

    /* if drawing lines is enabled */
    if (tablePtr->SplitLines && tablePtr->tabsNum > 0)
    {
    	int px = tablePtr->tabsMinSpace/2 - tablePtr->xOffset - 1;
	int hight = Tk_Height (tablePtr->tkwin);
	for (i=0; i<tablePtr->tabsNum-1; i++)
	{
	    if (tablePtr->tabsHidden[i])
	    {
		if (i==0 && tablePtr->itemPtr!=NULL)
		{
		    px += ITEM_TEXT_X(tablePtr->itemPtr);
		}
		else
		{
		    px += tablePtr->tabsMinSpace;
		}
	    }
	    else
	    {
		px += tablePtr->tabs[i];
	    }
	    if (px > 0)
	    {
	      XDrawLine (tablePtr->display, pixmap, tablePtr->defSplitLineGC,
	   		px,
			0,
			px, hight);
	    }
	}
    }
    if ((tablePtr->flags & UPDATE_H_SCROLLBAR) || oldMaxWidth != tablePtr->maxWidth)
    {
        TreeTableUpdateHScrollbar(tablePtr);
    	tablePtr->flags &= ~UPDATE_H_SCROLLBAR;
    }
    
    /*
     * Redraw the border to ensure it convers the text items of
     * the lines in the treetable.
     */
    Tk_Draw3DRectangle(tkwin, pixmap, 
		       tablePtr->normalBorder, 0, 0,
		       Tk_Width(tkwin), Tk_Height(tkwin),
		       tablePtr->borderWidth,
		       tablePtr->relief);
		       
    if (tablePtr->highlightWidth > 0 && tablePtr->highlightBgColor)
    {
      GC gc;

      if (tablePtr->flags & GOT_FOCUS)
      {
	gc = Tk_GCForColor(tablePtr->highlightColor, pixmap);
      } else
      {
	gc = Tk_GCForColor(tablePtr->highlightBgColor, pixmap);
      }
      Tk_DrawFocusHighlight(tkwin, gc, tablePtr->highlightWidth, pixmap);
    }
    
    /* if idle command is seted, execute it */
    if (tablePtr->idlecommand != NULL)
    {
	Tcl_Eval (tablePtr->interp, tablePtr->idlecommand);
    }
    
#ifdef USE_PIXMAP
    XCopyArea(tablePtr->display, pixmap, Tk_WindowId(tkwin),
	      tablePtr->defTextGC, 0, 0, Tk_Width(tkwin),
	      Tk_Height(tkwin), 0, 0);
    Tk_FreePixmap(tablePtr->display, pixmap);
#else
    Tcl_Release((ClientData) tablePtr->interp);
#endif
}

/*
 * This procedure is called, when an item is added or deleted and
 * the flag 'bestfit' is enabled
 */
static void
ComputeTabStops (TreeTable*tablePtr, TableItem *itemPtr)
{
    char *nextTab, *p = itemPtr->text;
    int len;
    int tx, px;
    int i = 0, width;
    int tabx, x, x1;
    
    if (tablePtr->tabs == NULL)
    {
	return;
    }
    
    /* first x position fot text displaying */
    tabx = tablePtr->inset - tablePtr->xOffset;
    x = tabx + itemPtr->indent*tablePtr->indentWidth;
    x1 = x + tablePtr->bitmapSpace;
    if (itemPtr->bitmap != None || itemPtr->image) {
	x1 += itemPtr->bitmapWidth;
    }
    tx = tabx;
    px = x1;
    
    while (p != NULL)
    {
	NextChar (nextTab, p, '\t', len);
	
	/*
	 * there is no tab stops availiable, so we don't need to
	 * recalc the list width
	 */
	width = Tk_TextWidth ((itemPtr->fontPtr != NULL)
			      ? itemPtr->fontPtr : tablePtr->defFontPtr,
			      p, len);
	
	
	/* Calculate best fit */
	if (tablePtr->BestFit)
	{
	    int xx = width+tablePtr->tabsMinSpace;
	    if (p == itemPtr->text)
	    {
		int x11;
		x11 = itemPtr->indent*tablePtr->indentWidth
			+ tablePtr->bitmapSpace
			+ itemPtr->bitmapWidth
			+ tablePtr->inset;
		xx += x11;
	    }
	    if (tablePtr->tabs[i] < xx)
	    {
		tablePtr->tabs[i] = xx;
	    }
	}
	
	/* calculate max. width */
	if (nextTab == NULL)
	{
	    if (itemPtr->lineWidth != px + width)
	    {
		itemPtr->lineWidth = px + width + tablePtr->xOffset;
	    }
	    if (tablePtr->maxWidth < itemPtr->lineWidth)
	    {
		tablePtr->maxWidth = itemPtr->lineWidth;
	    }
	}
	
	/* calculate the next tab stop position */
	if (nextTab)
	{
	    p = nextTab + 1;
	}
	else
	{
	    p = nextTab;
	}
	tx += tablePtr->tabs[i];

	px = tx;
	
	if (i<tablePtr->tabsNum-1)
	{
	    i++;
	}
    }
}

static int
TreeTableFindIndex_x (TableItem *itemPtr, TableItem *srcPtr,
		      int index, int* pos)
{
    TableItem *Ptr;
    int i = index;
    for (Ptr=itemPtr; Ptr; Ptr=Ptr->nextPtr)
    {
    	if (Ptr == srcPtr)
	{
	   *pos = i;
	   break;
	}
	i++;
	if (Ptr->succPtr != NULL)
	{
	   i = TreeTableFindIndex_x (Ptr->succPtr, srcPtr, i, pos);
	}
    }
    return i;
}

static int
TreeTableFindIndex(tablePtr, itemPtr) 
     TreeTable *tablePtr;    /* treetable to search in */
     TableItem *itemPtr;     /* the item we're looking for */
{
    int pos = -1;
    
    RET_CACHED_INDEX (itemPtr);
    
    TreeTableFindIndex_x (tablePtr->itemPtr, itemPtr, 0, &pos);
    
    /*SET_CACHED(itemPtr, pos);*/
    
    return pos;
}

/*
 *----------------------------------------------------------------------
 *
 * TreeTableSelectImageProc --
 *
 *	This procedure is invoked by the image code whenever the manager
 *	for an image does something that affects the size of contents
 *	of the image displayed in a tree table when it is selected.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May arrange for the button to get redisplayed.
 *
 *----------------------------------------------------------------------
 */
static void
TreeTableImageProc(clientData, x, y, width, height, imgWidth, imgHeight)
    ClientData clientData;		/* Pointer to widget record. */
    int x, y;				/* Upper left pixel (within image)
					 * that must be redisplayed. */
    int width, height;			/* Dimensions of area to redisplay
					 * (may be <= 0). */
    int imgWidth, imgHeight;		/* New dimensions of image. */
{
}
/*
 *----------------------------------------------------------------------
 *
 * ConfigureTreeTableItem --
 *
 *      Add a new item to the tree table.
 *      
 * Results:
 *      None.
 *
 * Side effects:
 *      New information gets added to tablePtr;  it will be redisplayed
 *      soon, but not immediately.
 *
 *----------------------------------------------------------------------
 */

static int
ConfigureTreeTableItem(tablePtr, itemPtr, argc, argv, flags, mode) 
     TreeTable *tablePtr;             /* table the item is located in */
     TableItem  *itemPtr;             /* Item to configure */
     int argc;                        /* number of arguments */
     char **argv;                     /* argument strings */
     int flags;                       /* Flags to pass to Tk_ConfigureWidget */
     int mode;
{
    XGCValues gcValues;
    XColor *oldtextfg;
#if USE_BITMAP_COLORS
    XColor *oldbitselfg, *oldbitselbg, *oldbitbg, *oldlinefg, *oldbitfg;
#endif
    Pixmap oldbitmap;
    Tk_Font oldfont;
    Tk_FontMetrics metrics;
    char *oldtext;
    int oldparent;
    int oldmaxwidth;
    int oldlinewidth;
    int oldindent;
    int recompute_height = 0;
    int recompute_width = 0, oldSuccNum, oldSeenNum;
    char *oldImageString = NULL;

    oldbitmap = itemPtr->bitmap;
    oldtext = itemPtr->text;
    oldtextfg = itemPtr->textFgColor;
#if USE_BITMAP_COLORS
    oldbitfg = itemPtr->bitmapFgColor;
    oldbitbg = itemPtr->bitmapBgColor;
    oldbitselfg = itemPtr->bitmapSelectFgColor;
    oldbitselbg = itemPtr->bitmapSelectBgColor;
    oldlinefg = itemPtr->lineFgColor;
#endif
    oldfont = itemPtr->fontPtr;
    oldmaxwidth = tablePtr->maxWidth;
    oldparent = itemPtr->parent;
    oldindent = itemPtr->indent;
    oldlinewidth = itemPtr->lineWidth;
    oldImageString = itemPtr->imageString;
    oldSuccNum = itemPtr->succNum;
    oldSeenNum = itemPtr->seenNum;
    
    if (Tk_ConfigureWidget(tablePtr->interp, tablePtr->tkwin, itemConfigSpecs,
			   argc, argv, (char *) itemPtr, flags) != TCL_OK) {
	
	return TCL_ERROR;
    }
    
    /** succnum field is read only */
    if (oldSuccNum != itemPtr->succNum)
    {
	itemPtr->succNum = oldSuccNum;
	Tcl_AppendResult (tablePtr->interp,
			"children number is readonly field and can't be changed",
			NULL);
	return TCL_ERROR;
    }
    /** seennum field is read only */
    if (oldSeenNum != itemPtr->seenNum)
    {
	itemPtr->seenNum = oldSeenNum;
	Tcl_AppendResult (tablePtr->interp,
			"seen number of children is readonly and can't be changed",
			NULL);
	return TCL_ERROR;
    }
    
    if (itemPtr->indent != oldindent) {
	recompute_width = 1;
    }
    if ((oldtextfg != itemPtr->textFgColor) || (oldfont != itemPtr->fontPtr)) {
	XColor *foreground;

	if (itemPtr->textFgColor != None) {
	    foreground = itemPtr->textFgColor;
	} else {
	    foreground = tablePtr->defTextFgColor;
	}
	gcValues.foreground = foreground->pixel;
	if (itemPtr->fontPtr != None) {
	    gcValues.font = Tk_FontId (itemPtr->fontPtr);
	} else {
	    gcValues.font = Tk_FontId (tablePtr->defFontPtr);
	}
	gcValues.graphics_exposures = False;
	if (itemPtr->textGC != None) {
	    Tk_FreeGC(tablePtr->display, itemPtr->textGC);
	}
	if (oldfont != itemPtr->fontPtr) {
	    Tk_GetFontMetrics (itemPtr->fontPtr, &metrics);
	    itemPtr->fontHeight = metrics.linespace;
	    recompute_height = 1;
	}
	itemPtr->textGC = Tk_GetGCColor(tablePtr->tkwin,
				   GCForeground|GCFont|GCGraphicsExposures,
				   &gcValues, foreground, NULL);
    }
#if USE_BITMAP_COLORS
    if ((oldbitfg != itemPtr->bitmapFgColor) || 
	(oldbitbg != itemPtr->bitmapBgColor)) {
	XColor *foreground, *background;

	if (itemPtr->bitmapFgColor != None) {
	    foreground = itemPtr->bitmapFgColor;
	} else {
	    foreground = tablePtr->defBitmapFgColor;
	}
	gcValues.foreground = foreground->pixel;
	if (itemPtr->bitmapBgColor != None) {
	    background = itemPtr->bitmapBgColor;
	} else {
	    background = tablePtr->defBitmapBgColor;
	}
	gcValues.background = background->pixel;
	gcValues.graphics_exposures = False;
	if (itemPtr->bitmapGC != None) {
	    Tk_FreeGC(tablePtr->display, itemPtr->bitmapGC);
	}
	itemPtr->bitmapGC = Tk_GetGCColor(tablePtr->tkwin, 
					  GCForeground|GCBackground|
					  GCGraphicsExposures, &gcValues,
					  foreground, background);
    }

    if ((oldbitselfg != itemPtr->bitmapSelectFgColor) ||
	(oldbitselbg != itemPtr->bitmapSelectBgColor)) {
	XColor *foreground, *background;

	if (itemPtr->bitmapSelectFgColor != None) {
	    foreground = itemPtr->bitmapSelectFgColor;
	} else {
	    foreground = tablePtr->defBitmapSelectFgColor;
	}
	gcValues.foreground = foreground->pixel;
	if (itemPtr->bitmapSelectBgColor != None) {
	    background = itemPtr->bitmapSelectBgColor;
	} else {
	    /* Zsolt Koppany 5-jul-96 */
	    if (!tablePtr->defBitmapSelectBgColor)
		background = tablePtr->defBitmapFgColor;
	    else
	        background = tablePtr->defBitmapSelectBgColor;
	}
	gcValues.background = background->pixel;
	gcValues.graphics_exposures = False;
	if (itemPtr->bitmapSelectGC != None) {
	    Tk_FreeGC(tablePtr->display, itemPtr->bitmapSelectGC);
	} 
	itemPtr->bitmapSelectGC = Tk_GetGCColor(tablePtr->tkwin,
						GCForeground|GCBackground|
						GCGraphicsExposures,
						&gcValues, foreground,
						background);
						
    }
#endif

    if (itemPtr->lineWidth == atoi(NO_TABLEITEM_LINEWIDTH)) {
	itemPtr->lineWidth = tablePtr->defLineWidth;
    }
#ifdef USE_BITMAP_COLOR
    if ((oldlinefg != itemPtr->lineFgColor) || 
	(oldlinewidth != itemPtr->lineWidth)) {
	XColor *foreground;

	if (itemPtr->lineFgColor != None) {
	    foreground = itemPtr->lineFgColor;
	} else {
	    foreground = tablePtr->defLineFgColor;
	}
	gcValues.foreground = foreground->pixel;
	gcValues.line_width = itemPtr->lineWidth;
	gcValues.graphics_exposures = False;
	if (itemPtr->lineGC != None) {
	    Tk_FreeGC(tablePtr->display, itemPtr->lineGC);
	}
	itemPtr->lineGC = Tk_GetGCColor(tablePtr->tkwin, 
					(GCForeground|GCGraphicsExposures
					 |GCLineWidth),
					&gcValues, foreground, NULL);
    }
#endif
    if ((oldtext != itemPtr->text) || (oldfont != itemPtr->fontPtr)) {
	if (itemPtr->text != (char *) NULL) {
	    itemPtr->textLength = strlen(itemPtr->text);
	} else {
	    itemPtr->textLength = 0;
	}
	recompute_width = 1;
    }
    
    if (oldbitmap != itemPtr->bitmap) {
	if (itemPtr->bitmap != None) {
	    int width, height;
	    Tk_SizeOfBitmap(tablePtr->display, itemPtr->bitmap,
			    &width, &height);
	    itemPtr->bitmapWidth = (short) width;
	    itemPtr->bitmapHeight = (short) height;
	}
	recompute_height = recompute_width = 1;
    }
    /* load image **/
    else if (oldImageString != itemPtr->imageString)
    {
	int width, height;
	Tk_Image image;
	
	image = Tk_GetImage(tablePtr->interp, tablePtr->tkwin,
		itemPtr->imageString, TreeTableImageProc, (ClientData) NULL);
	if (image == NULL)
	{
	    return TCL_ERROR;
	}
	/* free old image */
	if (itemPtr->image != NULL)
	{
	    Tk_FreeImage(itemPtr->image);
	}
	itemPtr->image = image;
	Tk_SizeOfImage(itemPtr->image, &width, &height);
	itemPtr->bitmapWidth  = (short) width;
	itemPtr->bitmapHeight = (short) height;
	recompute_height = recompute_width = 1;
    }
    
    if (! (mode & CONFIG_OPTIONS_ONLY) && recompute_height)
    {
	int oldlh;

	oldlh = tablePtr->lineHeight;
	TreeTableComputeLineHeight(tablePtr, itemPtr, 0);
	if (oldlh != tablePtr->lineHeight)
	{
	    tablePtr->flags |= UPDATE_V_SCROLLBAR;
	}
    }
    if ((mode & CONFIG_OPTIONS_ONLY) == 0 && recompute_width) {
	TreeTableComputeWidths(tablePtr, itemPtr, 0);
	if (oldmaxwidth != tablePtr->maxWidth)
	{
	    tablePtr->flags |= UPDATE_H_SCROLLBAR;
	}
    }	
    if (! (mode & CONFIG_OPTIONS_ONLY))
    {
    	TreeTableRedrawRange(tablePtr);
    }
    return TCL_OK;
}

/*
 * find items in a given range and append result to
 * the Tcl iterpreter result string
 */
/*{*/
static int
TreeTableGetItems_x (TreeTable *tablePtr, TableItem *itemPtr,
		     int index, int from, int to)
{
  TableItem *item = itemPtr;
  int i;
  
  for (i=index; i<=to && item != NULL; item = item->nextPtr)
  {
     if (i >= from && i<= to)
     {
	Tcl_AppendElement (tablePtr->interp, item->text);
     }
     i++;
     if (item->succPtr != NULL)
     {
	/* we don't need to seach in sub trees, when succ number does't
	 * succeed
	 */
	if (i+item->succNum < from)
	{
	    i += item->succNum;
	}
	else
        {
	    i = TreeTableGetItems_x (tablePtr, item->succPtr, i, from, to);
	}
     }
  }
  
  return i;
}

static void
TreeTableGetItems (TreeTable *tablePtr, int from, int to)
{
   TreeTableGetItems_x (tablePtr, tablePtr->itemPtr, 0,
			Min (from, to), Max(from, to));
}
/*}*/

/*{*/
static int
TreeTableIndex (TableItem *itemPtr, int index, int pos, TableItem **itemRet)
{
  TableItem *item = itemPtr;
  int i;
  
  for (i=pos; i<=index && item != NULL; item = item->nextPtr)
  {
     if (i == index)
     {
     	*itemRet = item;
	return -1;
     }
     i++;
     if (item->succPtr != NULL)
     {
	/* we don't need to seach in sub trees, when succ number does't
	 * succeed
	 */
	if (i+item->succNum < index)
	{
	    i += item->succNum;
	}
	else
        {
	    i = TreeTableIndex (item->succPtr, index, i, itemRet);
	    if (i < 0)
	    {
		return i;
	    }
	}
     }
  }
  
  return i;
}
static TableItem * TreeTableFindItem (TreeTable *tablePtr, int index)
{
    int pos = 0;
    TableItem* itemRet = NULL;
    TableItem* itemPtr = tablePtr->itemPtr;
   
    if (index < 0)
	return itemRet;
      
    RET_CACHED_ITEM(index);
    
    /* we can use the accelerator to speed up the location of
     * the item
     */
    FIND_IN_PATHFINDER(index, itemPtr, pos);
      
    TreeTableIndex (itemPtr, index, pos, &itemRet);
   
    if (itemRet)
    {
	SET_CACHED(itemRet, index);
    }
   
    return itemRet;
}
/*}*/

/*{*/
static int
TreeTableFindNotHiddenItem_x (TableItem *itemPtr, int index,
			      int pos, TableItem **itemRet)
{
  TableItem *item = itemPtr;
  int i;
  
  for (i=pos; i<=index && item != NULL; item = item->nextPtr)
  {
     if (i == index)
     {
     	*itemRet = item;
	return -1;
     }
     i++;
     if (item->succPtr != NULL)
     {
	/* don't force hidden sub trees */
	if ((item->flags & ITEM_HIDDEN_SUBTREE) || i+item->succNum < index)
	{
	    i += item->succNum;
	}
	else
        {
	    i = TreeTableFindNotHiddenItem_x (item->succPtr, index, i, itemRet);
	    if (i < 0)
	    {
		return i;
	    }
	}
     }
  }
  
  return i;
}
static TableItem * TreeTableFindNotHiddenItem (TreeTable *tablePtr, int index)
{
    TableItem* itemRet = NULL;
    TableItem* Ptr = tablePtr->itemPtr;
    int pos = 0;
   
    if (index < 0)
	return itemRet;
    
    FIND_IN_PATHFINDER(index, Ptr, pos);
    
    /*RET_CACHED_ITEM(index);*/
      
    TreeTableFindNotHiddenItem_x (Ptr, index, pos, &itemRet);
   
    /*SET_CACHED (itemRet, index);*/
   
    return itemRet;
}
/*}*/

/*
 * function return a boolean value if the item is inserted in a hidden
 * sub tree (not viewed)
 */
static int
TreeTableInsertListItem (register TreeTable* tablePtr,
			 register TableItem*newPtr,
			 register TableItem*prevPtr,
			 register int index,
			 register int *newIndex)
{
    register TableItem* parentPtr = NULL, *posPtr = NULL;
    register TableItem* thisPtr = NULL;
    register int hidden;

    /* add item after given position */
    if (prevPtr != NULL)
    {
	newPtr->parentPtr = prevPtr->parentPtr;
	newPtr->parent    = prevPtr->parent;
	newPtr->indent    = prevPtr->indent;
	newPtr->nextPtr   = prevPtr->nextPtr;
	prevPtr->nextPtr  = newPtr;
    }
    /* find position to add new item */
    else
    {
    /*
     * find parent
     */
    if (newPtr->parent >= 0)
    {
#ifdef USE_PARENT_CACHE
	parentPtr = GET_CACHED_PARENT(tablePtr, newPtr->parent);
#else
	parentPtr = TreeTableFindItem (tablePtr, newPtr->parent);
#endif
    	if (parentPtr != NULL)
	{
	    newPtr->parentPtr = parentPtr;
	    newPtr->indent    = parentPtr->indent + 1;
	    SET_PARENT_CACHE (parentPtr, newPtr->parent);
	}
	else
	{
	    fprintf (stderr, "TreeTableInsertListItem: parent %i not availiable\n",
	    		newPtr->parent);
	}
    }
    
    if (parentPtr != NULL)
    {
    	/* calculate the correct position of the new item */
	*newIndex = newPtr->parent+1;
	
    	/* first item on the list */
	if (parentPtr->succPtr == NULL)
	{
    	   parentPtr->succPtr = newPtr;
	}
	else
	{
	   /* find the previos item in the succ list */
	   posPtr = NULL;
	   if (parentPtr->succPtr != NULL)
	   {
		int i;
		/* if sorted insertion is enabled, search for the correct position
		 * in the children list
		 */
		if (tablePtr->sortedInsertion)
		{
		    posPtr = TreeTableSortFindPrev
		    		(parentPtr->succPtr,
		    			newPtr,
					tablePtr->sortColumn,
					tablePtr->sortNoCase,
					newIndex);
		}
		else
		{
		    GET_CACHED_ITEM(posPtr, index-1);
		    if (posPtr == NULL || posPtr->parentPtr != parentPtr)
	   	    {
			i = index - newPtr->parent - 1;
			thisPtr = parentPtr->succPtr;
			for (; thisPtr != NULL && i>0; thisPtr = thisPtr->nextPtr, i--)
			{
		   	    *newIndex += 1 + thisPtr->succNum;
			    posPtr = thisPtr;
			}
		    }
		    else
		    {
			*newIndex = index;
		    }
		}
	   }
	   
	   /* add item as first element in the list */
	   if (posPtr == NULL)
	   {
	   	newPtr->nextPtr = parentPtr->succPtr;
		parentPtr->succPtr = newPtr;
	   }
	   /* insert item in the middle or at end of list */
	   else
	   {
		newPtr->nextPtr = posPtr->nextPtr;
		posPtr->nextPtr = newPtr;
	   }
	}
    }
    else
    {
	/* first position */
	*newIndex = 0;
	
	/* insert at begin of list */
	if (index <= 0 || tablePtr->itemPtr == NULL)
	{
	    newPtr->nextPtr = tablePtr->itemPtr;
	    tablePtr->itemPtr = newPtr;
	    
	}
	else
	{
	    /* if sorted insertion is enabled, search for the correct position
	     * in the children list
	     */
	    if (tablePtr->sortedInsertion)
	    {
		posPtr = TreeTableSortFindPrev
				(tablePtr->itemPtr,
		    			newPtr,
					tablePtr->sortColumn,
					tablePtr->sortNoCase,
					newIndex);
	    }
	    else
	    {
	        /* find item before the insertion position */
	        posPtr = TreeTableFindItem (tablePtr, index - 1);
		if (posPtr)
		{
		    *newIndex = index;
		}
	    }
	    
	    /*
	     * insert the new item after found position
	     */
	    if (posPtr != (TableItem *) NULL) 
	    {
		/* we must be sure, that the item is added at root
		 * without any parent
		 */
		for (; posPtr->parentPtr != NULL; posPtr=posPtr->parentPtr)
		  ;
    
	        newPtr->parentPtr = posPtr->parentPtr;
		newPtr->indent    = posPtr->indent;
		
	        newPtr->nextPtr = posPtr->nextPtr;
		posPtr->nextPtr = newPtr;

	    }
	    /* insert item on top of list */
	    else
	    {
		newPtr->nextPtr = tablePtr->itemPtr;
		tablePtr->itemPtr = newPtr;
	    }
	}
    }
    }
    
    /* increment item number of parents to quicke searching */
    hidden = 0;
    for (parentPtr=newPtr->parentPtr; parentPtr; parentPtr=parentPtr->parentPtr)
    {
	if (parentPtr->flags&ITEM_HIDDEN_SUBTREE)
	{
	    hidden = 1;
	}
	parentPtr->succNum ++;
	if (!hidden)
	{
	    parentPtr->seenNum ++;
	}
    }
    
    FREE_RONG_CACHE (*newIndex);
    FREE_RONG_PARENT_CACHE(*newIndex);
    SET_CACHED (newPtr, *newIndex);
    
    /* Delete entries from Pathfinder after this position */
    DELETE_FROM_PATHFINDER(*newIndex);
    
    return hidden;
}
/*
 *----------------------------------------------------------------------
 *
 * TreeTableInsertItem --
 *
 *      Add a new item to the tree table.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      New information gets added to tablePtr;  it will be redisplayed
 *      soon, but not immediately.
 *
 *----------------------------------------------------------------------
 */


static TableItem*
TreeTableInsertItem(
     register TreeTable *tablePtr,   /* TreeTable that is to get the new
				      * items. */
     register int index,                      /* Add the new items before this
				      * line */
     register TableItem* prevPtr,
     register int argc,                       /* argument count for argv */
     register char **argv)                    /* argument strings */
{
    register TableItem *newPtr;
    int hidden;
    int oldmaxwidth, newIndex = -1;

    /*
     * Find the item before which the new ones will be inserted.
     */
    if (index <= 0) {
	index = 0;
    }
    if (index > tablePtr->numItems)
    {
	index = tablePtr->numItems;
    }
    
    newPtr = (TableItem *) Tcl_Alloc (sizeof(TableItem));
    /* FIXME: is this really required?  The code used to use calloc(),
       and I'm afraid to change it.  */
    memset (newPtr, 0, sizeof (TableItem));
    if (newPtr==NULL)
    {
	fprintf (stderr, "TreeTableInsertItem: FATAL ERROR: Can't allocate memory\n");
	return NULL;
    }

    /*
     * initialize new item 
     */
    newPtr->lineWidth    = tablePtr->defLineWidth;
    newPtr->parent       = -1;
    
    newPtr->fontHeight = tablePtr->defFontHeight;
    
    /*
     * Call lineconfigure for creation options 
     */
    if (ConfigureTreeTableItem(tablePtr, newPtr, argc,
			       argv,
			       TK_CONFIG_ARGV_ONLY,
			       CONFIG_OPTIONS_ONLY)
	!= TCL_OK)
    {
	ViewArgs ("couldn't add item", argc, argv, 0);
	
	TreeTableRemoveItem (tablePtr, newPtr, -1, NULL, 0, 0);
	return NULL;
    }
#if 0
    if (newPtr->text == NULL)
    {
    	ViewArgs ("FATAL ERROR: item->text == NULL", argc, argv);
	TreeTableRemoveItem (tablePtr, newPtr, -1, NULL, 0, 0);
	return NULL;
    }
#endif
    
    /*
     * Khamis: 26-03-97
     * Add item to the list structure
     */
    hidden = TreeTableInsertListItem (tablePtr, newPtr, prevPtr, index, &newIndex);
    InsertedNewPosition = newIndex;
    
    /* 
     * Update item count, adjust selection.
     */
    tablePtr->numItems++;
    
    /* correct anchor and active */
    if (tablePtr->selectAnchor >= index)
    {
	tablePtr->selectAnchor ++;
    }
    if (tablePtr->active >= index)
    {
	tablePtr->active ++;
    }
    
    /* recompute width and height if item is viewed and not in the
     * hidden sub tree
     */
    if (!hidden)
    {
	TreeTableComputeLineHeight(tablePtr, newPtr, 1);
	tablePtr->flags |= UPDATE_V_SCROLLBAR;

	oldmaxwidth = tablePtr->maxWidth;
	TreeTableComputeWidths(tablePtr, newPtr, 0);
	if (oldmaxwidth != tablePtr->maxWidth)
	{
	    tablePtr->flags |= UPDATE_H_SCROLLBAR;
	}
	
	TreeTableRedrawRange(tablePtr);
    }
    if (tablePtr->BestFit)
    {
	ComputeTabStops (tablePtr, newPtr);
    }
    return newPtr;
}

/*
 * add many items on the same position with same options **
 */
static int
TreeTableInsertItems (TreeTable* tablePtr,
		      int index, char* itemlist,
		      int argc, char **argv)
{
    TableItem*newPtr;
    char *p, *item;
    int i, oldmaxwidth, result, size;
    int nargc = argc + 2, targpos = nargc - 1;
    char **nargv = malloc (sizeof(char**)*nargc);
    int ret = TCL_OK;
    char buf[1024], *itembuf = buf;
    int bufsize = 1024;
    int length;
    
    for (i=0;i<argc; i++)
    {
    	nargv[i] = argv[i];
    }
    nargv[i++] = "-text";
    
    for (i = 0, newPtr=NULL, length=strlen(itemlist), p = itemlist; *p != 0; i++)
    {
    	char *prevlist = p;
	result = TclFindElement(tablePtr->interp, p, length, &item, &p, &size, NULL);
	if (result != TCL_OK)
	{
	    break;
	}
	length -= p - prevlist;
	if (size == 0)
	{
	    continue; /* don't add empty strings (core dump) */
	}
	
	if (size>bufsize)
	{
	    bufsize += size;
	    if (itembuf == buf)
	    {
	    	itembuf = malloc (size);
	    }
	    else
	    {
		itembuf = realloc (itembuf, size);
	    }
	}
	memcpy (itembuf, item, size);
	itembuf[size] = 0;
	nargv[targpos] = itembuf;
	
	newPtr = TreeTableInsertItem(tablePtr, index, newPtr, nargc, nargv);
	if (newPtr == NULL)
	{
	    ret = TCL_ERROR;
	    break;
	}
	if (!tablePtr->tabs)
	{
	    oldmaxwidth = tablePtr->maxWidth;
	    TreeTableComputeWidths(tablePtr, newPtr, 0);
	    if (oldmaxwidth != tablePtr->maxWidth)
	    {
		tablePtr->flags |= UPDATE_H_SCROLLBAR;
	    }
	}
    }
    if (itembuf != buf)
    {
    	free(itembuf);
    }
    free (nargv);
    return ret;
}

/*
 * Delete only the line indicated, which requires us to remove all
 * references to this node as a parent.
 */
static void
TreeTableDeleteRange(tablePtr, start, end) 
    TreeTable *tablePtr;    /* table in which target item resides */
    int start,end;          /* range of items to delete */ 
{
    TableItem *posPtr,*Ptr;
    int i;
    int oldmaxwidth, width = 0, children = 0;

    /* First check to make sure there any items to delete.  If there
     * are not, simply return TCL_OK */
    if (!tablePtr->numItems || start > end)
    {
	return;
    }
    
    /* delete all */
    if (start <= 0 && end >= tablePtr->numItems-1)
    {
	int j;
	children = 1;
	i = j = 0;
	for (posPtr=tablePtr->itemPtr; posPtr != NULL;)
	{
	    Ptr=posPtr->nextPtr;
	    j = i + 1 + posPtr->succNum;
	    TreeTableRemoveItem (tablePtr, posPtr, i, &width, 0, children);
	    i = j;
	    posPtr=Ptr;
	}
	FREE_RONG_CACHE(0);
 	FREE_RONG_PARENT_CACHE(0);
	tablePtr->itemPtr = NULL;
    }
    else for (i=end; i>=start; i--)
    {
    	posPtr = TreeTableFindItem(tablePtr,i);
	if (posPtr == NULL)
	{
	    continue;
	}
	TreeTableRemoveItem (tablePtr, posPtr, i, &width, 0, children);
    }
    
    /* if BestFit is enabled and no items are availiable
     * then set the tab stops to the default value
     */
    if (tablePtr->BestFit && tablePtr->numItems == 0)
    {
	for (i=0; i<tablePtr->tabsNum; i++)
	{
	    if (tablePtr->tabsHidden[i])
	    {
		tablePtr->tabs[i] = tablePtr->tabsMinSpace;
	    }
	    else
	    {
		tablePtr->tabs[i] = tablePtr->defTabs[i];
	    }
	}
    }
    
    /* if width is changed */
    if (width >= tablePtr->maxWidth)
    {
	oldmaxwidth = tablePtr->maxWidth;
	TreeTableComputeWidths (tablePtr, (TableItem *) NULL, 1);
	if (oldmaxwidth != tablePtr->maxWidth)
	{
	    tablePtr->flags |= UPDATE_H_SCROLLBAR;
	}
    }
    TreeTableComputeLineHeight (tablePtr, NULL, 0);
    tablePtr->flags |= UPDATE_V_SCROLLBAR;
    
    /* correct top index */
    if (tablePtr->topIndex > tablePtr->numViewed - tablePtr->numLines)
	tablePtr->topIndex = Max (0, tablePtr->numViewed - tablePtr->numLines);

    TreeTableRedrawRange(tablePtr);
}

/*
 * Delete the indicated item and all of its descendants.
 *
 * return count of deleted items
 */
static int
TreeTableRemoveItem(
     TreeTable *tablePtr,        /* table in which the target item resides */
     TableItem *itemPtr,         /* item to delete */
     int lineNum,
     int* lineWidth,
     int original_item,
     int children                /* only true if this is the original item
				  * marked for deletion.  Children of the
				  * target item will be deleted, but this
				  * flag will not be set for them. */
     )
{
    TableItem *Ptr, *tmpPtr, *prevPtr, *topPtr;
    int i, index;
    int deleted = 1, hidden, width;

    /*
     * go through and mark references to parents after us as being
     * one line less, and delete any children we have.  Have to use
     * indexes instead of following the linked list because the 
     * list itself will be modified as we travel through the loop.
     */
    if (itemPtr == (TableItem *) NULL)
    {
	fprintf(stderr, "TreeTableRemoveItem: Attempt to delete NULL item.\n");
	return 0;
    }
    
    /* correct selection range */
    if (lineNum == -1)
    {
	index = TreeTableFindIndex(tablePtr, itemPtr);
    }
    else
    {
	index = lineNum;
    }
    
    /*
     * no such node in tree anymore -- should not happen
     */
    if (index == -1)
    {
	fprintf(stderr, "Fatal Error: TreeTableRemoveItem: could not find index.\n");
    	fprintf(stderr, "line: %s, image: %s\n",
			itemPtr->text ? itemPtr->text : "NULL",
			itemPtr->imageString ? itemPtr->imageString : "NULL");
	TreeTableFreeItem(tablePtr, itemPtr);
	return 0;
    }
    if (tablePtr->active > index)
    {
      tablePtr->active --;
    }
    
    /* delete sub tree of this item */
    for (tmpPtr=itemPtr->succPtr, i=index+1; tmpPtr != NULL; )
    {
	int r;
	Ptr = tmpPtr->nextPtr;
    	r = TreeTableRemoveItem (tablePtr, tmpPtr, i, lineWidth, 0, 1);
	i   += r;
	deleted += r;
	tmpPtr = Ptr;
    }

    /* decrement succ number of parents */
    hidden = 0;
    for (Ptr=itemPtr->parentPtr; Ptr != NULL; Ptr=Ptr->parentPtr)
    {
	Ptr->succNum --;
	if (Ptr->flags & ITEM_HIDDEN_SUBTREE)
	{
	    hidden = 1;
	}
	if (!hidden)
	{
	    Ptr->seenNum --;
	}
    }

    if (!children)
    {
	FREE_RONG_CACHE(index);
 	FREE_RONG_PARENT_CACHE(index);
   
	/*
	 * Now remove this item from the list of items.
	 */
	if (itemPtr->parentPtr != NULL)
	{
	    topPtr = itemPtr->parentPtr->succPtr;
	}
	else
	{
	    topPtr = tablePtr->itemPtr;
	}
	if (topPtr != NULL)
	{
	    /* first item in the list */
	    if (topPtr == itemPtr)
	    {
		if (itemPtr->parentPtr != NULL)
		{
		    itemPtr->parentPtr->succPtr = itemPtr->nextPtr;
		}
		else
		{
		    tablePtr->itemPtr = itemPtr->nextPtr;
		}
	    }
	    else
	    {
		/* concat the tow parts */
		for (prevPtr=topPtr; prevPtr != NULL; prevPtr=prevPtr->nextPtr)
		{
		    if (prevPtr->nextPtr != NULL && prevPtr->nextPtr == itemPtr)
		    {
			prevPtr->nextPtr = itemPtr->nextPtr;
			break;
		    }
		}
		if (prevPtr == NULL)
		{
		    fprintf(stderr, "Fatal Error: TreeTableRemoveItem: previos list element not found");
		    exit (1);
		}
	    }
	}
	else
	{
	    fprintf(stderr, "fatal error: TreeTableRemoveItem: couldn't find root to delete item (%s)\n",
		    itemPtr->text);
	}
    }
    
    width = itemPtr->lineWidth;
    if (lineWidth && *lineWidth < width)
    {
	*lineWidth = width;
    }
    
    /* decrement number of items */
    tablePtr->numItems --;
    
    /* decrement number of selected items, if item is selected */
    if (itemPtr->flags&ITEM_SELECTED)
    {
	tablePtr->numSelected --;
    }
    
    /* Delete the entries of the Pathfinder after this index */
    DELETE_FROM_PATHFINDER(index);
    
    /* now we can destroy item */
    TreeTableFreeItem(tablePtr, itemPtr);
    
    /* correct anchor and active */
    if (tablePtr->selectAnchor > index)
    {
	tablePtr->selectAnchor --;
    }
    if (tablePtr->active > index)
    {
	tablePtr->active --;
    }
    
    if (original_item && !hidden)
    {
	int oldmaxwidth;
	if (width >= tablePtr->maxWidth)
	{
	    oldmaxwidth = tablePtr->maxWidth;
	    TreeTableComputeWidths(tablePtr, (TableItem *) NULL, 1);
	    if (oldmaxwidth != tablePtr->maxWidth) {
	        tablePtr->flags |= UPDATE_H_SCROLLBAR;
	    }
	}
	TreeTableComputeLineHeight(tablePtr, NULL, - deleted);
	tablePtr->flags |= UPDATE_V_SCROLLBAR;
	TreeTableRedrawRange(tablePtr);
    }
    
    return deleted;
}


/*
 *----------------------------------------------------------------------
 *
 * TreeTableFreeItem --
 *
 *      Free up the resources used by the item, then free the item itself.
 *      Note that this does nothing more than free the structure -- it 
 *      does not adjust the item list or anything else.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Memory gets freed.
 *
 *----------------------------------------------------------------------
 */

static void
TreeTableFreeItem(tablePtr, itemPtr)
     register TreeTable *tablePtr;  /* Treetable widget to modify. */
     TableItem *itemPtr;            /* Item to free */
{
    Tk_FreeOptions(itemConfigSpecs, (char *) itemPtr, tablePtr->display, 0);
    if (itemPtr->textGC != None) {
	Tk_FreeGC(tablePtr->display, itemPtr->textGC);
    }
    if (itemPtr->bitmapGC != None) {
	Tk_FreeGC(tablePtr->display, itemPtr->bitmapGC);
    }
#ifdef USE_BITMAP_COLOR
    if (itemPtr->bitmapSelectGC != None) {
	Tk_FreeGC(tablePtr->display, itemPtr->bitmapSelectGC);
    }
    if (itemPtr->lineGC != None) {
	Tk_FreeGC(tablePtr->display, itemPtr->lineGC);
    }
#endif
    if (itemPtr->bitmap != None) {
	Tk_FreeBitmap(tablePtr->display, itemPtr->bitmap);
    }
    /* free image from item */
    if (itemPtr->image != NULL) {
	Tk_FreeImage(itemPtr->image);
    }
    
    ckfree((char *) itemPtr);
}

static int
TreeTableToggle (TreeTable* tablePtr, int xx, int yy, int testonly)
{
    int ret;
    int pos = 0, rpos = 0;
    
    if (xx < 0 || yy < 0)
    {
       return ITEM_TEXT;
    }
    
    /* try to toggle or test if we can toggle */
    ret = TreeTableToggleSubTree (tablePtr, tablePtr->itemPtr, &rpos, &pos, xx, yy, testonly);
    
    /* index not found */
    if (ret >= 0)
        return ITEM_NOT_FOUND;
    
    /* bitmap clicked or text selection*/
    return ret;
}

static void
TreeTableToggleIt (TreeTable* tablePtr, TableItem*Ptr)
{
    TableItem *pPtr;
    int j, seen = 0, hidden = 0;
    if (Ptr->flags&ITEM_HIDDEN_SUBTREE)
    {
	Ptr->flags &= ~ITEM_HIDDEN_SUBTREE;
	
	/* recalculate seen items in sub tree and in parents too */
	Ptr->seenNum = TreeTableCountNotHidden_x (Ptr->succPtr);
	for (pPtr=Ptr->parentPtr; pPtr != NULL; pPtr=pPtr->parentPtr)
	{
	    if (pPtr->flags&ITEM_HIDDEN_SUBTREE)
	    {
		hidden = 1;
		break;
	    }
	    pPtr->seenNum += Ptr->seenNum;
	}
	seen = Ptr->seenNum;
    }
    else
    {
	Ptr->flags |= ITEM_HIDDEN_SUBTREE;
	
	/* no seen items in it's sub tree and in parents too */
	for (pPtr=Ptr->parentPtr; pPtr != NULL; pPtr=pPtr->parentPtr)
	{
	    if (pPtr->flags&ITEM_HIDDEN_SUBTREE)
	    {
		hidden = 1;
		break;
	    }
	    pPtr->seenNum -= Ptr->seenNum;
	}
	seen = - Ptr->seenNum;
	Ptr->seenNum = 0;
    }
     
    /* lost selected items in the sub tree */
    j = 0;
    TreeTableLostSel_x (tablePtr, Ptr->succPtr, &j, 0, tablePtr->numItems-1);
     
    /* recompute vertical */
#if 1
    if (hidden)
    {
    	return;
    }
    tablePtr->numViewed += seen;
#else
    TreeTableCountNotHidden (tablePtr, 1);
#endif
    
    /* recompute horizontal */
    TreeTableComputeWidths  (tablePtr, NULL, 1);
    
    tablePtr->flags |= UPDATE_V_SCROLLBAR|UPDATE_H_SCROLLBAR;
    TreeTableRedrawRange (tablePtr);
}

static int
TreeTableToggleSubTree (TreeTable* tablePtr, TableItem* itemPtr,
			int* rpos, int* pos, int xx, int yy, int testonly)
{
    TableItem *Ptr;
    
    for (Ptr=itemPtr; Ptr != NULL; Ptr=Ptr->nextPtr)
    {
	int x = ITEM_X(Ptr);
	int y = ITEM_Y(Ptr, *pos);

	if (yy >= y && yy <= y+tablePtr->lineHeight)
	{
	   /*
	    * the icon is clicked to toggle viewing a subtree,
	    */
	   if (xx >= x-2 && xx <= x+Ptr->bitmapWidth+2)
	   {
		if (Ptr->succPtr != NULL)
		{
		    if (testonly)
		    {
			if (Ptr->flags&ITEM_HIDDEN_SUBTREE)
			    return ITEM_HIDDEN;
		        else
			    return ITEM_VIEWED;
		    }
	     
	     	    TreeTableToggleIt (tablePtr, Ptr);
		    
		    /* we don't need to force forward */
		    if (Ptr->flags&ITEM_HIDDEN_SUBTREE)
		    {
			return ITEM_HIDDEN;
		    }
		    else
		    {
			return ITEM_VIEWED;
		    }
		}
		/* item bitmap without sub tree is clicked */ 
		else
		{
		    return ITEM_UNKNOWN;
		}
	   }
	   /* text is clicked */
	   else if (xx > x+Ptr->bitmapWidth)
	   {
		return ITEM_TEXT;
	   }
	   else
	   {
		return ITEM_NOT_FOUND;
	   }
	}

	/* incremente realy and virtual position */
	*pos += 1;
	*rpos += 1;
	
	/* call the function with the sub tree, if availiable */
	if (!(Ptr->flags&ITEM_HIDDEN_SUBTREE) && Ptr->succPtr != NULL)
	{
	    y = (((*pos+Ptr->seenNum) - tablePtr->topIndex) * tablePtr->lineHeight) +
		tablePtr->borderWidth;
	    if (y < yy)
	    {
	    	*pos += Ptr->seenNum;
		*rpos += Ptr->succNum;
	    }
	    else
	    {
		int i = TreeTableToggleSubTree (tablePtr, Ptr->succPtr, rpos, pos, xx, yy, testonly);
		if (i < 0)
		    return i;
	    }
	}
	else
	{
	    *rpos += Ptr->succNum;
	}
    }
    return *rpos;
}

static int
TreeTableToggleItem (TreeTable* tablePtr, TableItem *itemPtr, int hide)
{
    int ret = TCL_OK;
    
    /* have no children */
    if (itemPtr->succNum == 0)
    {
    	return ret;
    }
    
    /* already hidden or viewed */
    if (((itemPtr->flags&ITEM_HIDDEN_SUBTREE) && hide) ||
        (!(itemPtr->flags&ITEM_HIDDEN_SUBTREE) && !hide))
    {
	return ret;
    }
    TreeTableToggleIt (tablePtr, itemPtr);
    return ret;
}

/*
 *--------------------------------------------------------------
 *
 * TreeTableEventProc --
 *
 *      This procedure is invoked by the Tk dispatcher for various
 *      events on treetables.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      When the window gets deleted, internal structures get
 *      cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
TreeTableEventProc(clientData, eventPtr)
     ClientData clientData;      /* Information about window. */
     XEvent *eventPtr;           /* Information about event. */
{
    TreeTable *tablePtr = (TreeTable *) clientData;

    if (eventPtr->type == Expose) {
	TreeTableRedrawRange(tablePtr);
    } else if (eventPtr->type == DestroyNotify) {
	/* Zsolt Koppany, 3-jan-97 */
	if (tablePtr->tkwin != NULL)
	{
		if (tablePtr->setGrid) {
			Tk_UnsetGrid(tablePtr->tkwin);
		}
		tablePtr->tkwin = NULL;
		Tcl_DeleteCommand(tablePtr->interp,
			Tcl_GetCommandName(tablePtr->interp,tablePtr->widgetCmd));
	}

        if (tablePtr->flags & REDRAW_PENDING) {
            Tcl_CancelIdleCall(DisplayTreeTable, (ClientData) tablePtr);
        }
        Tcl_EventuallyFree((ClientData) tablePtr, DestroyTreeTable);
    } else if (eventPtr->type == ConfigureNotify) {
        Tcl_Preserve((ClientData) tablePtr);
	
        tablePtr->numLines = TREETABLE_NUM_LINES(tablePtr);
			   
        tablePtr->flags |= UPDATE_V_SCROLLBAR|UPDATE_H_SCROLLBAR;
        TreeTableRedrawRange(tablePtr);
        Tcl_Release((ClientData) tablePtr);
    } else if (eventPtr->type == FocusIn) {
      if (eventPtr->xfocus.detail != NotifyInferior) {
	tablePtr->flags |= GOT_FOCUS;
	TreeTableRedrawRange(tablePtr);
      }
    } else if (eventPtr->type == FocusOut) {
      if (eventPtr->xfocus.detail != NotifyInferior) {
	tablePtr->flags &= ~GOT_FOCUS;
	TreeTableRedrawRange(tablePtr);
      }
    }
}

/*
 *--------------------------------------------------------------
 *
 * GetTreeTableIndex --
 *
 *      Parse an index into a treetable and return either its value
 *      or an error.
 *
 * Results:
 *      A standard Tcl result.  If all went well, then *indexPtr is
 *      filled in with the index corresponding to
 *      string.  Otherwise an error message is left in interp->result.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static int
GetTreeTableIndex(tablePtr, string, endAfter, indexPtr)
     TreeTable *tablePtr;         /* TreeTable for which the index is being
				   * specified. */
     char *string;               /* Numerical index into tablePtr's item
				  * list, or "end" to refer to last item. */
     int endAfter;               /* 0 means "end" refers to the index of the
				  * last item, 1 means it refers to the
				  * item after the last one. */
     int *indexPtr;              /* Where to store converted index. */
{
    int length = strlen (string);
    if (string[0] == 'e' && strncmp(string, "end", length) == 0)
    {
            *indexPtr = tablePtr->numItems;
    }
    else if (string[0] == 'a' && strncmp(string, "active", length) == 0)
    {
            *indexPtr = tablePtr->active;
    }
    else if (string[0] == '@')
    {
	    int x, y;
	    char *p, *end;

	    p = string+1;
	    x = strtol(p, &end, 0);
	    if ((end == p) || (*end != ',')) {
	        goto badIndex;
	    }
	    p = end+1;
	    y = strtol(p, &end, 0);
	    if ((end == p) || (*end != 0)) {
	        goto badIndex;
	    }
	    *indexPtr = NearestTreeTableItem(tablePtr, y);
    }
    else if ((string[0] == 'a') && (strncmp(string, "anchor", length) == 0) && (length >= 2))
    {
	    *indexPtr = tablePtr->selectAnchor;
    }
    else
    {
        if (Tcl_GetInt(tablePtr->interp, string, indexPtr) != TCL_OK)
	{
            Tcl_ResetResult(tablePtr->interp);
            goto badIndex;
        }
    }
    
    if (endAfter)
    {
	if (*indexPtr > tablePtr->numItems)
	{
	     *indexPtr = tablePtr->numItems;
	}
    }
    else if (*indexPtr >= tablePtr->numItems)
    {
	*indexPtr = tablePtr->numItems-1;
    }
    if (tablePtr->numItems <= 0)
    {
	*indexPtr = 0;
    }
    else if (*indexPtr < 0)
    {
	*indexPtr = 0;
    }
    return TCL_OK;
 
    badIndex:
    Tcl_AppendResult(tablePtr->interp, "bad treetable index \"", string,
			     "\"", (char *) NULL);
    return TCL_ERROR;
}


/*
 *----------------------------------------------------------------------
 *
 * ChangeTreeTableView --
 *
 *      Change the view on a treetable widget.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      What's displayed on the screen is changed.  If there is a
 *      scrollbar associated with this widget, then the scrollbar
 *      is instructed to change its display too.
 *
 *----------------------------------------------------------------------
 */

static void
ChangeTreeTableView(register TreeTable *tablePtr,         /* Information about widget. */
    int index,                            /* Index of item in tablePtr */
    int redraw)
{
    int count;

    count = TreeTableCountNotHidden (tablePtr, 0);

    if (index >= (count - tablePtr->numLines))
    {
        index = count - tablePtr->numLines;
    }

    if (index < 0) {
        index = 0;
    }
    
    if (tablePtr->topIndex != index || redraw)
    {
        tablePtr->topIndex = index;
        if (!(tablePtr->flags & REDRAW_PENDING))
	{
            Tk_DoWhenIdle(DisplayTreeTable, (ClientData) tablePtr);
            tablePtr->flags |= REDRAW_PENDING;
        }
        tablePtr->flags |= UPDATE_V_SCROLLBAR;
    }
}



/*
 *----------------------------------------------------------------------
 *
 * ChangeTreeTableOffset --
 *
 *      Change the horizontal offset for a treetable.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The treetable may be redrawn to reflect its new horizontal
 *      offset.
 *
 *----------------------------------------------------------------------
 */

static void
ChangeTreeTableOffset(tablePtr, offset)
    register TreeTable *tablePtr;       /* Information about widget. */
    int offset;                         /* Desired new "xOffset" for
                                         * treetable. */
{
    int maxOffset;

    /*
     * Make sure that the new offset is within the allowable range, and
     * round it off to an even multiple of xScrollUnit.
     */

    maxOffset = tablePtr->maxWidth
    		+ (tablePtr->xScrollUnit-1)
            	- (Tk_Width(tablePtr->tkwin)
		   - 2 * (tablePtr->selBorderWidth+tablePtr->inset)
		   - tablePtr->xScrollUnit);
    if (offset > maxOffset) {
        offset = maxOffset;
    }
    if (offset < 0) {
        offset = 0;
    }
    offset -= offset%tablePtr->xScrollUnit;
    if (offset != tablePtr->xOffset)
    {
        tablePtr->xOffset = offset;
        tablePtr->flags |= UPDATE_H_SCROLLBAR;
        TreeTableRedrawRange(tablePtr);
    }
}


/*
 *----------------------------------------------------------------------
 *
 * TreeTableScanTo --
 *
 *      Given a point (presumably of the curent mouse location)
 *      drag the view in the window to implement the scan operation.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The view in the window may change.
 *
 *----------------------------------------------------------------------
 */

static void
TreeTableScanTo(tablePtr, x, y)
    register TreeTable *tablePtr;       /* Information about widget. */
    int x;                              /* X-coordinate to use for scan
                                         * operation. */
    int y;                              /* Y-coordinate to use for scan
                                         * operation. */
{
    int newTopIndex, newOffset;

    /*
     * Compute new top line for screen by amplifying the difference
     * between the current position and the place where the scan
     * started (the "mark" position).  If we run off the top or bottom
     * of the list, then reset the mark point so that the current
     * position continues to correspond to the edge of the window.
     * This means that the picture will start dragging as soon as the
     * mouse reverses direction (without this reset, might have to slide
     * mouse a long ways back before the picture starts moving again).
     */
    int count = TreeTableCountNotHidden (tablePtr, 0);
    
    newTopIndex = tablePtr->scanMarkYIndex
            - (10*(y - tablePtr->scanMarkY))/tablePtr->lineHeight;
    if (newTopIndex >= count)
    {
        newTopIndex = tablePtr->scanMarkYIndex = count-1;
        tablePtr->scanMarkY = y;
    } else if (newTopIndex < 0)
    {
        newTopIndex = tablePtr->scanMarkYIndex = 0;
        tablePtr->scanMarkY = y;
    }
    ChangeTreeTableView(tablePtr, newTopIndex, 0);

    /*
     * Compute new left edge for display in a similar fashion by amplifying
     * the difference between the current position and the place where the
     * scan started.
     */

    newOffset = tablePtr->scanMarkXOffset - (10*(x - tablePtr->scanMarkX));
    if (newOffset >= tablePtr->maxWidth) {
        newOffset = tablePtr->scanMarkXOffset = tablePtr->maxWidth;
        tablePtr->scanMarkX = x;
    } else if (newOffset < 0) {
        newOffset = tablePtr->scanMarkXOffset = 0;
        tablePtr->scanMarkX = x;
    }
    ChangeTreeTableOffset(tablePtr, newOffset);
}

/*{*/
/*
 * Caluclate the realy index of a view index
 */
static int
TreeTableRealyIndex_x (register TableItem *itemPtr, int*pos, int* rindex, int index)
{
    TableItem*Ptr;
    for (Ptr=itemPtr; *pos <= index && Ptr; Ptr=Ptr->nextPtr)
    {
    	/* found realy position */
	if (*pos == index)
	{
	    return -1;
	}
	*pos += 1;
	*rindex += 1;
	if (!(Ptr->flags&ITEM_HIDDEN_SUBTREE) && Ptr->succPtr != NULL)
	{
	    int ret;
	    
	    if (*pos+Ptr->seenNum < index)
	    {
		*pos += Ptr->seenNum;
		*rindex += Ptr->succNum;
	    }
	    else
	    {
		ret = TreeTableRealyIndex_x (Ptr->succPtr, pos, rindex, index);
		if (ret < 0)
		    return ret;
	    }
	}
	else
	{
	    *rindex += Ptr->succNum;
	}
    }
    return *rindex;
}

static int
TreeTableRealyIndex (register TreeTable *tablePtr, int index)
{
    TableItem*Ptr=tablePtr->itemPtr;
    int rindex, pos, ret;
    
    /* there is nothing hidden, so the realy index is eq. viewed index */
    if (tablePtr->numViewed == tablePtr->numItems)
    {
	return index;
    }
    
    FIND_IN_PATHFINDER(index, Ptr, pos);
    
    /* calculate the realy index */
    rindex = pos = 0;
    ret = TreeTableRealyIndex_x(Ptr, &pos, &rindex, index);
    
    if (ret < 0)
       return rindex;
       
    return -1;
}
/*}*/

/*{*/
/*
 * calculate the viewed index of a realy index
 */
static int
TreeTableViewedIndex_x (register TableItem *itemPtr, int*pos, int* vindex, int index)
{
    TableItem*Ptr;
    for (Ptr=itemPtr; *pos <= index && Ptr; Ptr=Ptr->nextPtr)
    {
    	/* found viewed position */
	if (*pos == index)
	{
	    return -1;
	}
	*pos += 1;
	*vindex += 1;
	if (Ptr->succPtr != NULL)
	{
	    if (!(Ptr->flags&ITEM_HIDDEN_SUBTREE))
	    {
		int ret;
	    
		if (*pos+Ptr->succNum < index)
		{
		    *pos += Ptr->succNum;
		    *vindex += Ptr->seenNum;
		}
		else
		{
		    ret = TreeTableViewedIndex_x (Ptr->succPtr, pos, vindex, index);
		    if (ret < 0)
			return ret;
		}
	    }
	    /* if item is in a hidden sub tree, then return his first viewed parent */
	    else if (*pos+Ptr->succNum > index - 1)
	    {
	    	*vindex -= 1;
		return -1;
	    }
	    else
	    {
		*pos += Ptr->succNum;
	    }
	}
    }
    return *vindex;
}
static int
TreeTableViewedIndex (register TreeTable *tablePtr, int index)
{
    int vindex, pos, ret;
    
    /* there is nothing hidden */
    if (tablePtr->numViewed == tablePtr->numItems)
    {
	return index;
    }
    
    /* calculate the viewed index of a realy index */
    vindex = pos = 0;
    ret = TreeTableViewedIndex_x(tablePtr->itemPtr, &pos, &vindex, index);

    if (ret < 0)
       return vindex;
       
    return -1;
}
/*}*/

/*
 *----------------------------------------------------------------------
 *
 * NearestTreeTableItem --
 *
 *      Given a y-coordinate inside a treetable, compute the index of
 *      the item under that y-coordinate (or closest to that
 *      y-coordinate).
 *
 * Results:
 *      The return value is an index of an element of itemPtr.  If
 *      itemPtr has no elements, then 0 is always returned.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int
NearestTreeTableItem(tablePtr, y)
    register TreeTable *tablePtr;      /* Information about widget. */
    int y;                             /* Y-coordinate in tablePtr's window. */
{
    int index;
    int count;
    
    count = TreeTableCountNotHidden (tablePtr, 0);
    
    if (y < 0) y = 0;
    index = (y - tablePtr->borderWidth)/tablePtr->lineHeight;
    
    /* correct the index in usable range */
    if (index < 0)
    {
        index = 0;
    }
    if (index > tablePtr->numLines)
    {
	index = tablePtr->numLines - 1;
    }
    index += tablePtr->topIndex;
    if (index >= count)
    {
        index = count - 1;
    }

    /* return realy index in the tree */
    return TreeTableRealyIndex (tablePtr, index);
}


/*
 *  TreeTableComputeLineHeight
 *
 *          Cycles through all of the items in the table and computes
 *          the maximum line height.
 *
 *
 */
static int
TreeTableComputeLineHeight_x (TableItem* itemPPtr, TableItem* thisPtr, int maxHeight)
{
	TableItem* itemPtr = itemPPtr;
	int height = maxHeight;
	
	while (itemPtr != (TableItem *) NULL)
	{
	    if (itemPtr->fontHeight > height)
	    {
		height = itemPtr->fontHeight;
	    }
	    if (itemPtr->bitmapHeight > height)
	    {
		height = itemPtr->bitmapHeight;
	    }
	
	    /* force only the first item */
	    if (thisPtr != NULL)
		break;
		
	    if (itemPtr->succPtr != NULL)
	    {
		height =
			TreeTableComputeLineHeight_x
				(itemPtr->succPtr, thisPtr, height);
	    }
	    itemPtr = itemPtr->nextPtr;
	}
	return height;
}
static void
TreeTableComputeLineHeight(TreeTable *tablePtr, TableItem* thisPtr, int cnt)
{
    register TableItem *itemPtr;
    register int maxHeight;


    if (cnt == 0)
    {
	if (thisPtr != NULL)
	{
	    itemPtr = thisPtr;
	}
	else
	{
	    itemPtr = tablePtr->itemPtr;
	}
	maxHeight =
		TreeTableComputeLineHeight_x
			(itemPtr, thisPtr, tablePtr->defFontHeight);
    	tablePtr->lineHeight = maxHeight + 1 + 2*tablePtr->selBorderWidth;
	TreeTableCountNotHidden(tablePtr, 1);
    }
    else
    {
    	tablePtr->numViewed += cnt;
    }
}


static void
TreeTableComputeWidths_x (TreeTable* tablePtr, TableItem* itemPtr,
			  TableItem* thisPtr, int checkall)
{
    TableItem* travPtr;
    
    if (thisPtr != NULL)
    {
	travPtr = thisPtr;
    }
    else
    {
	travPtr = itemPtr;
    }
    
    for (; travPtr != (TableItem *) NULL;
	travPtr = travPtr->nextPtr)
    {
	if (tablePtr->BestFit && tablePtr->tabs != NULL)
	{
	    ComputeTabStops (tablePtr, travPtr);
        }
	else
	{
	    /* find max. width only */
	    if (thisPtr != NULL || checkall == 2)
		travPtr->lineWidth =
			Tk_TextWidth (travPtr->fontPtr != NULL ? travPtr->fontPtr : tablePtr->defFontPtr,
			     travPtr->text, 
			     travPtr->textLength)
			+ (travPtr->indent*tablePtr->indentWidth)
			+ travPtr->bitmapWidth 
			+ tablePtr->bitmapSpace
			+ tablePtr->inset
			- tablePtr->xOffset;
	    if (travPtr->lineWidth > tablePtr->maxWidth)
	    {
		tablePtr->maxWidth = travPtr->lineWidth;
	    }
	}
	
	/*
	 * calulate the max. width relative to only one given item
	 */
	if (thisPtr != NULL)
	{
	    return;
	}
	
	if (!(travPtr->flags&ITEM_HIDDEN_SUBTREE) && travPtr->succPtr != NULL)
	{
	    TreeTableComputeWidths_x (tablePtr, travPtr->succPtr, NULL, checkall);
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TreeTableComputeWidths --
 *
 *      This procedure is invoked to recompute the width of either 
 *      a single item (if itemPtr != NULL), or the entire table
 *      (fontChanged = 1).  
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      tablePtr->maxWidth may be modified 
 *
 *----------------------------------------------------------------------
 */
static void
TreeTableComputeWidths(tablePtr, itemPtr, checkAll)
     TreeTable *tablePtr;        /* TreeTable whos geometry is to be
				  * recomputed. */
     TableItem *itemPtr;         /* Item in the table which has changed
				  * or NULL if none has. */
     int checkAll;               /* Non-zero means to check the entire
				  * list of table items in order to 
				  * recompute the max width.
				  *
				  * 0: compute only the width only for
				  *    given 'itemPtr'
				  * 1: find maximum width of lines
				  * 2: recompute line width
				  * 3: If BestFit enabled, recompute the
				  *    best fit for tab stops
				  */
{
    tablePtr->xScrollUnit = Tk_TextWidth (tablePtr->defFontPtr, "0", 1);
    
    if (checkAll == 3 && tablePtr->BestFit && tablePtr->tabs != NULL)
    {
	int i;
	for (i=0; i<tablePtr->tabsNum; i++)
	{
	    if (tablePtr->tabsHidden[i])
	    {
		tablePtr->tabs[i] = tablePtr->tabsMinSpace;
	    }
	    else
	    {
		tablePtr->tabs[i] = tablePtr->defTabs[i];
	    }
        }
    }
    /* we don't need to compute width if tab stops availiable */
    else if (tablePtr->tabs != NULL)
    {
	return;
    }
    
    if (itemPtr == (TableItem *) NULL)
    {
	tablePtr->maxWidth = 10;
    }
    
    TreeTableComputeWidths_x (tablePtr, tablePtr->itemPtr, itemPtr, checkAll);
}


static int
TreeTableSelectFromTo (
	TreeTable*tablePtr,
	TableItem *itemPtr,
	int pos,
	int start,
	int end)
{
    TableItem*Ptr=itemPtr;
    int i = pos;
    
    /* Accelerate the location of the fist item 
     */
    FIND_IN_PATHFINDER (start, Ptr, i);
    
    for (; i <= end && Ptr; Ptr=Ptr->nextPtr)
    {
	if (i >= start && ! (Ptr->flags&ITEM_SELECTED))
	{
	    Ptr->flags |= ITEM_SELECTED;
	    
	    if (tablePtr->selectFirst > i)
	    {
		tablePtr->selectFirst = i;
	    }
	    if (tablePtr->numSelected == 0)
	    {
	    	tablePtr->lastSelected = Ptr;
	    }
	    tablePtr->numSelected ++;
	}
	
	i++;
	if (Ptr->succPtr != NULL)
	{
	    if (!(Ptr->flags&ITEM_HIDDEN_SUBTREE))
	    {
#if 1
		i = TreeTableSelectFromTo
				(tablePtr, Ptr->succPtr,
		    		 i,
				 start, end);
#else
	    	/* we don't need to force the sub tree, if there
		 * well be no changes done
		 */
		if (i+Ptr->succNum+1 < start && i+Ptr->succNum+1 < old_first)
		{
		    i += Ptr->succNum;
		}
		else
		{
		    i = TreeTableSelectFromTo
		    		(tablePtr, Ptr->succPtr,
		    		 i,
				 start, end);
		}
#endif
	    }
	    else
	    {
		i += Ptr->succNum;
	    }
	}	
    }
    return i;
}

static int
TreeTableCountSelected (TableItem* itemPtr, int cnt)
{
    TableItem* Ptr;
    int i = cnt;
    for (Ptr=itemPtr; Ptr!=NULL; Ptr=Ptr->nextPtr)
    {
       if (Ptr->flags&ITEM_SELECTED)
          i++;
       if (Ptr->succPtr != NULL && !(Ptr->flags&ITEM_HIDDEN_SUBTREE))
       {
            i = TreeTableCountSelected (Ptr->succPtr, i);
       }
    }
    return i;
}

static int
TreeTableSetText (TableItem* itemPtr, char*argv[], int cnt)
{
    TableItem* Ptr;
    int i = cnt;
    for (Ptr=itemPtr; Ptr!=NULL; Ptr=Ptr->nextPtr)
    {
	if (Ptr->flags&ITEM_SELECTED)
	{
	    argv[i++] = Ptr->text;
	}
	if (Ptr->succPtr != NULL)
	{
            i = TreeTableSetText (Ptr->succPtr, argv, i);
	}
    }
    return i;
}
/*
 *----------------------------------------------------------------------
 *
 * TreeTableFetchSelection --
 *
 *      This procedure is called back by Tk when the selection is
 *      requested by someone.  It returns part or all of the selection
 *      in a buffer provided by the caller.
 *
 * Results:
 *      The return value is the number of non-NULL bytes stored
 *      at buffer.  Buffer is filled (or partially filled) with a
 *      NULL-terminated string containing part or all of the selection,
 *      as given by offset and maxBytes.  The selection is returned
 *      as a Tcl list with one list element for each item in the
 *      treetable.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int
TreeTableFetchSelection(clientData, offset, buffer, maxBytes)
     ClientData clientData;              /* Information about treetable
					  * widget. */
     int offset;                         /* Offset within selection of first
					  * byte to be returned. */
     char *buffer;                       /* Location in which to place
					  * selection. */
     int maxBytes;                       /* Maximum number of bytes to place
					  * at buffer, not including terminating
					  * NULL character. */
{
    register TreeTable *tablePtr = (TreeTable *) clientData;
    char **argv, *selection;
    int length, count, argc;

    if (!tablePtr->exportSelection) {
        return -1;
    }

    /*
     * Use Tcl_Merge to format the treetable items into a suitable
     * Tcl list.
     */
    argc = TreeTableCountSelected (tablePtr->itemPtr, 0);
    if (argc == 0)
    {
	return -1;
    }
    argv = (char **) ckalloc((unsigned) (argc*sizeof(char *)));
    TreeTableSetText (tablePtr->itemPtr, argv, 0);
    
    selection = Tcl_Merge(argc, argv);

    /*
     * Copy the requested portion of the selection to the buffer.
     */

    length = strlen(selection);
    count = length - offset;
    if (count <= 0) {
        count = 0;
        goto done;
    }
    if (count > maxBytes) {
        count = maxBytes;
    }
    memcpy((VOID *) buffer, (VOID *) (selection + offset), count);

  done:
    buffer[count] = '\0';
    ckfree(selection);
    ckfree((char *) argv);
    return count;
}



/*
 *----------------------------------------------------------------------
 *
 * TreeTableLostSelection --
 *
 *      This procedure is called back by Tk when the selection is
 *      grabbed away from a treetable widget.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The existing selection is unhighlighted, and the window is
 *      marked as not containing a selection.
 *
 *----------------------------------------------------------------------
 */
static int
TreeTableLostSel_x (register TreeTable *tablePtr,
		  register TableItem* itemPtr,
		  int *pos, int first, int last)
{
    register TableItem* Ptr;
    int j = 0;
    
    for (Ptr=itemPtr; tablePtr->numSelected > 0 && *pos<=last && Ptr != NULL;
		Ptr=Ptr->nextPtr)
    {
	if (*pos >= first && *pos<=last && (Ptr->flags&ITEM_SELECTED))
	{
	    Ptr->flags &= ~ITEM_SELECTED;
	    tablePtr->numSelected --;
	    j++;
	}
	*pos += 1;
	if (Ptr->succPtr != NULL)
	{
            j += TreeTableLostSel_x (tablePtr, Ptr->succPtr, pos, first, last);
	}
    }
    return j;
}

static int
TreeTableLostSel (register TreeTable *tablePtr, int first, int last)
{
    TableItem* Ptr=tablePtr->itemPtr;
    int i=0, ret;
    
    if (tablePtr->numSelected == 1 && tablePtr->lastSelected != NULL)
    {
	tablePtr->lastSelected->flags &= ~ITEM_SELECTED;
	tablePtr->lastSelected = NULL;
	
	tablePtr->numSelected = 0;
	ret = 1;
    }
    else
    {
    	FIND_IN_PATHFINDER(first, Ptr, i);
    	return TreeTableLostSel_x(tablePtr, Ptr, &i, first, last);
    }
    tablePtr->numSelected = 0;
    
    return ret;
}

static void
TreeTableLostSelection(clientData)
     ClientData clientData;              /* Information about listbox widget. */
{
    register TreeTable *tablePtr = (TreeTable *) clientData;

    /* set items as not selected */
    if (tablePtr->exportSelection &&
        TreeTableLostSel (tablePtr, 0, tablePtr->numItems-1) > 0)
    {
	TreeTableRedrawRange(tablePtr);
    }
}



/*
 *----------------------------------------------------------------------
 *
 * TreeTableRedrawRange --
 *
 *      Ensure that a given range of elements is eventually redrawn on
 *      the display (if those elements in fact appear on the display).
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Information gets redisplayed.
 *
 *----------------------------------------------------------------------
 */

#if 0
static void
TreeTableRedrawRange(
     register TreeTable *tablePtr  /* Information about widget. */
     )
{
    if ((tablePtr->tkwin != NULL) && Tk_IsMapped(tablePtr->tkwin)
	&& ! (tablePtr->flags & REDRAW_PENDING))
    {
	Tk_DoWhenIdle(DisplayTreeTable, (ClientData) tablePtr);
	tablePtr->flags |= REDRAW_PENDING;
    }
}
#endif
/*
 * Calculate the seen number in a sub tree 
 */
static int
TreeTableCountNotHidden_x (register TableItem*itemPtr)
{
    register TableItem*Ptr;
    register int i = 0;
    
    for (Ptr=itemPtr; Ptr != NULL; Ptr=Ptr->nextPtr)
    {
	i += 1 + Ptr->seenNum;
    }
    return i;
}

/*
 * Calculate the number of seen items on the tree
 */
static int
TreeTableCountNotHidden (register TreeTable*tablePtr, int flag)
{
     if (flag || tablePtr->numViewed == -1)
     {
	register TableItem*Ptr;
	register int i = 0;
	for (Ptr=tablePtr->itemPtr; Ptr != NULL; Ptr=Ptr->nextPtr)
	{
	    i += 1 + Ptr->seenNum;
	}
	tablePtr->numViewed = i;
     }
     return tablePtr->numViewed;
}

/*
 *----------------------------------------------------------------------
 *
 * TreeTableUpdateVScrollbar --
 *
 *      This procedure is invoked whenever information has changed in
 *      a treetable in a way that would invalidate a vertical scrollbar
 *      display.  If there is an associated scrollbar, then this command
 *      updates it by invoking a Tcl command.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      A Tcl command is invoked, and an additional command may be
 *      invoked to process errors in the command.
 *
 *----------------------------------------------------------------------
 */
static void
TreeTableUpdateVScrollbar(
     register TreeTable *tablePtr          /* Information about widget. */
     )
{
    char string[60];
    int result, last;
    int count;

    if (tablePtr->yScrollCmd == NULL)
    {
        return;
    }
    count = TreeTableCountNotHidden (tablePtr, 0);
    
    last = tablePtr->topIndex + tablePtr->numLines - 1;
    if (last >= count)
    {
        last = count-1;
    }
    if (last < tablePtr->topIndex)
    {
        last = tablePtr->topIndex;
    }
    sprintf(string, " %d %d %d %d",
    	    count,
	    tablePtr->numLines,
            tablePtr->topIndex,
	    last);
	    
    Tcl_Preserve((ClientData) tablePtr->interp);
    result = Tcl_VarEval(tablePtr->interp, tablePtr->yScrollCmd, string,
			 (char *) NULL);
    if (result != TCL_OK)
    {
        Tcl_AddErrorInfo(tablePtr->interp,
			 "\n    (vertical scrolling command executed by listbox)");
        Tk_BackgroundError(tablePtr->interp);
    }
    Tcl_Release((ClientData) tablePtr->interp);
}


/*
 *----------------------------------------------------------------------
 *
 * TreeTableUpdateHScrollbar --
 *
 *      This procedure is invoked whenever information has changed in
 *      a treetable in a way that would invalidate a horizontal scrollbar
 *      display.  If there is an associated horizontal scrollbar, then
 *      this command updates it by invoking a Tcl command.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      A Tcl command is invoked, and an additional command may be
 *      invoked to process errors in the command.
 *
 *----------------------------------------------------------------------
 */
static void
TreeTableUpdateHScrollbar(
	register TreeTable *tablePtr /* Information about widget. */
	)
{
    char string[60];
    Tcl_Interp *interp;
    int result, windowWidth;
    double first, last;

    if (tablePtr->xScrollCmd == NULL)
    {
        return;
    }
    windowWidth = TREETABLE_WIDTH(tablePtr);
    if (tablePtr->maxWidth == 0)
    {
	first = 0;
	last = 1.0;
    }
    else
    {
	first = ((double)tablePtr->xOffset)/((double) tablePtr->maxWidth);
	last = (tablePtr->xOffset + windowWidth)
		/((double) tablePtr->maxWidth);
	if (last > 1.0)
	{
	    last = 1.0;
	}
    }

    sprintf(string, " %g %g", first, last);

    /*
     * We must hold onto the interpreter because the data referred to at
     * tablePtr might be freed as a result of the call to Tcl_VarEval.
     */
    
    interp = tablePtr->interp;
    Tcl_Preserve((ClientData) interp);
    result = Tcl_VarEval(interp, tablePtr->xScrollCmd, string,
	    (char *) NULL);
    if (result != TCL_OK) {
	Tcl_AddErrorInfo(interp,
		"\n    (horizontal scrolling command executed by listbox)");
	Tcl_BackgroundError(interp);
    }
    Tcl_Release((ClientData) interp);
}

enum
{
    SRCH_EXACT = 1,
    SRCH_GLOB,
    SRCH_REGEXP,
    SRCH_BEGINS
};

typedef struct
{
    char *pattern, *varName;
    Tcl_DString line;
    int backwards, exact, noCase;
    int matchLength;
    int patLength;
    int index, stopIndex;
    int lineNum;
    Tcl_RegExp regexp;
    int (* cmpProc) (char*, char*, int);
    int code;
} searchParam_t;

static int
TreeTableSearchCmd_x (TreeTable* tablePtr, TableItem* linePtr,
						searchParam_t* prm)
{
    int found = 0;
    char *p;
    
    if (prm->exact != SRCH_EXACT && prm->exact != SRCH_BEGINS)
    {
	Tcl_DStringInit(&prm->line);
    }
    
    for (;linePtr;)
    {
	/* only by searching forwards */
	if (! prm->backwards)
	{
	    if (prm->lineNum < prm->index || prm->lineNum > prm->stopIndex)
	    {
		goto nextLine;
	    }
	}
	else
	{
	    if (prm->lineNum >= tablePtr->numItems) {
	        /*
	         * Don't search the dummy last line of the text.
	         */
	        goto nextLine;
	    }
	    linePtr = TreeTableFindNotHiddenItem(tablePtr, prm->lineNum);
	    if (linePtr == NULL)
	    {
	        goto nextLine;
	    }
	}
	
	if (prm->exact != SRCH_EXACT && prm->exact != SRCH_BEGINS)
	{
	    Tcl_DStringAppend(&prm->line, linePtr->text, linePtr->textLength);
	}
	if (prm->exact == SRCH_REGEXP)
	{
	    Tcl_DStringSetLength(&prm->line, Tcl_DStringLength(&prm->line)-1);
	}

	/*
	 * If we're ignoring case, convert the line to lower case.
	 */
	if (prm->noCase && prm->exact != SRCH_EXACT && prm->exact != SRCH_BEGINS)
	{
	    for (p = Tcl_DStringValue(&prm->line); *p != 0; p++)
	    {
		if (isupper(UCHAR(*p)))
		{
		    *p = tolower(UCHAR(*p));
		}
	    }
	}
	
	if (prm->exact == SRCH_EXACT)
	{
	    if (prm->patLength == linePtr->textLength && prm->cmpProc (linePtr->text, prm->pattern, prm->patLength) == 0)
	    {
		found = 1;  /* string found */
		goto done;
	    }
	}
	else if (prm->exact == SRCH_BEGINS)
	{
	    if (prm->cmpProc (linePtr->text, prm->pattern, prm->patLength) == 0)
	    {
		found = 1;  /* string found */
		goto done;
	    }
	}
	else if (prm->exact == SRCH_REGEXP)
	{
		switch (prm->matchLength = Tcl_RegExpExec(tablePtr->interp, prm->regexp,
					Tcl_DStringValue(&prm->line),
					Tcl_DStringValue(&prm->line)))
		{
		case -1:
			prm->code = TCL_ERROR;
			goto done;
			break;
		case 0:                 /* Not found ! */
			break;
		default:		/* found */
			found = 1;
			goto done;
			break;
			
		}
	}
	else if (prm->exact == SRCH_GLOB)
	{
		if (Tcl_StringMatch(Tcl_DStringValue(&prm->line), prm->pattern))
		{
		    found = 1;
		    goto done;
		}
	}
	
	/*
	 * Go to the next (or previous) line;
	 */
	nextLine:
	if (prm->backwards)
	{
	    prm->lineNum--;
	    if (prm->lineNum < prm->stopIndex)
	    {
		prm->code = TCL_OK;
		return -1;
	    }
	    else if (prm->lineNum < 0)
	    {
		prm->lineNum = tablePtr->numItems-1;
	    }
	}
	else
	{
	    prm->lineNum++;
	    if (linePtr->succPtr != NULL && !(linePtr->flags&ITEM_HIDDEN_SUBTREE))
	    {
		int ret;
		ret = TreeTableSearchCmd_x (tablePtr, linePtr->succPtr, prm);
		if (ret < 0)
		{
		    return ret;
		}
	    }
	    else
	    {
		prm->lineNum += linePtr->succNum;
	    }
	    linePtr = linePtr->nextPtr;
	    if (linePtr == NULL)
	    {
		return 0;
	    }
	    
	    if (prm->lineNum > prm->stopIndex)
	    {
		return -1;
	    }
	    else if (prm->lineNum >= tablePtr->numItems)
	    {
		prm->lineNum = 0;
	    }
	}
	
	if (prm->exact != SRCH_EXACT && prm->exact != SRCH_BEGINS)
	{
	    Tcl_DStringSetLength(&prm->line, 0);
	}
    }
    done:
    
    if (found)
    {
	char tmp[20];
	sprintf (tmp, "%i", prm->lineNum);
	if (prm->varName != NULL)
	{
	    sprintf(tmp, "%d", prm->matchLength);
	    if (Tcl_SetVar(tablePtr->interp, prm->varName, tmp, TCL_LEAVE_ERR_MSG) == NULL)
	    {
		prm->code = TCL_ERROR;
		return -1;
	    }
	}
	Tcl_AppendElement (tablePtr->interp, tmp);
	return -1;
    }
    return 0;
}

static int
TreeTableSearchCmd(TreeTable *tablePtr, int argc, char **argv)
{
    searchParam_t prm;
    int i, argsLeft;
    size_t length;
    int c;
    char *arg, *p;
    Tcl_DString patDString;

    /*
     * Parse switches and other arguments.
     */
    memset (&prm, 0, sizeof (prm));
    prm.exact = SRCH_EXACT;
    prm.backwards = 0;
    prm.noCase = 0;
    prm.varName = NULL;
    for (i = 2; i < argc; i++) {
	arg = argv[i];
	if (arg[0] != '-') {
	    break;
	}
	length = strlen(arg);
	if (length < 2) {
	    badSwitch:
	    Tcl_AppendResult(tablePtr->interp, "bad switch \"", arg,
		    "\": must be -forward, -backward, -exact, -begins, -glob, -regexp, ",
		    "-nocase, -count, or --", (char *) NULL);
	    return TCL_ERROR;
	}
	c = arg[1];
	if ((c == 'b') && (strncmp(argv[i], "-backwards", length) == 0)) {
	    prm.backwards = 1;
	} else if ((c == 'c') && (strncmp(argv[i], "-count", length) == 0)) {
	    if (i >= (argc-1)) {
		tablePtr->interp->result = "no value given for \"-count\" option";
		return TCL_ERROR;
	    }
	    i++;
	    prm.varName = argv[i];
	} else if ((c == 'e') && (strncmp(argv[i], "-exact", length) == 0)) {
	    prm.exact = SRCH_EXACT;
	} else if ((c == 'b') && (strncmp(argv[i], "-begins", length) == 0)) {
	    prm.exact = SRCH_BEGINS;
	} else if ((c == 'f') && (strncmp(argv[i], "-forwards", length) == 0)) {
	    prm.backwards = 0;
	} else if ((c == 'n') && (strncmp(argv[i], "-nocase", length) == 0)) {
	    prm.noCase = 1;
	} else if ((c == 'r') && (strncmp(argv[i], "-regexp", length) == 0)) {
	    prm.exact = SRCH_REGEXP;
	} else if ((c == 'g') && (strncmp(argv[i], "-glob", length) == 0)) {
	    prm.exact = SRCH_GLOB;
	} else if ((c == '-') && (strncmp(argv[i], "--", length) == 0)) {
	    i++;
	    break;
	} else {
	    goto badSwitch;
	}
    }
    argsLeft = argc - (i+2);
    if ((argsLeft != 0) && (argsLeft != 1))
    {
	Tcl_AppendResult(tablePtr->interp, "wrong # args: should be \"",
		argv[0], " search ?switches? pattern index ?stopIndex?",
		(char *) NULL);
	return TCL_ERROR;
    }
    prm.pattern = argv[i];

    /*
     * Convert the pattern to lower-case if we're supposed to ignore case.
     */
    if (prm.noCase && prm.exact != SRCH_EXACT && prm.exact != SRCH_BEGINS)
    {
	Tcl_DStringInit(&patDString);
	Tcl_DStringAppend(&patDString, prm.pattern, -1);
	prm.pattern = Tcl_DStringValue(&patDString);
	for (p = prm.pattern; *p != 0; p++)
	{
	    if (isupper(UCHAR(*p)))
	    {
		*p = tolower(UCHAR(*p));
	    }
	}
    }

    /* verify if the list isn't empty */
    if (tablePtr->itemPtr == NULL) {
    	return TCL_OK;
    }
    
    if (GetTreeTableIndex(tablePtr, argv[i+1], 0, &prm.index) != TCL_OK)
    {
	return TCL_ERROR;
    }
    if (argsLeft == 1)
    {
	if (GetTreeTableIndex(tablePtr, argv[i+2], 0, &prm.stopIndex) != TCL_OK)
	{
	    return TCL_ERROR;
	}
	if (prm.stopIndex >= tablePtr->numItems)
	{
	    prm.stopIndex = tablePtr->numItems-1;
	}
    }
    else
    {
	prm.stopIndex = tablePtr->numItems-1;
    }
    
    if (prm.index >= tablePtr->numItems)
    {
    	prm.index = tablePtr->numItems-1;
    }
    
    /*
     * Scan through all of the lines of the text circularly, starting
     * at the given index.
     */
    prm.matchLength = prm.patLength = 0;	/* Only needed to prevent compiler
					 * warnings. */
    if (prm.exact != SRCH_REGEXP)
    {
	prm.patLength = strlen(prm.pattern);
    }
    else
    {
	prm.regexp = Tcl_RegExpCompile(tablePtr->interp, prm.pattern);
	if (prm.regexp == NULL)
	{
	    return TCL_ERROR;
	}
    }
    
    prm.lineNum = 0;
    prm.cmpProc = (int(*)(char*,char*,int)) ((prm.noCase) ? strncasecmp : strncmp);
    TreeTableSearchCmd_x (tablePtr, tablePtr->itemPtr, &prm);
    
    if (prm.exact != SRCH_EXACT && prm.exact != SRCH_BEGINS)
    {
	Tcl_DStringFree(&prm.line);
	if (prm.noCase)
	{
	    Tcl_DStringFree(&patDString);
	}
    }
    return prm.code;
}

/*
 * Sort items on a column number, options:
 * -column <num>:	column number to sort items on it
 * -nocase:		ignore lower/upper case
 */
#define PosTabStr(s, p, n) i=0; p=s; while(i < n && *p) {\
	if (*p == '\t') i++; \
	p++; \
	if (n == i) \
		break; \
	}

static int sortColNum;
static int sortNoCase;

static int
TreeTableSortCmp(TableItem **Ptr11, TableItem **Ptr22)
{
    TableItem *Ptr1=*Ptr11;
    TableItem *Ptr2=*Ptr22;
    register char *p, *q;
    int ret, i;

    if (sortColNum <= 0)
    {
	p = Ptr1->text;
	q = Ptr2->text;
    }
    else
    {
	PosTabStr (Ptr1->text, p, sortColNum);
	PosTabStr (Ptr2->text, q, sortColNum);
    }
    if (*p == '~') p ++;
    if (*q == '~') q ++;
    
    if (sortNoCase)
    {
    	/* Check first the only first characters! */
    	if ((ret = (tolower(*p) - tolower(*q))) || *p == '\0')
	{
		return ret;
	}
	else
	{
		ret = strcasecmp (p + 1, q + 1);
	}
	/* collapse the string comparision, if the text
	 * is identical
	 */
	if (ret == 0 && sortColNum > 0)
	{
	    ret = strcasecmp (Ptr1->text, Ptr2->text);
	}
    }
    else
    {
    	/* Check first the only first characters! */
	if (*p != *q || *p == '\0')
	{
	    return *p - *q;
	}
	else
	    ret = strcmp (q + 1, q + 1);

	/* collapse the string comparision, if the text
	 * is identical
	 */
	if (ret == 0 && sortColNum > 0)
	{
	    ret = strcmp (Ptr1->text, Ptr2->text);
	}
    }
    return ret;
}

static TableItem *
TreeTableSortFindPrev(TableItem *listPtr, TableItem *itemPtr, int col, int noCase, int *index)
{
    TableItem *Ptr, *posPtr=NULL;
    char *p, *q;
    int ret, i;
    for (Ptr=listPtr; Ptr; Ptr=Ptr->nextPtr)
    {
    	if (col <= 0)
	{
	    p = itemPtr->text;
	    q = Ptr->text;
	}
	else
	{
	    PosTabStr (itemPtr->text, p, col);
	    PosTabStr (Ptr->text, q, col);
	}
    	if (noCase)
	{
	    ret = strcasecmp (p, q);
	    /* collapse the string comparision, if the text
	     * is identicale
	     */
	    if (ret == 0 && col > 0)
	    {
	    	ret = strcasecmp (itemPtr->text, Ptr->text);
	    }
	}
	else
	{
	    if (*p != *q)
	    	ret = *p - *q;
	    else
		ret = strcmp (q, q);
	    /* collapse the string comparision, if the text
	     * is identicale
	     */
	    if (ret == 0 && col > 0)
	    {
	    	ret = strcmp (itemPtr->text, Ptr->text);
	    }
	}
	if (ret <= 0)
	{
	    break;
	}
	if (index)
	{
	    *index += 1 + Ptr->succNum;
	}
	posPtr = Ptr;
    }
    return posPtr;
}
void qqsort(char *base, int n, int size, int (*compar) (const void*,const void*));
static int
TreeTableSort_x(TreeTable *tablePtr, TableItem* parentPtr, int col, int noCase)
{
    TableItem *Ptr, *cutPtr, **list;
    int ret, size, i;
    
    /* sort children of parent children items recursively */
    if (parentPtr)
	Ptr=parentPtr->succPtr;
    else
	Ptr=tablePtr->itemPtr;
    for (; Ptr; Ptr=Ptr->nextPtr)
    {
	if (Ptr->succPtr != NULL)
	{
	    ret = TreeTableSort_x (tablePtr, Ptr, col, noCase);
	    if (ret != TCL_OK)
	    	return ret;
	}
    }
    
    if (parentPtr == NULL)
    {
	cutPtr = tablePtr->itemPtr;
    }
    else
    {
	cutPtr = parentPtr->succPtr;
    }
    
    /* 
     * Make list to use qsort for sorting the list, this is
     * more faster as to insert an item on sorted list
     */
    for (size=0, Ptr=cutPtr; Ptr; Ptr = Ptr->nextPtr, size++)
    {
	/* Empty Loop */
    }
    /*
     * No Items or One Item
     */
    if (size<=1)
    {
	return TCL_OK;
    }
    
    /* cut list from tree table */
    if (parentPtr == NULL)
    {
	tablePtr->itemPtr = NULL;
    }
    else
    {
	parentPtr->succPtr = NULL;
    }
	
    list = malloc (size * sizeof (TableItem*));
    for (i=0, Ptr=cutPtr; Ptr; Ptr = Ptr->nextPtr, i++)
    {
	list [i] = Ptr;
    }
    sortColNum = col;
    sortNoCase = noCase;
    qqsort ((char*) list, size, sizeof (TableItem*), (int (*)(const void*,const void*)) TreeTableSortCmp);
    
    /*
     * now add the sorted list to the tree
     */
    if (parentPtr == NULL)
    {
	tablePtr->itemPtr = list[0];
    }
    else
    {
	parentPtr->succPtr = list[0];
    }
    for (i=1; i<size; i++)
    {
	list[i-1]->nextPtr = list[i];
    }
    list[i-1]->nextPtr = NULL;
    
    /* get allocated memory free */
    free (list);
    return TCL_OK;
}
static int
TreeTableSort(TreeTable *tablePtr, int argc, char **argv)
{
    int col = -1, noCase = 0;
    char * arg, c;
    int i, length, ret;
    for (i=0; i<argc; i++)
    {
    	arg = argv[i];
	c = arg[1];
    	length = strlen (arg);
	if (((c == 'c') && (strncmp(argv[i], "-column", length) == 0))
	    || ((c == 't') && (strncmp(argv[i], "-tab", length) == 0))) {

	    if (i >= (argc-1)) {
		Tcl_AppendResult(tablePtr->interp, "no value given for \"", arg,
				    "\" option",
				    (char *) NULL);
		return TCL_ERROR;
	    }
	    i++;
	    col = atoi (argv[i]);
	    /* if sortedInsertion flag is enabled, then 
	     * save the column number to sort new inserted items
	     * at the same column
	     */
	    if (tablePtr->sortedInsertion)
	    {
		tablePtr->sortColumn = col;
	    }
	} else if ((c == 'n') && (strncmp(argv[i], "-nocase", length) == 0)) {
	    noCase = 1;
	} else {
	    Tcl_AppendResult(tablePtr->interp, "bad switch \"", arg,
		    "\": must be -column <num>, -tab <num> or -nocase ",
		    (char *) NULL);
	    return TCL_ERROR;
	}
    }
    
    /* free cached items */
    FREE_CACHE();
    FREE_PARENT_CACHE();
    
    ret = TreeTableSort_x (tablePtr, NULL, col, noCase);
    TreeTableRedrawRange (tablePtr);
    
    return ret;
}

/*
 * Get the BBox of a treetable item. The important thing here is
 * that the coordinates are returned as the coordinates of the
 * virtual position, that meens the hidden items haven't bbox coordinates
 * and coordinates requierment of an hidden item produces an ERROR.
 */
static int
TreeTableGetBBox (TreeTable *tablePtr, int index, int *x, int *y, int *w, int *h)
{
    TableItem *Ptr;
    *x = *y = *w = *h = 0;
    Ptr = TreeTableFindItem (tablePtr, index);
    if (Ptr == NULL)
    {
	Tcl_AppendResult (tablePtr->interp, "bad treetable index for BBox", NULL);
	return TCL_ERROR;
    }
    Ptr = TreeTableFindNotHiddenItem (tablePtr, index);
    if (Ptr == NULL)
    {
	Tcl_AppendResult (tablePtr->interp,
				"Can't get BBox of item, "
				"the item is in a hidden sub tree", NULL);
	return TCL_ERROR;
    }
    
    *x = ITEM_X(Ptr);
    *y = ITEM_Y(Ptr, index);
    *w = *x + ITEM_WIDTH(Ptr);
    *h = *y + ITEM_HEIGHT();

    return TCL_OK;
}
/*******************************************************************/
/** QUICKSORT ******************************************************/
/** It can be added to another separate file !!!!!                **/
/*******************************************************************/
static void qst(char *base, char *max);
/*
 * qsort.c: Our own version of the system qsort routine which is faster by an
 * average of 25%, with lows and highs of 10% and 50%. The THRESHold below is
 * the insertion sort threshold, and has been adjusted for records of size 48
 * bytes. The MTHREShold is where we stop finding a better median.
 */

#define THRESH  4		       /* threshold for insertion */
#define MTHRESH 6		       /* threshold for median */

static int qsz;			       /* size of each record */
static int thresh;		       /* THRESHold in chars */
static int mthresh;		       /* MTHRESHold in chars */

static int	(*qcmp) (const void*,const void*); /* the comparison routine */
static void	qst (char *base, char *max);
/*
 * qqsort: First, set up some global parameters for qst to share.  Then,
 * quicksort with qst(), and then a cleanup insertion sort ourselves.  Sound
 * simple? It's not...
 */

void
qqsort(char *base, int n, int size, int (*compar) (const void*,const void*))
{
	register char *i;
	register char *j;
	register char *lo;
	register char *hi;
	register char *min;
	register char c;
	char   *max;

	if (n <= 1)
		return;
	qsz = size;
	qcmp = compar;
	thresh = qsz * THRESH;
	mthresh = qsz * MTHRESH;
	max = base + n * qsz;
	if (n >= THRESH) {
		qst(base, max);
		hi = base + thresh;
	} else {
		hi = max;
	}
	/*
	 * First put smallest element, which must be in the first THRESH, in the
	 * first position as a sentinel.  This is done just by searching the
	 * first THRESH elements (or the first n if n < THRESH), finding the min,
	 * and swapping it into the first position.
	 */
	for (j = lo = base; (lo += qsz) < hi;) {
		if ((*qcmp) (j, lo) > 0)
			j = lo;
	}
	if (j != base)
	{		       /* swap j into place */
		for (i = base, hi = base + qsz; i < hi;)
		{
			c = *j;
			*j++ = *i;
			*i++ = c;
		}
	}
	/*
	 * With our sentinel in place, we now run the following hyper-fast
	 * insertion sort. For each remaining element, min, from [1] to [n-1],
	 * set hi to the index of the element AFTER which this one goes. Then, do
	 * the standard insertion sort shift on a character at a time basis for
	 * each element in the frob.
	 */
	for (min = base; (hi = min += qsz) < max;)
	{
		while ((*qcmp) (hi -= qsz, min) > 0);
		if ((hi += qsz) != min)
		{
			for (lo = min + qsz; --lo >= min;)
			{
				c = *lo;
				for (i = j = lo; (j -= qsz) >= hi; i = j)
					*i = *j;
				*i = c;
			}
		}
	}
}

/*
 * qst: Do a quicksort.  First, find the median element, and put that one in
 * the first place as the discriminator.  (This "median" is just the median
 * of the first, last and middle elements).  (Using this median instead of
 * the first element is a big win). Then, the usual partitioning/swapping,
 * followed by moving the discriminator into the right place.  Then, figure
 * out the sizes of the two partions, do the smaller one recursively and the
 * larger one via a repeat of this code.  Stopping when there are less than
 * THRESH elements in a partition and cleaning up with an insertion sort (in
 * our caller) is a huge win. All data swaps are done in-line, which is
 * space-losing but time-saving. (And there are only three places where this
 * is done).
 */
static void qst(char *base, char *max)
{
	register char *i;
	register char *j;
	register char *jj;
	register char *mid;
	register int ii;
	register char c;
	char   *tmp;
	int     lo;
	int     hi;

	lo = (int)(max - base);		/* number of elements as chars */
	do
	{
		/*
		 * At the top here, lo is the number of characters of elements in the
		 * current partition.  (Which should be max - base). Find the median
		 * of the first, last, and middle element and make that the middle
		 * element.  Set j to largest of first and middle.  If max is larger
		 * than that guy, then it's that guy, else compare max with loser of
		 * first and take larger.  Things are set up to prefer the middle,
		 * then the first in case of ties.
		 */
		mid = i = base + qsz * ((unsigned) (lo / qsz) >> 1);
		if (lo >= mthresh)
		{
			j = ((*qcmp) ((jj = base), i) > 0 ? jj : i);
			if ((*qcmp) (j, (tmp = max - qsz)) > 0)
			{
				/* switch to first loser */
				j = (j == jj ? i : jj);
				if ((*qcmp) (j, tmp) < 0)
					j = tmp;
			}
			if (j != i)
			{
				ii = qsz;
				do
				{
					c = *i;
					*i++ = *j;
					*j++ = c;
				} while (--ii);
			}
		}
		/* Semi-standard quicksort partitioning/swapping */
		for (i = base, j = max - qsz;;)
		{
			while (i < mid && (*qcmp) (i, mid) <= 0)
				i += qsz;
			while (j > mid)
			{
				if ((*qcmp) (mid, j) <= 0)
				{
					j -= qsz;
					continue;
				}
				tmp = i + qsz;	       /* value of i after swap */
				if (i == mid)
				{		       /* j <-> mid, new mid is j */
					mid = jj = j;
				}
				else
				{		       /* i <-> j */
					jj = j;
					j -= qsz;
				}
				goto swap;
			}
			if (i == mid)
			{
				break;
			}
			else
			{		       /* i <-> mid, new mid is i */
				jj = mid;
				tmp = mid = i;	       /* value of i after swap */
				j -= qsz;
			}
		swap:
			ii = qsz;
			do
			{
				c = *i;
				*i++ = *jj;
				*jj++ = c;
			} while (--ii);
			i = tmp;
		}
		/*
		 * Look at sizes of the two partitions, do the smaller one first by
		 * recursion, then do the larger one by making sure lo is its size,
		 * base and max are update correctly, and branching back. But only
		 * repeat (recursively or by branching) if the partition is of at
		 * least size THRESH.
		 */
		i = (j = mid) + qsz;
		if ((lo = (int)(j - base)) <= (hi = (int)(max - i)))
		{
			if (lo >= thresh)
				qst(base, j);
			base = i;
			lo = hi;
		}
		else
		{
			if (hi >= thresh)
				qst(i, max);
			max = j;
		}
	} while (lo >= thresh);
}

void
ViewArgs (char *reason, int argc, char *argv[], int mode)
{
    int i;
    if (reason)
    {
	fprintf (stderr, "%s\nused arguments:\n", reason);
    }
    for (i=0; i<argc; i++)
    {
	if (mode)
	{
	    fprintf (stderr, "%s ", argv[i]);
	}
	else
	{
	    fprintf (stderr, "argv[%i] = [%s]\n", i, argv[i]);
	}
    }
    if (mode && mode != 2)
    {
	fprintf (stderr, "\n");
    }
}
#ifdef TEST_TREE

#ifdef USE_CACHE
static void
ViewCache(TreeTable*tablePtr)
{
	int i;
	for (i=0; i<MAX_CACHED; i++)
	{
	    fprintf (stderr, "cache(%i)=[%i, %s]\n",
		i,
		tablePtr->cachedPos [i],
		tablePtr->cachedItem[i] != NULL
			? tablePtr->cachedItem[i]->text
			: "NULL");
	}
	printf ("parentcache[%i, %s]\n",
		tablePtr->cachedParent,
		tablePtr->cachedParentPtr
			? tablePtr->cachedParentPtr->text
			: "NULL");
}
#endif /* USE_CACHE */
#endif /* TEST_TREE */
